module camera_module
  use rtglobal_module
  use amrray_module
  use dust_module
  use lines_module
  use gascontinuum_module
  use stars_module
  use montecarlo_module
  use ioput_module
  use userdef_module
  use mathroutines_module
  use constants_module
  !
  ! This module computes images using ray-tracing. There are two modes of
  ! viewing: one with the observer at infinity (good for most astrophysical
  ! purposes) and one with a local observer (good for public relations movies
  ! for instance). 
  !
  ! ----------------------------------------------------------------------------
  !
  ! Main settings
  !
  !   What do we truly transfer?
  !
  !       1 = The actual radiative transfer [default]
  !      -1 = The total column depth is 'traced' along each ray. 
  !      -2 = The total optical depth is 'traced' along each ray.
  !      -3 = The total optical depth is 'traced' along each ray, and
  !           the point is found where tau_tot-tau=1, i.e. where the
  !           optical depth toward the user is 1 at that frequency.
  !
  integer :: camera_tracemode=1
  !
  !   Do we do first-order integration of the transfer equation (=cell
  !   based, i.e. the default) or second order (=corner based, i.e. much
  !   smoother, and for line radiative transfer allows capturing 
  !   Doppler crossings).
  !
  logical :: camera_secondorder=.false.
  logical :: camera_interpol_jnu=.true.
  !
  !   If we have very narrow lines (i.e. intrinsic line width smaller
  !   than cell-to-cell doppler shifts), then the ray-tracing may go wrong
  !   if no sub-stepping is done in the transfer equation when the line
  !   doppler-shifts through the frequency-of-sight. With this flag we can
  !   switch on this careful kind of integration of the transfer equation.
  !   This assures that the result of the ray-tracing is correct, even 
  !   in the presence of strong velocity gradients and narrow lines. 
  !
  !   NOTE: This does NOT assure that the pixel size of the image is 
  !         sufficiently small to catch the regions of the image where 
  !         the line is visible. In extreme cases (very narrow line, very
  !         strong velocity gradients) these regions can be very narrow,
  !         and thus easily missed with coarse pixel resolution. An example
  !         of such a case would be a keplerian disk around a massive object,
  !         in which T_disk <<< T_virial. See Pontoppidan et al. 2009. 
  !
  !   NOTE: We use the variable lines_widthmargin from the lines_module
  !         to determine the window around the line center within which
  !         we insist on doing sub-stepping should it become necessary.
  !
  logical :: camera_catch_doppler_line=.false.
  !
  !   For the doppler catching algorithm (see above): How fine velocity
  !   resolution do you wish for the sub-stepping? Also in same units.
  !   Currently set to 0.2, which is more than fine enough. Also here
  !   you may experiment with courser values (try 0.4, for instance).
  !   The default is the safe value.
  !
  double precision :: camera_catch_doppler_resolution=0.2d0
  !
  !   This is the main switch for the perspective of the viewer: if this
  !   variable is false, the observer-at-infinity mode is used. If this is
  !   true, the local-observer mode is used.
  !
  logical :: camera_localobserver=.false.
  integer :: camera_localobs_projection=1
  !
  !   If this integer is set to 1, then the stars are included in the images.
  !   In this version they are only included as point sources.
  !
  integer :: camera_incl_stars
  !
  !   Dump subpixeling diagnostics?
  !
  logical :: camera_diagnostics_subpix=.false.
  !
  ! For camera_tracemode=-3 we must set the camera_taustop=1d99 before
  ! the first trace, and then at the second trace it will stop at
  ! optical depth = camera_taustop, where taustop = tautot-tausurface 
  ! so that we can figure out where that point is.
  !
  double precision :: camera_tausurface=1.d0
  double precision, allocatable :: camera_taustop(:)
  double precision, allocatable :: camera_xstop(:),camera_ystop(:),camera_zstop(:)
  double precision, allocatable :: camera_dstop(:)
  double precision, allocatable :: camera_tausurface_x(:,:,:)
  double precision, allocatable :: camera_tausurface_y(:,:,:)
  double precision, allocatable :: camera_tausurface_z(:,:,:)
  !
  ! Variables and arrays
  !
  !   The point to which the camera points is a 3-D point in space. This
  !   is used in both viewing modes. For the observer-at-infinity mode this
  !   is merely responsible for a shift of the camera pointing such that the
  !   center of the image plane is at the pointing position (but see 
  !   camera_zoomcenter_x/y below), while for the local-observer mode this
  !   position is used relative to the camera_observer_position (see below)
  !   to determine the camera viewing direction.
  !
  double precision :: camera_pointing_position(1:3)
  !
  !   The size of the image, from image center to image boundary. I.e. the
  !   total size (from edge to edge) is twice this value. 
  !
  double precision :: camera_image_halfsize_x
  double precision :: camera_image_halfsize_y
  !
  !   In the image plane you can specify where you center your 'ccd camera'.
  !   For the observer-at-infinity mode you can get a similar effect by
  !   adapting the camera_pointing_position. It can then nevertheless be 
  !   useful for zooming-in, not requiring any 3-D transformations to 
  !   convert a sub-field in an existing image into a new camera_pointing_position.
  !   For the local-observer mode the zoomcenter is in fact not reproducable
  !   with a different pointing position because of spherical aberrations. The
  !   zoomcenter stuff is done in the image plane, i.e. the location of the
  !   'ccd camera' in the focal plane of the 'telescope'.
  !
  double precision :: camera_zoomcenter_x
  double precision :: camera_zoomcenter_y
  !
  !   For local observer mode, if a fisheye projection (for planetarium
  !   dome, OMNIMAX) is chosen, his is the shift between the pointing
  !   direction and the zenith of the projection.
  !
  double precision :: camera_localobs_zenith
  !
  !   The orientation of the camera. The center of the image stays put, but
  !   the pixels of the camera are rotated around this point.
  !
  double precision :: camera_pointing_degr_posang
  double precision :: camera_pointing_cos_posang
  double precision :: camera_pointing_sin_posang
  !
  !   For the local-observer mode: this is the position of the observer.
  !   This has no meaning in the observer-at-infinity mode.
  !
  double precision :: camera_observer_position(1:3)
  !
  !   For the flux-conserving mode of imaging, this is the criterion for how
  !   much each pixel should be refined to assure flux conservation. This is
  !   typically of order 1, with smaller values leading to more accurate images
  !   (i.e. each pixel representing the true flux better), but also requiring
  !   more computational power. This does not have meaning for (future??) 
  !   Monte-Carlo style images.
  !
  double precision :: camera_refine_criterion
  double precision :: camera_spher_cavity_relres=0.05d0
!  double precision :: camera_min_aspectratio=0.05d0
  double precision :: camera_min_drr=0.003d0
  double precision :: camera_max_dangle=0.3d0
  double precision :: camera_min_dangle=0.05d0
  integer :: camera_subpixeling_npixfine,camera_subpixeling_npixtot
  !
  !   The number of pixels in x and y image-plane directions.
  !
  integer :: camera_image_nx,camera_image_ny
  !
  !   Flags and variables for the setting of the camera_frequencies array
  !
  integer :: camera_theinu,camera_range_nlam
  double precision :: camera_lambdamic,camera_lambdamic1
  logical :: camera_loadcolor
  logical :: camera_loadlambdas
  logical :: camera_setfreq_global
  !
  !   For the flux-conserving capabilities of the ray-traced images this number
  !   gives the maximum number of times a pixel is allowed to recursively split
  !   into 2x2 sub-pixels. If this number is set to 0, then the ray-traced images
  !   are one-ray-per-pixel: this means that details in the model that are on
  !   scales smaller than the pixel can resolve may be missed entrely, and may
  !   lead to under-estimation of the flux, or if a very bright un-resolved 
  !   detail is hit by coincidence it may lead to over-estimation of the flux
  !   in this pixel. The flux-conserving scheme to avoid this is to recursively
  !   refine the pixel into sub-pixels and once the required resolution is
  !   achieved, add all contributions up (weighted by their sub-pixel size)
  !   to get the 'true' flux of the pixel which is then converted into an
  !   average intensity of the pixel.
  !
  integer :: camera_nrrefine 
  !
  !   For the observer-at-infinity mode these represent the angular position
  !   of the observer at infinity in units of degrees. If theta=0, then the
  !   observer views the stuff from the top (x=y=0, z=infinity). If theta=
  !   pi the observer views from the bottom (x=y=0, z=-infinity). If theta=
  !   pi/2 then the viewer is in the z=0 plane. Other values are obvious from
  !   these specifications. For values of 0<theta<pi the value of phi determines
  !   the location in the remaining degree of freedom. Example: if theta=pi/2
  !   and phi=0 then the observer is located at y=-infinity, x=0, z=0. If theta=pi/2
  !   and phi=pi/2 the observer is x=-infinity, y=0, z=0. 
  !
  double precision :: camera_observer_degr_theta=0.d0
  double precision :: camera_observer_degr_phi=0.d0
  !
  !   Precalculated cosine and sine of the above variables.
  !
  double precision :: camera_observer_cos_theta=0.d0
  double precision :: camera_observer_cos_phi=0.d0
  double precision :: camera_observer_sin_theta=0.d0
  double precision :: camera_observer_sin_phi=0.d0
  !
  !   The spectrum 
  !
  double precision, allocatable :: camera_spectrum(:)
  !
  !   Flag for use or no use of the aperture option for spectra
  ! 
  logical :: camera_use_aperture_info=.false.
  double precision :: camera_observer_distance_pc=0.d0
  integer :: camera_aperture_nf=0
  double precision, allocatable :: camera_aperture_freq(:)
  double precision, allocatable :: camera_aperture_radius_collectarea(:)
  !
  !   For multiple image vantages (optional). This is useful for making
  !   movies where the observer is changing position. Once can then simply
  !   load these positions into these arrays and do all the image rendering
  !   at once.
  !
  double precision, allocatable :: cameras_pt_pos(:,:)
  double precision, allocatable :: cameras_img_hs_x(:),cameras_img_hs_y(:)
  double precision, allocatable :: cameras_zmc_x(:),cameras_zmc_y(:)
  double precision, allocatable :: cameras_pt_degr_pa(:)
  double precision, allocatable :: cameras_obs_pos(:,:)
  double precision, allocatable :: cameras_obs_degr_th(:)
  double precision, allocatable :: cameras_obs_degr_ph(:)
  integer :: cameras_nr_images=0
  !
  !    Some arrays for the images and for rendering
  !
  double precision, allocatable :: camera_intensity(:)
  double precision, allocatable :: camera_rect_image(:,:,:)
  double precision, allocatable :: camera_dustdens(:),camera_dusttemp(:)
  double precision, allocatable :: camera_dustkappa_a(:,:),camera_dustkappa_s(:,:)
  double precision, allocatable :: camera_alpha_a(:),camera_alpha_s(:)
  logical :: camera_warn_resolution
  !
  !    Arrays for templates of stellar source spectra
  !
  double precision, allocatable :: camera_stellarsrc_templates(:,:)
  !
  !    Flag to force RADMC-3D to precompute the source function for all frequencies
  !    of the camera_frequencies array at once.
  !
  logical :: camera_scatsrc_allfreq=.false.
  !
  !    For multi-frequency images with scattering, this flag tells whether
  !    the Monte Carlo scattering source array is also frequency-dispersed, 
  !    or whether it is monochromatic (i.e. one single freq-bin) to save
  !    memory.
  !
  logical :: camera_mcscat_monochromatic=.false.
  !
  !    Integer specifying if we are/were in a star sphere, and if yes,
  !    which one.
  !
  integer :: camera_istar=0
  !
  !    With how many pixels (in each direction) do we need to resolve the
  !    star(s) if we treat them as spheres?
  !
  double precision :: camera_starsphere_nrpix = 20
  !
  !    For backward compatibility with RADMC, here is the circular image stuff
  !
  integer :: cim_nr,cim_np
  double precision, allocatable :: cim_rc(:),cim_ri(:),cim_pc(:),cim_pi(:)
  !
  !    Arrays for the second order integration with corner vertices
  !
  !    NOTE: snu either means S_nu (source function) or j_nu (emissivity)
  !          depending on the setting of camera_interpol_jnu.
  !
  double precision, allocatable :: camera_vertex_snu(:)
  double precision, allocatable :: camera_vertex_anu(:)
  double precision, allocatable :: camera_cell_snu(:)
  double precision, allocatable :: camera_cell_anu(:)
  !
  !    Arrays for doppler catching algorithm
  !
  double precision, allocatable :: camera_vertex_doppler(:)
  double precision, allocatable :: camera_vertex_turb(:)
  double precision, allocatable :: camera_vertex_temp(:)
  double precision, allocatable :: camera_vertex_line_jnubase(:,:)
  double precision, allocatable :: camera_vertex_line_anubase(:,:)
  double precision, allocatable :: camera_cell_doppler(:)
!  double precision, allocatable :: camera_cell_turb(:)
!  double precision, allocatable :: camera_cell_temp(:)
  double precision, allocatable :: camera_cell_line_jnubase(:,:)
  double precision, allocatable :: camera_cell_line_anubase(:,:)
  integer :: camera_vertex_lines_nractivetot=0
  !
  double precision :: camera_local_doppler_curr,camera_local_turb_curr,camera_local_temp_curr
  double precision, allocatable :: camera_local_line_jnubase_curr(:)
  double precision, allocatable :: camera_local_line_anubase_curr(:)
  !
  double precision :: camera_local_doppler_prev,camera_local_turb_prev,camera_local_temp_prev
  double precision, allocatable :: camera_local_line_jnubase_prev(:)
  double precision, allocatable :: camera_local_line_anubase_prev(:)
  !
  double precision :: camera_local_doppler_end,camera_local_turb_end,camera_local_temp_end
  double precision, allocatable :: camera_local_line_jnubase_end(:)
  double precision, allocatable :: camera_local_line_anubase_end(:)
  !
contains


!-------------------------------------------------------------------------
!                   INITIALIZE THE CAMERA MODULE
!-------------------------------------------------------------------------
subroutine camera_init()
  use amr_module
  implicit none
  integer :: ierr,inu,ispec,itempl,iinu
  double precision :: temp,eps
  !
  ! If you wish to use the doppler-catching of lines, then we must be in second-order mode.
  !
  if(camera_catch_doppler_line) then
     camera_secondorder = .true.
  endif
  !
  ! Check that the camera frequency array or choice of frequencies from the
  ! global frequency array is made.
  !
  if((camera_nrfreq.le.0).or.(.not.allocated(camera_frequencies))) then
     write(stdo,*) 'ERROR in camera module: cannot init camera if '
     write(stdo,*) '      the camera_frequencies(:) array is not set.'
     stop
  endif
  !
  ! If already initialized before, deallocate some stuff
  !
  call camera_partial_cleanup()
  !
  ! Check some stuff
  !
  if(camera_image_nx.le.0) then 
     write(stdo,*) 'ERROR in camera module: image_nx = ',camera_image_nx
     stop
  endif
  if(camera_image_ny.le.0) then 
     write(stdo,*) 'ERROR in camera module: image_ny = ',camera_image_ny
     stop
  endif
  if(rt_incl_dust) then
     if(.not.allocated(dust_kappa_abs)) then
        write(stdo,*) 'ERROR in camera module: dust opacities not yet set.'
        stop
     endif
     if(.not.allocated(dustdens)) then
        write(stdo,*) 'ERROR in camera module: dust density array not yet allocated.'
        stop
     endif
     if(.not.allocated(dusttemp)) then
        write(stdo,*) 'ERROR in camera module: dust temperature array not yet allocated.'
        stop
     endif
  else
     if(scattering_mode.ne.0) then
        write(stdo,*) 'ERROR in camera module: if dust not included, cannot have scattering_mode.ne.0.'
        stop
     endif
  endif
  if(rt_incl_lines) then
     if(.not.allocated(lines_aud)) then
        write(stdo,*) 'ERROR in camera module: lines included, but no line data read.'
        stop
     endif
     !
     ! ************ DO MORE CHECKS **************
     !
  endif
  !
  ! Allocate the arrays for the dust 
  !
  if(rt_incl_dust) then
     allocate(camera_dustdens(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_dustdens() array'
        stop
     endif
     allocate(camera_dusttemp(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_dusttemp() array'
        stop
     endif
     allocate(camera_dustkappa_a(1:camera_nrfreq,1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_dustkappa_a()'
        stop 
     endif
     allocate(camera_dustkappa_s(1:camera_nrfreq,1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_dustkappa_s()'
        stop 
     endif
     allocate(camera_alpha_a(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_alpha_a()'
        stop 
     endif
     allocate(camera_alpha_s(1:dust_nr_species),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate camera_alpha_s()'
        stop 
     endif
     !
     ! Get the dust opacities 
     !
     temp    = 100.d0     ! Arbitrary temperature...
     do inu=1,camera_nrfreq
        do ispec=1,dust_nr_species
           camera_dustkappa_a(inu,ispec) =    &
                find_dust_kappa_interpol(camera_frequencies(inu),ispec,temp,1,0,0)
           camera_dustkappa_s(inu,ispec) =    &
                find_dust_kappa_interpol(camera_frequencies(inu),ispec,temp,0,1,0)
        enddo
     enddo
     !
     ! If scattering_mode is 0, then put kappa_s to zero
     ! 
     if(scattering_mode.eq.0) then
!        write(stdo,*) 'In Camera initialization: found that scattering_mode==0, so putting kappa_scat=0.'
        do inu=1,camera_nrfreq
           do ispec=1,dust_nr_species
              camera_dustkappa_s(inu,ispec) = 0.d0
           enddo
        enddo
     endif
  endif
  !
  ! Allocate arrays for the stellar source templates, and
  ! map the templates onto these frequencies
  !
  if(stellarsrc_nrtemplates.gt.0) then
     allocate(camera_stellarsrc_templates(1:camera_nrfreq,1:stellarsrc_nrtemplates),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in camera module: Could not allocate stellarsrc templates.'
        stop
     endif
     do itempl=1,stellarsrc_nrtemplates
        do inu=1,camera_nrfreq
           call hunt(freq_nu,freq_nr,camera_frequencies(inu),iinu)
           if((iinu.lt.0).or.(iinu.gt.freq_nr)) then
              camera_stellarsrc_templates(inu,itempl) = 0.d0
           else
              eps = ( camera_frequencies(inu) - freq_nu(iinu) ) / &
                    ( freq_nu(iinu+1) - freq_nu(iinu) )
              if((eps.lt.0.d0).or.(eps.gt.1.d0)) stop 674
              camera_stellarsrc_templates(inu,itempl) = &
                   (1.d0-eps)*stellarsrc_templates(iinu,itempl) + &
                          eps*stellarsrc_templates(iinu+1,itempl)
           endif
        enddo
     enddo
  endif
  !
  ! Now allocate further arrays
  !
  allocate(camera_spectrum(1:camera_nrfreq),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in camera module: Could not allocate spectrum array.'
     stop
  endif
  allocate(camera_rect_image(1:camera_image_nx,1:camera_image_ny,1:camera_nrfreq),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in camera module: Could not allocate camera_rect_image() array'
     stop
  endif
  allocate(camera_intensity(1:camera_nrfreq),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR in camera module: Could not allocate camera_intensity() array'
     stop
  endif
  !
  ! If we wish to do second order integration of the formal transfer equation,
  ! then we must set up the vertex grid (corner-based grid), or at least
  ! check that it is set up.
  !
  if(camera_secondorder) then
     !
     ! Set up the corner vertex grid
     !
     if(.not.allocated(amr_vertex_cells)) then
        write(stdo,*) 'Setting up the vertices (cell corners) for second order transfer...'
        call amr_set_up_vertices()
        !#########################
        !#########################
        ! MAYBE THERE IS A BUG HERE: WE MUST MAKE A SPECIAL TREATMENT OF THE MIDPLANE IF MIRROR SYMMETRY IS PRESENT.
        ! THIS CAN ONLY EASILY BE FIXED INSIDE AMR_SET_UP_VERTICES VIA AN OPTIONAL ARGUMENT
        ! HOWEVER, THIS IS ONLY A MINOR PROBLEM, SINCE SRC/ALP REMAINS CORRECT; ONLY THE OPACITY OF THE CELL IS SMALLER
        !#########################
        !#########################
     endif
     !
     ! Allocate the j_nu and alpha_nu arrays at the vertices. Note that
     ! it would take presumably too much memory to store the entire
     ! jnu(inu,ivert) array. So we store just one frequency at a time.
     ! This means also that this works only in the one-wavelength-at-
     ! a-time raytracing mode.
     !
     if(.not.allocated(camera_vertex_snu))                    &
          allocate(camera_vertex_snu(1:amr_nr_vertices_max))
     if(.not.allocated(camera_vertex_anu))                    &
          allocate(camera_vertex_anu(1:amr_nr_vertices_max))
     if(.not.allocated(camera_cell_snu))                      &
          allocate(camera_cell_snu(1:amr_nr_vertices_max))
     if(.not.allocated(camera_cell_anu))                      &
          allocate(camera_cell_anu(1:amr_nr_vertices_max))
     !
     ! If we want to do doppler-catching of lines, then we must allocate
     ! further arrays
     !
     if(camera_catch_doppler_line) then
        if(.not.allocated(camera_vertex_doppler))                    &
             allocate(camera_vertex_doppler(1:amr_nr_vertices_max))
        if(.not.allocated(camera_cell_doppler))                    &
             allocate(camera_cell_doppler(1:amr_nrleafs_max))
        if(.not.allocated(camera_vertex_turb))                    &
             allocate(camera_vertex_turb(1:amr_nr_vertices_max))
!        if(.not.allocated(camera_cell_turb))                    &
!             allocate(camera_cell_turb(1:amr_nrleafs_max))
        if(.not.allocated(camera_vertex_temp))                    &
             allocate(camera_vertex_temp(1:amr_nr_vertices_max))
!        if(.not.allocated(camera_cell_temp))                    &
!             allocate(camera_cell_temp(1:amr_nrleafs_max))
     endif
     !
  endif
  !
end subroutine camera_init



!-------------------------------------------------------------------------
!                     FINALIZE THE CAMERA MODULE
!-------------------------------------------------------------------------
subroutine camera_cleanup()
  implicit none
  !
  ! Clean up the private arrays
  !
  call camera_partial_cleanup()
  !
  ! Clean up all the rest. These are the arrays that may have 
  ! to be set by the user or elsewhere in the program.
  !
  ! ...The wavelength specification arrays
  !
  if(allocated(camera_frequencies)) deallocate(camera_frequencies)
  camera_nrfreq = 0
  !
  ! ...The movie-related arrays
  ! 
  if(allocated(cameras_pt_pos)) deallocate(cameras_pt_pos)
  if(allocated(cameras_img_hs_x)) deallocate(cameras_img_hs_x)
  if(allocated(cameras_img_hs_y)) deallocate(cameras_img_hs_y)
  if(allocated(cameras_zmc_x)) deallocate(cameras_zmc_x)
  if(allocated(cameras_zmc_y)) deallocate(cameras_zmc_y)
  if(allocated(cameras_pt_degr_pa)) deallocate(cameras_pt_degr_pa)
  if(allocated(cameras_obs_pos)) deallocate(cameras_obs_pos)
  if(allocated(cameras_obs_degr_th)) deallocate(cameras_obs_degr_th)
  if(allocated(cameras_obs_degr_ph)) deallocate(cameras_obs_degr_ph)
  !
  ! ...The aperture-related arrays
  !
  if(allocated(camera_aperture_freq)) deallocate(camera_aperture_freq)
  if(allocated(camera_aperture_radius_collectarea)) deallocate(camera_aperture_radius_collectarea)
  !
  ! ...Any circular image stuff
  !
  if(allocated(cim_rc)) deallocate(cim_rc)
  if(allocated(cim_ri)) deallocate(cim_ri)
  if(allocated(cim_pc)) deallocate(cim_pc)
  if(allocated(cim_pi)) deallocate(cim_pi)
  !
  ! ...Any vertex grid arrays for the camera module
  !
  if(allocated(camera_vertex_snu)) deallocate(camera_vertex_snu)
  if(allocated(camera_vertex_anu)) deallocate(camera_vertex_anu)
  if(allocated(camera_cell_snu)) deallocate(camera_cell_snu)
  if(allocated(camera_cell_anu)) deallocate(camera_cell_anu)
  !
  ! ...Any doppler catching stuff
  !
  if(allocated(camera_vertex_doppler)) deallocate(camera_vertex_doppler)
  if(allocated(camera_cell_doppler)) deallocate(camera_cell_doppler)
  if(allocated(camera_vertex_turb)) deallocate(camera_vertex_turb)
!  if(allocated(camera_cell_turb)) deallocate(camera_cell_turb)
  if(allocated(camera_vertex_temp)) deallocate(camera_vertex_temp)
!  if(allocated(camera_cell_temp)) deallocate(camera_cell_temp)
  if(allocated(camera_vertex_line_jnubase)) deallocate(camera_vertex_line_jnubase)
  if(allocated(camera_vertex_line_anubase)) deallocate(camera_vertex_line_anubase)
  if(allocated(camera_cell_line_jnubase)) deallocate(camera_cell_line_jnubase)
  if(allocated(camera_cell_line_anubase)) deallocate(camera_cell_line_anubase)
  if(allocated(camera_local_line_jnubase_curr)) deallocate(camera_local_line_jnubase_curr)
  if(allocated(camera_local_line_anubase_curr)) deallocate(camera_local_line_anubase_curr)
  if(allocated(camera_local_line_jnubase_prev)) deallocate(camera_local_line_jnubase_prev)
  if(allocated(camera_local_line_anubase_prev)) deallocate(camera_local_line_anubase_prev)
  if(allocated(camera_local_line_jnubase_end)) deallocate(camera_local_line_jnubase_end)
  if(allocated(camera_local_line_anubase_end)) deallocate(camera_local_line_anubase_end)
  !
end subroutine camera_cleanup


!-------------------------------------------------------------------------
!               RESET THE CAMERA MODULE FOR NEXT ACTION
!
! Here we only clean up the arrays that contain temporary information
! and the arrays that contain results. All these arrays can be deleted
! at the start of a new action without loss of important information.
!-------------------------------------------------------------------------
subroutine camera_partial_cleanup()
  implicit none
  if(allocated(camera_rect_image)) deallocate(camera_rect_image)
  if(allocated(camera_spectrum)) deallocate(camera_spectrum)
  if(allocated(camera_dustdens)) deallocate(camera_dustdens)
  if(allocated(camera_dusttemp)) deallocate(camera_dusttemp)
  if(allocated(camera_intensity)) deallocate(camera_intensity)
  if(allocated(camera_dustkappa_a)) deallocate(camera_dustkappa_a)
  if(allocated(camera_dustkappa_s)) deallocate(camera_dustkappa_s)
  if(allocated(camera_alpha_a)) deallocate(camera_alpha_a)
  if(allocated(camera_alpha_s)) deallocate(camera_alpha_s)
  if(allocated(camera_stellarsrc_templates)) deallocate(camera_stellarsrc_templates)
  if(allocated(camera_xstop)) deallocate(camera_xstop)
  if(allocated(camera_ystop)) deallocate(camera_ystop)
  if(allocated(camera_zstop)) deallocate(camera_zstop)
  if(allocated(camera_dstop)) deallocate(camera_dstop)
  if(allocated(camera_taustop)) deallocate(camera_taustop)
  if(allocated(camera_tausurface_x)) deallocate(camera_tausurface_x)
  if(allocated(camera_tausurface_y)) deallocate(camera_tausurface_y)
  if(allocated(camera_tausurface_z)) deallocate(camera_tausurface_z)
end subroutine camera_partial_cleanup


!-------------------------------------------------------------------------
!            SET THE STARTING POINT AND DIRECTION OF RAY IN IMAGE
!-------------------------------------------------------------------------
subroutine camera_set_ray(px,py,x,y,z,dirx,diry,dirz,distance)
  implicit none
  double precision :: px,py
  double precision :: x,y,z
  double precision :: xbk,ybk,zbk
  double precision :: dirx,diry,dirz
  double precision :: distance,dum,shift,dumy,dumz,cs,sn
  !
  ! First find out how far we should push the starting points of the ray
  ! back to be absolutely sure that the start behind the object.
  !
  if(camera_localobserver) then
     shift = sqrt( camera_observer_position(1)**2 + &
                   camera_observer_position(2)**2 + &
                   camera_observer_position(3)**2 ) 
  else
     shift = sqrt( camera_pointing_position(1)**2 + &
                   camera_pointing_position(2)**2 + &
                   camera_pointing_position(3)**2 ) 
  endif
  !
  ! If star_maxexterndist>0
  !
  if(star_maxexterndist.gt.0.d0) then
     shift = shift + 2*star_maxexterndist
  endif
  !
  ! Grid dependent stuff
  !
  if(igrid_type.lt.100) then
     !
     ! Regular and AMR grids
     !
     if(igrid_coord.lt.100) then
        !
        ! Cartesian coordinates
        !
        if(igrid_coord.eq.10) then
           if(camera_observer_cos_theta.eq.0.d0) then
              write(stdo,*) 'ERROR: In plane-parallel grid you cannot choose'
              write(stdo,*) '       cos(theta)=0 for the camera.'
              stop
           endif
           shift = shift + sqrt(3.d0)*max(abs(amr_grid_xi(1,3)),&
                            abs(amr_grid_xi(amr_grid_nz+1,3)))/ &
                            abs(camera_observer_cos_theta)
        elseif(igrid_coord.eq.20) then
           if(camera_observer_cos_theta.eq.0.d0) then
              write(stdo,*) 'ERROR: In pencil-parallel grid you cannot choose'
              write(stdo,*) '       cos(theta)=0 for the camera.'
              stop
           endif
           shift = shift + sqrt(3.d0)*max(abs(amr_grid_xi(1,2)),&
                            abs(amr_grid_xi(amr_grid_ny+1,2)),  &
                            abs(amr_grid_xi(1,3)),              &
                            abs(amr_grid_xi(amr_grid_nz+1,3)))/ &
                            abs(camera_observer_cos_theta)
        else
           shift = shift + sqrt(3.d0)*max(abs(amr_grid_xi(1,1)),&
                                       abs(amr_grid_xi(amr_grid_nx+1,1)),&
                                       abs(amr_grid_xi(1,2)),&
                                       abs(amr_grid_xi(amr_grid_ny+1,2)),&
                                       abs(amr_grid_xi(1,3)),&
                                       abs(amr_grid_xi(amr_grid_nz+1,3)))
        endif
     elseif(igrid_coord.lt.200) then
        !
        ! Spherical coordinates
        !
        shift = shift + amr_grid_xi(amr_grid_nx+1,1)
     else
        stop 612
     endif
  else
     write(stdo,*) 'ERROR in camera module: non-regular grids not yet ready'
     stop
  endif
  shift = shift*3        ! For safety, put it even further back
  !
  ! Switch between perspective view from a local observer (if localobserver==.true.),
  ! or from an observer 'infinitely far' away (if localobserver==.false.)
  !
  if(camera_localobserver) then
     !
     ! Local perspective: an observer close to (or inside) the object.
     ! Use the camera_observer_position() array to find the rotation
     ! and other things. Moreover, px and py are now to be interpreted
     ! as angular coordinates (in radian).
     !
     ! First make the position, which is at the observer at the moment
     ! 
     x   = camera_observer_position(1)
     y   = camera_observer_position(2)
     z   = camera_observer_position(3)
     !
     ! Now make the 3-D direction vector, depending on which
     ! projection type we use.
     !
     if(camera_localobs_projection.eq.1) then
        !
        ! Simply project onto a flat screen
        !
        dirx = -px
        diry = -py
        dum  = dirx**2 + diry**2
        if(dum.gt.10.d0) then
           write(stdo,*) 'ERROR in camera module: angular coordinates of image'
           write(stdo,*) '      are out of range.'
           stop
        endif
        dum  = 1.d0/sqrt(dum + 1.d0)
        dirx = dirx*dum
        diry = diry*dum
        dirz = dum
     elseif(camera_localobs_projection.eq.2) then
        !
        ! Project onto a sphere: the center of the image is one point on the
        ! sphere. The px and py now point the way along the sphere to the
        ! direction we want to have. 
        !
        dum  = sqrt(px**2+py**2)
        if(dum.gt.0.d0) then
           dirx = -px/dum
           diry = -py/dum
           dirz = cos(dum)
           dum  = sin(dum)
           dirx = dirx*dum
           diry = diry*dum
        else
           dirx = 0.d0
           diry = 0.d0
           dirz = 1.d0
        endif
        !
     else
        write(stdo,*) 'ERROR in camera module: projection type ',camera_localobs_projection,' not known.'
        stop
     endif
     !
     ! Apply a rotation to allow the zenith of the projection to be above 
     ! the pointing direction (so that the viewer in a dome does not have
     ! to always look straight up).
     !
     cs   = cos(camera_localobs_zenith/57.295780d0)
     sn   = sin(camera_localobs_zenith/57.295780d0)
     dumy = diry
     dumz = dirz
     diry =  cs * dumy - sn * dumz
     dirz =  sn * dumy + cs * dumz
     !
     ! The 'position angle' is such that positive z is 'up' and negative z
     ! is 'down'. With that definition the camera is always oriented
     ! 'horizontally' as one would hold a handheld camera. Only with 
     ! the camera_pointing_degr_posang one can change this.
     !
     ! The exact distance for the ray to travel is precisely 'shift'
     !
     distance = shift
     !
  else
     !
     ! Observer at infinite distance (usually OK for astronomical observations).
     ! Use the angles theta and phi to rotate.
     !
     ! First make the starting position in the x-y plane, with z=0
     !
     x = px
     y = py
     z = 0.d0
     !
     ! Now make the direction vector
     !
     dirx = 0.d0
     diry = 0.d0
     dirz = 1.d0   ! Note: direction of light propatation, i.e. toward observer
     !
     ! Set distance to observer to infinity (=1d99)
     !
     distance = 1d99
     !
     ! For the camera at infinity we must, in addition to the direction
     ! vectors (see below) also rotate the x,y,z positions, while for the
     ! perspective view we only rotate the direction vector. 
     !
     ! NOTE: We use the right-hand-rule for the orientation of the 
     !       x,y,z axes. So if you look at the x-y plane from the top,
     !       with the z-axis pointing toward you, then the x and y 
     !       axes are pointing in the usual way: positive x toward the
     !       right, positive y toward the top. Note that this is also
     !       the convention for the images.
     !
     ! First a rotation of the camera itself. The camera is rotated
     ! in clockwise direction. Therefore, on the camera CCD any image is 
     ! rotated in counter-clockwise direction. 
     !
     xbk = x
     ybk = y
     x   = camera_pointing_cos_posang * xbk + camera_pointing_sin_posang * ybk
     y   =-camera_pointing_sin_posang * xbk + camera_pointing_cos_posang * ybk
     !
     ! Then a rotation in the y,z plane, i.e. going from pole-on
     ! to more face-on (if a disk is assumed to be present in the x-y plane)
     !
     ybk = y
     zbk = z
     y   = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
     z   = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
     !
     ! Then a rotation in the x,y plane, rotating horizontally around
     ! the object. If the camera looks toward the object, then the
     ! camera now moves to the left (clockwise around the object).
     !
     xbk = x
     ybk = y
     x   = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
     y   =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     !
     ! Now shift the entire stuff toward the point of focus
     !
     x   = x + camera_pointing_position(1)
     y   = y + camera_pointing_position(2)
     z   = z + camera_pointing_position(3)
     !
  endif
  !
  ! Now perform the rotation of the direction vectors
  ! 
  ! NOTE: We use the right-hand-rule for the orientation of the 
  !       x,y,z axes. So if you look at the x-y plane from the top,
  !       with the z-axis pointing toward you, then the x and y 
  !       axes are pointing in the usual way: positive x toward the
  !       right, positive y toward the top. Note that this is also
  !       the convention for the images.
  !
  ! First a rotation of the camera itself. The camera is rotated
  ! in clockwise direction. Therefore, on the camera CCD any image is 
  ! rotated in counter-clockwise direction. Note: this is only
  ! necessary for the direction vectors in the perspective mode, hence
  ! the if-statement.
  !
  if(camera_localobserver) then
     xbk  = dirx
     ybk  = diry
     dirx = camera_pointing_cos_posang * xbk + camera_pointing_sin_posang * ybk
     diry =-camera_pointing_sin_posang * xbk + camera_pointing_cos_posang * ybk
  endif
  !
  ! Then a rotation in the y,z plane, i.e. going from pole-on
  ! to more face-on (if a disk is assumed to be present in the x-y plane)
  !
  ybk  = diry
  zbk  = dirz
  diry = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
  dirz = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
  !
  ! Then a rotation in the x,y plane, rotating horizontally around
  ! the object. If the camera looks toward the object, then the
  ! camera now moves to the left (clockwise around the object).
  !
  xbk  = dirx
  ybk  = diry
  dirx = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
  diry =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
  if(abs(dirx**2+diry**2+dirz**2-1.d0).gt.1d-6) then
     write(stdo,*) 'ERROR in camera module: direction vector not OK.'
     write(stdo,*) dirx,diry,dirz
     stop
  endif
  !
  ! Now shift
  !
  x = x - shift * dirx
  y = y - shift * diry
  z = z - shift * dirz
  !
end subroutine camera_set_ray


!-------------------------------------------------------------------------
!            SET THE STARTING POINT AND DIRECTION FOR THE STARS
!-------------------------------------------------------------------------
subroutine camera_set_ray_stars_pntsrc(istar,x,y,z,dirx,diry,dirz,px,py,distance)
  implicit none
  integer :: istar
  double precision :: x,y,z
  double precision :: xbk,ybk,zbk
  double precision :: dirx,diry,dirz
  double precision :: px,py
  double precision :: distance
  !
  ! Check
  !
  if(.not.allocated(star_pos)) then
     write(stdo,*) 'ERROR in camera module: star positions not allocated'
     stop
  endif
  if((istar.lt.1).or.(istar.gt.nstars)) then
     write(stdo,*) 'ERROR in camera module: istar out of range'
     stop
  endif
  !
  ! The starting position is clearly the star position
  !
  ! NOTE: The stars are treated as point sources!!!
  !
  x   = star_pos(1,istar)
  y   = star_pos(2,istar)
  z   = star_pos(3,istar)
  !
  ! Check which perspective we use
  !
  if(camera_localobserver) then
     !
     ! If the local perspective is used
     !
     ! Make the direction vector, pointing toward the observer
     ! At the same time also compute the distance
     !
     dirx  = camera_observer_position(1) - star_pos(1,istar)
     diry  = camera_observer_position(2) - star_pos(2,istar)
     dirz  = camera_observer_position(3) - star_pos(3,istar)
     distance = sqrt( dirx**2 + diry**2 + dirz**2 )
     dirx  = dirx / distance
     diry  = diry / distance
     dirz  = dirz / distance
     !
     ! Find the location on the image
     !
     ! A rotation in the x,y plane, rotating horizontally around
     ! the object. 
     !
     xbk   = -dirx
     ybk   = -diry
     px    = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
     py    = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     !
     ! Then a rotation in the y,z plane
     !
     ybk   = py
     zbk   = -dirz
     py    = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
     !
  else
     !
     ! If observer-at-infinity is used
     !
     ! Make the direction vector
     !
     dirx = 0.d0
     diry = 0.d0
     dirz = 1.d0
     ybk  = diry
     zbk  = dirz
     diry = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
     dirz = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
     xbk  = dirx
     ybk  = diry
     dirx = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
     diry =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     if(abs(dirx**2+diry**2+dirz**2-1.d0).gt.1d-6) then
        write(stdo,*) 'ERROR in camera module: direction vector not OK.'
        write(stdo,*) dirx,diry,dirz
        stop
     endif
     !
     ! Find the location on the image
     !
     ! A rotation in the x,y plane, rotating horizontally around
     ! the object. 
     !
     xbk   = x - camera_pointing_position(1)
     ybk   = y - camera_pointing_position(2)
     px    = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
     py    = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     !
     ! Then a rotation in the y,z plane
     !
     ybk   = py
     zbk   = z - camera_pointing_position(3)
     py    = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
     !
     ! The distance is infinite
     ! 
     distance = 1d99
  endif
  !
end subroutine camera_set_ray_stars_pntsrc


!-------------------------------------------------------------------------
!                    RAY TRACE THROUGH THE MODEL
!
! This is the normal ray-tracing routine (see also the subroutine
! camera_ray1d_raytrace which is an alternative). 
!-------------------------------------------------------------------------
subroutine camera_serial_raytrace(nrfreq,inu0,inu1,x,y,z,dx,dy,dz,distance,   &
                           celldxmin,intensity)
  implicit none
  double precision :: x,y,z,dx,dy,dz,dum,distance,celldxmin
  double precision :: r,theta,phi
  double precision :: intensity(1:nrfreq)
  integer :: ispec,nrfreq,deepestlevel
  integer :: inu,levelnext,itemplate,inu0,inu1,iddr
  logical :: arrived
  double precision :: xp,xp1,dummy,dummy2,epstau
  double precision :: dumm,dumr,dumt,dump,dumx,dumy,dumz
  double precision :: src(1:nrfreq),alp(1:nrfreq)
  type(amr_branch), pointer :: a,b
  doubleprecision :: bxi(1:2,1:3),bxc(1:3)
  doubleprecision :: snu_prev,anu_prev,jnu_prev,snu_curr,anu_curr,jnu_curr
  doubleprecision :: ln_anu_prev,ln_jnu_prev,ln_anu_curr,ln_jnu_curr
  doubleprecision :: dtau1,theomax,e0,e1,ax,bx,qdr,proj
  doubleprecision :: nstp,ap,jp,sp,ac,jc,sc,nu,nu0,nu0_prev,nu0_curr
  doubleprecision :: duc,eps,eps1,eps0,temp,resol,margin
  integer :: nsteps,iline,ilactive,istep
  integer :: ixx,iyy,izz,bc_idir,bc_ilr
  double precision :: xbk,ybk,zbk,spx,spy,spz,rx1,ry1,rz1,rx0,ry0,rz0
  !
  ! Do checks
  !
  if(nrfreq.ne.camera_nrfreq) stop 9301
  if(debug_check_all.eq.1) then
     dum = sqrt(dx**2 + dy**2 + dz**2)
     if(abs(dum-1.d0).gt.1d-6) stop 3487
     if(.not.rt_incl_dust) then
        write(stdo,*) 'ERROR in camera_serial_raytrace: dust is not activated...'
        stop
     endif
!     if(.not.allocated(camera_ray_jnu)) stop 4560
!     if(.not.allocated(camera_ray_alpnu)) stop 4561
  endif
  !
  ! Set defaults
  !
  snu_prev = 0.d0
  anu_prev = 0.d0
  jnu_prev = 0.d0
  snu_curr = 0.d0
  anu_curr = 0.d0
  jnu_curr = 0.d0
  if(camera_catch_doppler_line) then
     ln_anu_prev = 0.d0
     ln_jnu_prev = 0.d0
     ln_anu_curr = 0.d0
     ln_jnu_curr = 0.d0
     camera_local_doppler_prev = 0.d0
     camera_local_turb_prev = 0.d0
     camera_local_temp_prev = 0.d0
     camera_local_line_jnubase_prev(:) = 0.d0
     camera_local_line_anubase_prev(:) = 0.d0
     camera_local_doppler_curr = 0.d0
     camera_local_turb_curr = 0.d0
     camera_local_temp_curr = 0.d0
     camera_local_line_jnubase_curr(:) = 0.d0
     camera_local_line_anubase_curr(:) = 0.d0
     camera_local_doppler_end = 0.d0
     camera_local_turb_end = 0.d0
     camera_local_temp_end = 0.d0
     camera_local_line_jnubase_end(:) = 0.d0
     camera_local_line_anubase_end(:) = 0.d0
  endif
  !
  ! Check the dust scattering
  !
  if(scattering_mode.ge.1) then
     if(scattering_mode.eq.1) then
        !
        ! Isotropic scattering mode
        !
        if(mcscat_nrdirs.ne.1) then
           if(mcscat_nrdirs.eq.0) then
              write(stdo,*) 'ERROR: For isotropic scattering mode we must have'
              write(stdo,*) '       the scattering source function computed.'
              stop
           else
              write(stdo,*) 'ERROR: For isotropic scattering mode we need not'
              write(stdo,*) '       (and should not) use multiple directions.'
              stop
           endif
        endif
     elseif(scattering_mode.ge.2) then
        !
        ! Non-isotropic scattering mode. Check if the directions are
        ! set properly and that our current direction is equal to that
        ! of the monte carlo pre-computed source function.
        ! Note: Use mcscat_current_dir to select among the pre-computed
        !       discrete directions. 
        !
        if(mcscat_current_dir.gt.mcscat_nrdirs) stop 3012
        if(.not.allocated(mcscat_dirs)) stop 3013
        if((abs(mcscat_dirs(1,mcscat_current_dir)-dx).gt.1d-6).or.     &
           (abs(mcscat_dirs(2,mcscat_current_dir)-dy).gt.1d-6).or.     &
           (abs(mcscat_dirs(3,mcscat_current_dir)-dz).gt.1d-6)) then
           write(stdo,*) 'ERROR in camera module: attempting to use'
           write(stdo,*) '      scattering source for a direction that'
           write(stdo,*) '      is not the direction of the selected '
           write(stdo,*) '      precomputed scattering source function.'
           stop
        endif
     endif
     !
     ! General checks for scattering
     !
     if(.not.allocated(mcscat_scatsrc)) stop 3005
     if(camera_mcscat_monochromatic) then
        !
        ! Scattering is done one frequency at a time, so the scattering
        ! source function has only one bin in frequency space.
        !
        ! Check that inu1.eq.inu0, because if not, then this method cannot
        ! work properly.
        !
        if(inu0.ne.inu1) then
           write(stdo,*) 'INTERNAL ERROR in ray-tracing with scattering:'
           write(stdo,*) '      If you use monochromatic scattering source '
           write(stdo,*) '      function, then you can ray trace only one'
           write(stdo,*) '      frequency at a time. However, I obtained'
           write(stdo,*) '      inu0=',inu0,' and inu1=',inu1
           stop
        endif
        !
        ! Check that mc_nrfreq.eq.1
        !
        if(mc_nrfreq.ne.1) then
           write(stdo,*) 'INTERNAL ERROR in ray-tracing with scattering:'
           write(stdo,*) '      mc_nrfreq.ne.1 while the monochromatic'
           write(stdo,*) '      style of scattering Monte Carlo is used.'
           stop
        endif
        !
     else
        !
        ! Scattering is done spectrally dispersed, i.e. the scattering 
        ! source function has been computed for all camera frequencies
        ! beforehand and stored for all these frequencies in presumably
        ! a very large array. This method is presumably not so useful
        ! for large models, because it would require too much memory
        ! space. But it may be faster than the frequency-by-frequency
        ! method (i.e. the monochromatic method, see above).
        !
        if(mc_nrfreq.ne.camera_nrfreq) then
           write(stdo,*) 'ERROR in camera module when using scattering: '
           write(stdo,*) '      the frequency grid of the camera is not'
           write(stdo,*) '      equal to that of the montecarlo...' 
           stop
        endif
        do inu=1,camera_nrfreq
           if(camera_frequencies(inu).ne.mc_frequencies(inu)) then
              write(stdo,*) 'ERROR in camera module when using scattering: '
              write(stdo,*) '      the frequency grid of the camera is not'
              write(stdo,*) '      equal to that of the montecarlo...' 
              stop
           endif
        enddo
     endif
  endif
  !
  ! Set starting value of celldxmin, i.e. the size of the smallest cell
  ! we will pass through. Note that for cartesian coordinates, since cell 
  ! sizes will then always be cubes of the base size times (1/2)^level
  ! we can use just the deepest level to compute celldxmin.
  !
  ! NOTE: In the AMR module we now changed the level-numbering: Now
  !       the base grid level is level=0 (used to be 1). Hence we must
  !       now set deepestlevel = 0 here, and the multiplication factor
  !       is now (1/2)^level instead of (1/2)^(level-1).
  !
  celldxmin     = 1d99
  deepestlevel  = 0
  !
  ! Set the starting point and the direction
  !
  ray_cart_x    = x
  ray_cart_y    = y
  ray_cart_z    = z
  ray_cart_dirx = dx
  ray_cart_diry = dy
  ray_cart_dirz = dz
  !
  ! If mirror symmetry is switched on, then we must assure that 
  ! z is above the midplane
  !
  if(amrray_mirror_equator) then
     if(ray_cart_z.eq.0.d0) then
        ray_cart_dirz = abs(ray_cart_dirz)
     elseif(ray_cart_z.lt.0.d0) then
        ray_cart_z    = -ray_cart_z
        ray_cart_dirz = -ray_cart_dirz
     endif
  endif
  !
  ! Put the distance from the starting point to the observer here
  ! If observer at infinity, simply put distance=1d99
  !
  ray_dsend     = distance 
  !
  ! Check if we start inside or outside the domain
  !
  if(igrid_type.lt.100) then 
     !
     ! We use a regular or AMR grid
     !
     if(igrid_coord.lt.100) then
        !
        ! We use cartesian coordinates
        !
        if(amr_tree_present) then
           !
           ! We use the AMR tree
           !
           call amr_findcell(x,y,z,a)
           if(associated(a)) then
              !
              ! We start inside the domain
              !
              ray_index     = a%leafindex
              ray_indexnext = 0
           else
              !
              ! We start outside the domain, so all indices must be 0
              !
              ray_index     = 0
              ray_indexnext = 0
           endif
        else
           !
           ! We have a regular grid
           !
           call amr_findbasecell(x,y,z,ixx,iyy,izz)
           if(ixx.gt.0) then
              ray_index     = ixx+((iyy-1)+(izz-1)*amr_grid_ny)*amr_grid_nx
              ray_indexnext = 0
           else
              ray_index     = 0
              ray_indexnext = 0
           endif
        endif
        !
     elseif(igrid_coord.lt.200) then
        !
        ! We use spherical coordinates
        !
        call amr_xyz_to_rthphi(ray_cart_x,ray_cart_y,ray_cart_z,r,theta,phi)
        if(amr_tree_present) then
           !
           ! We use the AMR tree
           !
           call amr_findcell(r,theta,phi,a)
           if(associated(a)) then
              !
              ! We start inside the domain
              !
              ray_index     = a%leafindex
              ray_indexnext = 0
           else
              !
              ! We start outside the domain, so all indices must be 0
              !
              ray_index     = 0
              ray_indexnext = 0
           endif
        else
           !
           ! We have a regular grid
           !
           call amr_findbasecell(r,theta,phi,ixx,iyy,izz)
           if(ixx.gt.0) then
              ray_index     = ixx+((iyy-1)+(izz-1)*amr_grid_ny)*amr_grid_nx
              ray_indexnext = 0
           else
              ray_index     = 0
              ray_indexnext = 0
           endif
        endif
!        write(stdo,*) 'FINDING CELL NOT YET READY FOR SPHERICAL COORDINATES'
!        stop
        !
     else
        write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
        stop
     endif
  else
     write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
     stop
  endif
  !
  ! Now trace
  !
  arrived = .false.
  do while(.not.arrived)
     !
     ! Back up the current position
     !
     ray_prev_x     = ray_cart_x
     ray_prev_y     = ray_cart_y
     ray_prev_z     = ray_cart_z
     !
     ! Find new position
     !
     if(igrid_type.lt.100) then 
        !
        ! We use a regular or AMR grid
        !
        if(igrid_coord.lt.100) then
           !
           ! We use cartesian coordinates
           !
           call amrray_find_next_location_cart(ray_dsend,            &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived,              &
                levelnext=levelnext)
           !
           ! Check if this cell is the smallest so far
           !
           deepestlevel = max(deepestlevel,levelnext)
           !
           ! If we went through one of the boundaries of the
           ! domain, we might need to implement the thermal
           ! boundary, if present
           !
           if(incl_thermbc.ne.0) then
              if((ray_index.eq.0).and.(ray_indexnext.ne.0)) then
                 !
                 ! First determine which boundary we just went
                 ! through 
                 !
                 if(amrray_icross.ge.0) stop 3902  ! Self-consistency check
                 if(amrray_icross.eq.-1) then
                    bc_idir = 1
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-2) then
                    bc_idir = 1
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-3) then
                    bc_idir = 2
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-4) then
                    bc_idir = 2
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-5) then
                    bc_idir = 3
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-6) then
                    bc_idir = 3
                    bc_ilr  = 2
                 else
                    stop 4944
                 endif
                 !
                 ! Now set the intensity to the thermal
                 ! emission of the boundary, if the
                 ! thermal boundary is switched on
                 !
                 if(thermal_bc_active(bc_ilr,bc_idir)) then
                    do inu=1,camera_nrfreq
                       intensity(inu) = bplanck(thermal_bc_temp(bc_ilr,bc_idir), &
                                                camera_frequencies(inu))
                    enddo
                 endif
              endif
           endif
           !
        elseif(igrid_coord.lt.200) then
           !
           ! We use spherical coordinates
           !
           call amrray_find_next_location_spher(ray_dsend,           &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived)
           !
           ! Check if this cell is the smallest so far. However, this is far
           ! more subtle than for the Cartesian case where all cells are
           ! forced to be precisely cubic. Here, for spherical coordinates,
           ! we can have very flattened cells. For instance, near the inner
           ! rim of the coordinate system we may have very much refinement
           ! in R (apart from AMR!), so the cells are nearly like sheats
           ! perpendicular to radially outward rays. Or near the midplane
           ! we could have strong refinement for some other reason.
           ! So we must take projection into account.
           !
           ! Note: We approximate sin(theta)=theta for theta<pi/2
           !       and pi-theta for theta>pi/2
           !
           if(ray_indexnext.gt.0) then
              if(amr_tree_present) then
                 !
                 ! We get our information about the cell dimensions 
                 ! from the AMR tree
                 !
                 b => amr_index_to_leaf(ray_indexnext)%link
                 if(.not.associated(b)) stop 7130
                 !
                 ! Get the cell walls and centers
                 !
                 do iddr=1,3
                    bxc(iddr)   = amr_finegrid_xc(b%ixyzf(iddr),iddr,b%level)
                    bxi(1,iddr) = amr_finegrid_xi(b%ixyzf(iddr),iddr,b%level)
                    bxi(2,iddr) = amr_finegrid_xi(b%ixyzf(iddr)+1,iddr,b%level)
                 enddo
              else
                 !
                 ! We get our information about the cell dimensions
                 ! from the regular grid 
                 !
                 call amr_regular_get_ixyz(ray_indexnext,amrray_ix_next,amrray_iy_next,amrray_iz_next)
                 if(amrray_ix_next.le.0) stop 7130
                 bxc(1)   = amr_finegrid_xc(amrray_ix_next,1,0)
                 bxc(2)   = amr_finegrid_xc(amrray_iy_next,2,0)
                 bxc(3)   = amr_finegrid_xc(amrray_iz_next,3,0)
                 bxi(1,1) = amr_finegrid_xi(amrray_ix_next,1,0)
                 bxi(2,1) = amr_finegrid_xi(amrray_ix_next+1,1,0)
                 bxi(1,2) = amr_finegrid_xi(amrray_iy_next,2,0)
                 bxi(2,2) = amr_finegrid_xi(amrray_iy_next+1,2,0)
                 bxi(1,3) = amr_finegrid_xi(amrray_iz_next,3,0)
                 bxi(2,3) = amr_finegrid_xi(amrray_iz_next+1,3,0)
              endif
              !
              ! The smallest size in R direction
              !
              dumr = bxi(2,1) - bxi(1,1)
              dumr = max(dumr,bxc(1)*camera_min_drr)
              !
              ! The smallest size in Theta direction
              !
              dumt = (bxi(2,2) - bxi(1,2))
              dumt = bxc(1)*max(min(dumt,camera_max_dangle),camera_min_dangle)
              !
              ! The smallest size in Phi direction
              !
              if(bxc(2).lt.1.d0) then
                 dumm = bxc(2)
              elseif(pi-bxc(2).lt.1.d0) then
                 dumm = pi-bxc(2)
              else
                 dumm = 1.d0
              endif
              dump = dumm*abs(bxi(2,2) - bxi(1,2))
              dump = bxc(1)*max(min(dump,camera_max_dangle),camera_min_dangle)
              !
              ! Compute the projections of the direction on the coordinate
              ! vector
              !
              r     = sqrt(ray_cart_x**2+ray_cart_y**2+ray_cart_z**2)
              proj  = abs( ray_cart_dirx*ray_cart_x +  &
                           ray_cart_diry*ray_cart_y +  &
                           ray_cart_dirz*ray_cart_z ) / r
!              proj = max(proj,camera_min_aspectratio)
              dummy = abs(ray_cart_dirz)
              !
              ! Compute estimates (!) of the projection of cell theta-width
              ! and cell phi-width onto image plane. This is just a simple
              ! estimation based on the z-component of the direction vector.
              ! If we view things face-on (ray_cart_dirz=1 or -1) then 
              ! the theta-size will be subject to projection, while
              ! if we view things edge-on (ray_cart_dirz=0) then 
              ! the phi-size will be subject to projection
              !
              celldxmin = min(celldxmin,max(dumr, &
                   min(dumt,dump,(dummy*dumt+(1.d0-dummy)*dump)*proj)))
           else
              !
              ! Check if the ray passes within the central hole of
              ! the coordinate system. 
              !
              dummy = ray_cart_dirx*ray_cart_x +  &
                      ray_cart_diry*ray_cart_y +  &
                      ray_cart_dirz*ray_cart_z 
              dumx  = ray_cart_x - dummy*ray_cart_dirx
              dumy  = ray_cart_y - dummy*ray_cart_diry
              dumz  = ray_cart_z - dummy*ray_cart_dirz
              dumr  = sqrt(dumx**2+dumy**2+dumz**2)
              if(dumr.le.amr_grid_xi(1,1)) then
                 celldxmin = min(celldxmin,                            &
                      camera_spher_cavity_relres * amr_grid_xi(1,1) )
              endif
           endif
           !
        else
           write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
           stop
        endif
        !
        ! Set the camera_istar, which is used for treating finite-size
        ! stars (instead of point source stars)
        !
        camera_istar = amrray_ispherehit
        !   
     else
        write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
        stop
     endif
     !
     ! Path length
     ! 
     ! ######### CHECK: WHY NOT USE ray_ds from the above routines ??? ########
     !
     ray_ds    = sqrt( (ray_cart_x-ray_prev_x)**2 + (ray_cart_y-ray_prev_y)**2 + (ray_cart_z-ray_prev_z)**2 )
     !
     ! Do a simple safet check
     !
     if(ray_index.gt.nrcellsmax) stop 6271
     !
     ! Check if we are/were in a star
     !
     if(camera_istar.le.0) then
        !
        ! Normal situation: we are not inside a star (istar=0), or we will
        ! hit the star but are not inside the star yet (istar<0).
        !
        ! Decide whether we do 1st order or 2nd order integration
        !
        if(camera_secondorder) then
           !
           ! Second order integration. Corner-based (=vertex-based) scheme.
           ! Here the continuum sources have all been pre-calculated on
           ! the cell corners. For the lines: they are also pre-calculated
           ! unless the camera_catch_doppler_line==.true., because in
           ! that case the line stuff may have to be done locally at smaller
           ! steps. 
           !
           ! Switch between integration of radiation field and 
           ! integration of the opacity.
           !
           if((camera_tracemode.eq.1).or.(camera_tracemode.le.-2)) then
              !
              ! Do formal radiative transfer
              !
              inu = inu0
              !
              ! Backup snu and anu at previous crossing
              !
              snu_prev = snu_curr
              anu_prev = anu_curr
              jnu_prev = jnu_curr
              !
              ! Also backup stuff specifically for lines, if the
              ! doppler catching method is switched on
              !
              if(camera_catch_doppler_line) then
                 ln_anu_prev = ln_anu_curr
                 ln_jnu_prev = ln_jnu_curr
                 camera_local_doppler_prev = camera_local_doppler_curr
                 camera_local_turb_prev = camera_local_turb_curr
                 camera_local_temp_prev = camera_local_temp_curr
                 camera_local_line_jnubase_prev(:) = camera_local_line_jnubase_curr(:)
                 camera_local_line_anubase_prev(:) = camera_local_line_anubase_curr(:)
              endif
              !
              ! Find snu and anu at new crossing
              !
              ! NOTE: If camera_interpol_jnu is set, then snu is actually
              !       j_nu (the emissivity). It will be translated below.
              !
              if(igrid_type.lt.100) then 
                 if(igrid_coord.lt.100) then
                    !
                    ! Cartesian
                    !
                    call camera_find_srcalp_interpol(       &
                         ray_cart_x,ray_cart_y,ray_cart_z,  &
                         snu_curr,anu_curr)
                 elseif(igrid_coord.lt.200) then
                    !
                    ! Spherical 
                    !
                    call amr_xyz_to_rthphi(ray_cart_x,ray_cart_y,ray_cart_z, &
                                           r,theta,phi)
                    call camera_find_srcalp_interpol(    &
                         r,theta,phi,snu_curr,anu_curr)
                 else
                    stop 498
                 endif
              else
                 stop 499
              endif
              !
              ! If camera_interpol_jnu is set, then what we get back
              ! from the camera_find_srcalp_interpol() is in fact the
              ! emissivity j_nu instead of the source function S_nu.
              ! So translate either snu to jnu or vice versa.
              !
              if(camera_interpol_jnu) then
                 jnu_curr = snu_curr
                 snu_curr = jnu_curr / (anu_curr+1d-99)
              else
                 jnu_curr = snu_curr * (anu_curr+1d-99)
              endif
              !
              if(ray_index.gt.0) then
                 !
                 ! Decide if we do "normal" second order integration, or
                 ! if we take special care of Doppler shifts, i.e. catching
                 ! doppler jumps.
                 !
                 if(camera_catch_doppler_line) then
                    !
                    ! Special kind of integration: With doppler-catching.
                    ! This means that we are going to check at each integration
                    ! segment along the ray if the Doppler shift between the
                    ! start and finish of this ray segment is larger than 
                    ! some fraction of the local thermal+microturbulent line
                    ! width. If so, we will do substeps along this ray segment.
                    !
                    nu = camera_frequencies(inu)
                    !
                    ! Call the line module to make the final computation
                    ! from source function and base alpha to j_nu and 
                    ! alpha_nu for the endpoint of the current ray segment.
                    ! This is for the line contributions only. The continuum
                    ! stuff is already done before.
                    !
                    call lines_jnu_anu_from_base(camera_vertex_lines_nractivetot, &
                         camera_local_line_jnubase_curr(:),&
                         camera_local_line_anubase_curr(:),&
                         camera_local_doppler_curr,&
                         camera_local_turb_curr,&
                         camera_local_temp_curr,&
                         camera_frequencies(inu),ln_anu_curr,ln_jnu_curr)
                    !
                    ! Now check if there is any line that Doppler-jumps
                    ! over the current wavelength-of-sight...
                    !
                    ! Do a loop over all active lines and do that check.
                    !
                    ! **** NOTE: This can be done more efficiently! ****
                    !
                    nsteps = 0
                    do ispec=1,lines_nr_species
                       !
                       ! Determine the local line width for this molecule,
                       ! times some safety margin. To be on the safe side
                       ! we take the smallest possible value.
                       !
                       temp  = min(camera_local_temp_curr, &
                                   camera_local_temp_prev )
                       duc   = ( min(camera_local_turb_curr, &
                                 camera_local_turb_prev ) + &
                                 sqrt(2*kk*temp/lines_umass(ispec)) ) / cc
                       margin = duc * lines_widthmargin
                       resol  = duc * camera_catch_doppler_resolution
                       !
                       ! Now loop over all active lines of this molecule
                       !
                       do ilactive=1,active_nrlines(ispec)
                          !
                          ! Which lines is this?
                          !
                          iline = active_lines(ilactive,ispec)
                          !
                          ! What is the rest-frame frequency?
                          !
                          nu0   = lines_nu0(iline,ispec)
                          !
                          ! What is the doppler-shifted frequency, i.e. the
                          ! frequency of line center in the lab frame (prev
                          ! and curr)? Note: If we use double-precision, we
                          ! are fine here.
                          !
                          nu0_prev = nu0 * ( 1.d0 + camera_local_doppler_prev )
                          nu0_curr = nu0 * ( 1.d0 + camera_local_doppler_curr )
                          !
                          ! Does a jump happen (distinguish two scenarios)?
                          !
                          if(nu0_curr.gt.nu0_prev) then
                             if((nu0_curr-nu0_prev.gt.resol*nu).and.  &
                                (nu0_curr.gt.nu*(1.d0-margin)).and.   &
                                (nu0_prev.lt.nu*(1.d0+margin))) then
                                !
                                ! Yes, we have a potentially dangerous jump!
                                ! Compute the required refinement, where 
                                ! for simplicity we simply refine the entire
                                ! ray segment (i.e. a slightly simpler
                                ! procedure than in Pontoppidan et al.)
                                !
                                nstp = (nu0_curr-nu0_prev)/(resol*nu+1.d-99)+1.d0
                                if(nstp.gt.10000) then
                                   write(stdo,*) 'ERROR during Doppler-catching of lines:'
                                   write(stdo,*) '      Local line width is so small that '
                                   write(stdo,*) '      I require more than 10000 refinement'
                                   write(stdo,*) '      steps in a single cell! This is'
                                   write(stdo,*) '      suspicious, so I stop.'
                                   write(stdo,*) '      nu       = ',nu
                                   write(stdo,*) '      nu0_line = ',nu0
                                   write(stdo,*) '      du/c     = ',duc
                                   write(stdo,*) '      nstp     = ',nstp
                                   stop
                                endif
                                if(nstp.gt.nsteps) nsteps=nstp
                             endif
                          else
                             if((nu0_prev-nu0_curr.gt.resol*nu).and.  &
                                (nu0_prev.gt.nu*(1.d0-margin)).and.   &
                                (nu0_curr.lt.nu*(1.d0+margin))) then
                                !
                                ! Yes, we have a potentially dangerous jump!
                                ! Compute the required refinement, where 
                                ! for simplicity we simply refine the entire
                                ! ray segment (i.e. a slightly simpler
                                ! procedure than in Pontoppidan et al.)
                                !
                                nstp = (nu0_prev-nu0_curr)/(resol*nu+1.d-99)+1.d0
                                if(nstp.gt.10000) then
                                   write(stdo,*) 'ERROR during Doppler-catching of lines:'
                                   write(stdo,*) '      Local line width is so small that '
                                   write(stdo,*) '      I require more than 10000 refinement'
                                   write(stdo,*) '      steps in a single cell! This is'
                                   write(stdo,*) '      suspicious, so I stop.'
                                   write(stdo,*) '      nu       = ',nu
                                   write(stdo,*) '      nu0_line = ',nu0
                                   write(stdo,*) '      du/c     = ',duc
                                   write(stdo,*) '      nstp     = ',nstp
                                   stop
                                endif
                                if(nstp.gt.nsteps) nsteps=nstp
                             endif
                          endif
                       enddo
                    enddo
                    !
                    ! Check if we have to do a substepping or
                    ! if we can integrate this ray segment in 
                    ! one step.
                    !
                    if(nsteps.eq.0) then
                       !
                       ! No substeps required (no doppler jump occurring).
                       ! Normal second order integration step of dust+lines
                       ! over the entire ray segment.
                       !
                       ! Create the total alpha_nu and j_nu
                       !
                       ap = anu_prev + ln_anu_prev
                       ac = anu_curr + ln_anu_curr
                       jp = jnu_prev + ln_jnu_prev
                       jc = jnu_curr + ln_jnu_curr
                       sp = jp/(ap+1.d-99)
                       sc = jc/(ac+1.d-99)
                       !
                       if(camera_tracemode.le.-2) then
                          !
                          ! Integration of the optacity and/or finding of the
                          ! tau=1 surface
                          !
                          dtau1     = 0.5d0 * ( ap + ac ) * ray_ds
                          if(camera_tracemode.eq.-2) then
                             !
                             ! Integration of the optacity to obtain the total optical depth
                             ! along the ray
                             !
                             intensity(inu) = intensity(inu) + dtau1
                          else
                             !
                             ! Find the tau=camera_taustop point in order to create 
                             ! a "tau=1" surface (or better: "tau=camera_tausurface").
                             !
                             dummy2 = intensity(inu) + dtau1
                             if(dummy2.ge.camera_taustop(inu)) then
                                !
                                ! We arrived at the taustop point
                                !
                                ! Self-consistency check
                                !
                                if(camera_nrrefine.gt.0) stop 9245
                                !
                                ! If we arrived here because camera_taustop(inu).le.0
                                ! then do a special treatment, otherwise continue as
                                ! normal
                                !
                                if(camera_taustop(inu).le.0.d0) then
                                   camera_xstop(inu) = -1d91
                                   camera_ystop(inu) = -1d91
                                   camera_zstop(inu) = -1d91
                                   camera_dstop(inu) = -1d91
                                else
                                   !
                                   ! Do a linear interpolation between tau and tau+dtau1
                                   !
                                   epstau = (camera_taustop(inu)-intensity(inu))/dtau1
                                   if((epstau.gt.1.d0).or.(epstau.lt.0.d0)) then
                                      write(stdo,*) epstau,dtau1,camera_taustop(inu)
                                      stop 9223
                                   endif
                                   !
                                   ! Find the 3-D position of the point where the linear
                                   ! interpolation put the tau=1 point
                                   !
                                   camera_xstop(inu) = ray_prev_x + epstau * (ray_cart_x - ray_prev_x)
                                   camera_ystop(inu) = ray_prev_y + epstau * (ray_cart_y - ray_prev_y)
                                   camera_zstop(inu) = ray_prev_z + epstau * (ray_cart_z - ray_prev_z)
                                   !
                                   ! Now perform the usual rotation into the image plane, so that
                                   ! the "image" is now this distace to the tau=1 plane.
                                   !
                                   if(.not.camera_localobserver) then
                                      !
                                      ! Observer at infinity: the "distance" is measured toward the 
                                      ! plane going through (0,0,0) perpendicular to the line of sight, 
                                      ! with positive value being toward the observer (more positive = 
                                      ! closer to the observer).
                                      !
                                      ! A rotation in the x,y plane, rotating horizontally around
                                      ! the pointing position.
                                      !
                                      xbk = camera_xstop(inu) - camera_pointing_position(1)
                                      ybk = camera_ystop(inu) - camera_pointing_position(2)
                                      !spx = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
                                      spy = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
                                      !
                                      ! Then a rotation in the y,z plane
                                      !
                                      ybk = spy
                                      zbk = camera_zstop(inu) - camera_pointing_position(3)
                                      !spy = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
                                      spz =-camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
                                      !
                                      ! Take the z-value and put this into camera_dstop(inu)
                                      !
                                      camera_dstop(inu) = spz
                                   else 
                                      !
                                      ! Local observer: the "distance" is the true distance to the 
                                      ! observer (more positive = farther away from the observer).
                                      ! It can be useful for analysis purposes to find the tau=1 surface 
                                      ! to some specific point in the model
                                      !
                                      camera_dstop(inu) =                                            &
                                           sqrt((camera_observer_position(1)-camera_xstop(inu))**2 + &
                                           (camera_observer_position(2)-camera_ystop(inu))**2 +      &
                                           (camera_observer_position(3)-camera_zstop(inu))**2)
                                   endif
                                endif
                                !
                                ! Reset the taustop to infinity
                                !
                                camera_taustop(inu) = 1d91
                                !
                             endif
                             !
                             ! Continue integrating the optical depth
                             !
                             intensity(inu) = dummy2
                          endif
                       else
                          !
                          ! Integration of the radiative transfer equation.
                          ! Second order integration (i.e. integrating
                          ! a first order approximation of the opacity and source
                          ! functions).
                          !
                          dtau1     = 0.5d0 * ( ap + ac ) * ray_ds
                          theomax   = 0.5d0 * ( jp + jc ) * ray_ds
                          if(dtau1.gt.1.d-6) then
                             xp = exp(-dtau1)
                             e0 = 1.d0 - xp
                             e1 = dtau1 - e0
                             bx = e1 / dtau1
                             ax = e0 - bx
                          else
                             ax = 0.5d0 * dtau1
                             bx = 0.5d0 * dtau1
                             xp = 1.d0 - dtau1 
                          endif
                          if(dtau1.gt.1e-9) then
                             qdr = ax * sp + bx * sc
                          else
                             qdr = theomax
                          endif
                          qdr = min(qdr,theomax)
                          intensity(inu) = xp * intensity(inu) + qdr
                       endif
                    else
                       !
                       ! Yes, we have a doppler jump here: A line shifts
                       ! across our current wavelength by more than
                       ! camera_catch_doppler_resolution times the local
                       ! (thermal+turbulent) line width.
                       !
                       ! So we have to do sub-stepping to resolve this
                       ! line doppler shift.
                       !
                       ap = anu_prev + ln_anu_prev
                       jp = jnu_prev + ln_jnu_prev
                       sp = jp/(ap+1.d-99)
                       !
                       ! Memorize the "current" quantities at the end of
                       ! the present ray segment
                       !
                       camera_local_line_jnubase_end(:) = camera_local_line_jnubase_curr(:)
                       camera_local_line_anubase_end(:) = camera_local_line_anubase_curr(:)
                       camera_local_doppler_end         = camera_local_doppler_curr
                       camera_local_turb_end            = camera_local_turb_curr
                       camera_local_temp_end            = camera_local_temp_curr
                       !
                       ! Do loop over substepping
                       !
                       do istep=1,nsteps
                          !
                          ! Fraction of the ray segment that we are now
                          !
                          eps  = (istep*1.d0)/nsteps
                          eps1 = 1.d0-eps
                          !
                          ! Find new "current" quantities, by linear interpolation
                          ! between begin and end of this segment ("prev" and "end")
                          !
                          camera_local_line_jnubase_curr(:) = eps1 * camera_local_line_jnubase_prev(:)  + &
                                                               eps * camera_local_line_jnubase_end(:)
                          camera_local_line_anubase_curr(:) = eps1 * camera_local_line_anubase_prev(:) + &
                                                               eps * camera_local_line_anubase_end(:)
                          camera_local_doppler_curr         = eps1 * camera_local_doppler_prev         + &
                                                               eps * camera_local_doppler_end
                          camera_local_turb_curr            = eps1 * camera_local_turb_prev            + &
                                                               eps * camera_local_turb_end
                          camera_local_temp_curr            = eps1 * camera_local_temp_prev            + &
                                                               eps * camera_local_temp_end
                          !
                          ! Now calculate the alpha_nu and j_nu again
                          !
                          call lines_jnu_anu_from_base(camera_vertex_lines_nractivetot, &
                               camera_local_line_jnubase_curr(:),&
                               camera_local_line_anubase_curr(:),&
                               camera_local_doppler_curr,&
                               camera_local_turb_curr,&
                               camera_local_temp_curr,&
                               camera_frequencies(inu),ln_anu_curr,ln_jnu_curr)
                          ac = eps1*anu_prev + eps*anu_curr + ln_anu_curr
                          jc = eps1*jnu_curr + eps*jnu_curr + ln_jnu_curr
                          sc = jc/(ac+1.d-99)
                          !
                          ! Now the integration step
                          !
                          if(camera_tracemode.le.-2) then
                             !
                             ! Integration of the optacity and/or finding of the
                             ! tau=1 surface
                             !
                             dtau1 = 0.5d0 * ( ap + ac ) * ray_ds / nsteps
                             if(camera_tracemode.eq.-2) then
                                !
                                ! Integration of the optacity to obtain the optical depth
                                ! (useful for debugging etc)
                                !
                                intensity(inu) = intensity(inu) + dtau1
                             else
                                !
                                ! Find the tau=camera_taustop point in order to create 
                                ! a "tau=1" surface (or better: "tau=camera_tausurface").
                                !
                                dummy2 = intensity(inu) + dtau1
                                if(dummy2.ge.camera_taustop(inu)) then
                                   !
                                   ! We arrived at the taustop point
                                   !
                                   ! Self-consistency check
                                   !
                                   if(camera_nrrefine.gt.0) stop 9245
                                   !
                                   ! If we arrived here because camera_taustop(inu).le.0
                                   ! then do a special treatment, otherwise continue as
                                   ! normal
                                   !
                                   if(camera_taustop(inu).le.0.d0) then
                                      camera_xstop(inu) = -1d91
                                      camera_ystop(inu) = -1d91
                                      camera_zstop(inu) = -1d91
                                      camera_dstop(inu) = -1d91
                                   else
                                      !
                                      ! Do a linear interpolation between tau and tau+dtau1
                                      !
                                      epstau = (camera_taustop(inu)-intensity(inu))/dtau1
                                      if((epstau.gt.1.d0).or.(epstau.lt.0.d0)) then
                                         write(stdo,*) epstau,dtau1,camera_taustop(inu)
                                         stop 9223
                                      endif
                                      !
                                      ! Find the 3-D position of the point where the linear
                                      ! interpolation put the tau=1 point. Note that we must
                                      ! here take into account the substepping (doppler catching)
                                      !
                                      eps0 = (istep*1.d0-1.d0)/nsteps
                                      rx0  = ray_prev_x + eps0 * (ray_cart_x - ray_prev_x)
                                      ry0  = ray_prev_y + eps0 * (ray_cart_y - ray_prev_y)
                                      rz0  = ray_prev_z + eps0 * (ray_cart_z - ray_prev_z)
                                      rx1  = ray_prev_x + eps * (ray_cart_x - ray_prev_x)
                                      ry1  = ray_prev_y + eps * (ray_cart_y - ray_prev_y)
                                      rz1  = ray_prev_z + eps * (ray_cart_z - ray_prev_z)
                                      camera_xstop(inu) = rx0 + epstau * (rx1 - rx0)
                                      camera_ystop(inu) = ry0 + epstau * (ry1 - ry0)
                                      camera_zstop(inu) = rz0 + epstau * (rz1 - rz0)
                                      !
                                      ! Now perform the usual rotation into the image plane, so that
                                      ! the "image" is now this distace to the tau=1 plane.
                                      !
                                      if(.not.camera_localobserver) then
                                         !
                                         ! Observer at infinity: the "distance" is measured toward the 
                                         ! plane going through (0,0,0) perpendicular to the line of sight, 
                                         ! with positive value being toward the observer (more positive = 
                                         ! closer to the observer).
                                         !
                                         ! A rotation in the x,y plane, rotating horizontally around
                                         ! the pointing position.
                                         !
                                         xbk = camera_xstop(inu) - camera_pointing_position(1)
                                         ybk = camera_ystop(inu) - camera_pointing_position(2)
                                         !spx = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
                                         spy = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
                                         !
                                         ! Then a rotation in the y,z plane
                                         !
                                         ybk = spy
                                         zbk = camera_zstop(inu) - camera_pointing_position(3)
                                         !spy = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
                                         spz =-camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
                                         !
                                         ! Take the z-value and put this into "intensity(inu)"
                                         !
                                         camera_dstop(inu) = spz
                                      else 
                                         !
                                         ! Local observer: the "distance" is the true distance to the 
                                         ! observer (more positive = farther away from the observer).
                                         ! It can be useful for analysis purposes to find the tau=1 surface 
                                         ! to some specific point in the model
                                         !
                                         camera_dstop(inu) =                                            &
                                              sqrt((camera_observer_position(1)-camera_xstop(inu))**2 + &
                                              (camera_observer_position(2)-camera_ystop(inu))**2 +      &
                                              (camera_observer_position(3)-camera_zstop(inu))**2)
                                      endif
                                      !
                                      ! Reset the taustop to infinity
                                      !
                                      camera_taustop(inu) = 1d91
                                      !
                                   endif
                                endif
                                !
                                ! Continue integrating the optical depth
                                !
                                intensity(inu) = dummy2
                             endif
                          else
                             !
                             ! Integration of the radiative transfer equation.
                             ! Second order integration (i.e. integrating
                             ! a first order approximation of the opacity and source
                             ! functions).
                             !
                             dtau1     = 0.5d0 * ( ap + ac ) * ray_ds / nsteps
                             theomax   = 0.5d0 * ( jp + jc ) * ray_ds / nsteps
                             if(dtau1.gt.1.d-6) then
                                xp = exp(-dtau1)
                                e0 = 1.d0 - xp
                                e1 = dtau1 - e0
                                bx = e1 / dtau1
                                ax = e0 - bx
                             else
                                ax = 0.5d0 * dtau1
                                bx = 0.5d0 * dtau1
                                xp = 1.d0 - dtau1 
                             endif
                             if(dtau1.gt.1e-9) then
                                qdr = ax * sp + bx * sc
                             else
                                qdr = theomax
                             endif
                             qdr = min(qdr,theomax)
                             intensity(inu) = xp * intensity(inu) + qdr
                          endif
                          !
                          ! Now shift curr to prev
                          !
                          ap = ac
                          jp = jc
                          sp = sc
                          !
                          ! Now next substep
                          !
                       enddo
                    endif
                 else
                    !
                    ! Normal kind of integration (no sub-stepping and thus
                    ! no doppler catching). Here the anu and jnu already 
                    ! contain the line contributions.
                    !
                    if(camera_tracemode.le.-2) then
                       !
                       ! Integration of the optacity and/or finding of the
                       ! tau=1 surface
                       !
                       dtau1 = 0.5d0 * ( anu_prev + anu_curr ) * ray_ds
                       if(camera_tracemode.eq.-2) then
                          !
                          ! Integration of the optacity to obtain the optical depth
                          ! (useful for debugging etc)
                          !
                          intensity(inu) = intensity(inu) + dtau1
                       else
                          !
                          ! Find the tau=camera_taustop point in order to create 
                          ! a "tau=1" surface (or better: "tau=camera_tausurface").
                          !
                          dummy2 = intensity(inu) + dtau1
                          if(dummy2.ge.camera_taustop(inu)) then
                             !
                             ! We arrived at the taustop point
                             !
                             ! Self-consistency check
                             !
                             if(camera_nrrefine.gt.0) stop 9245
                             !
                             ! If we arrived here because camera_taustop(inu).le.0
                             ! then do a special treatment, otherwise continue as
                             ! normal
                             !
                             if(camera_taustop(inu).le.0.d0) then
                                camera_xstop(inu) = -1d91
                                camera_ystop(inu) = -1d91
                                camera_zstop(inu) = -1d91
                                camera_dstop(inu) = -1d91
                             else
                                !
                                ! Do a linear interpolation between tau and tau+dtau1
                                !
                                epstau = (camera_taustop(inu)-intensity(inu))/dtau1
                                if((epstau.gt.1.d0).or.(epstau.lt.0.d0)) then
                                   write(stdo,*) epstau,dtau1,camera_taustop(inu)
                                   stop 9223
                                endif
                                !
                                ! Find the 3-D position of the point where the linear
                                ! interpolation put the tau=1 point
                                !
                                camera_xstop(inu) = ray_prev_x + epstau * (ray_cart_x - ray_prev_x)
                                camera_ystop(inu) = ray_prev_y + epstau * (ray_cart_y - ray_prev_y)
                                camera_zstop(inu) = ray_prev_z + epstau * (ray_cart_z - ray_prev_z)
                                !
                                ! Now perform the usual rotation into the image plane, so that
                                ! the "image" is now this distace to the tau=1 plane.
                                !
                                if(.not.camera_localobserver) then
                                   !
                                   ! Observer at infinity: the "distance" is measured toward the 
                                   ! plane going through (0,0,0) perpendicular to the line of sight, 
                                   ! with positive value being toward the observer (more positive = 
                                   ! closer to the observer).
                                   !
                                   ! A rotation in the x,y plane, rotating horizontally around
                                   ! the pointing position.
                                   !
                                   xbk = camera_xstop(inu) - camera_pointing_position(1)
                                   ybk = camera_ystop(inu) - camera_pointing_position(2)
                                   !spx = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
                                   spy = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
                                   !
                                   ! Then a rotation in the y,z plane
                                   !
                                   ybk = spy
                                   zbk = camera_zstop(inu) - camera_pointing_position(3)
                                   !spy = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
                                   spz =-camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
                                   !
                                   ! Take the z-value and put this into "intensity(inu)"
                                   !
                                   camera_dstop(inu) = spz
                                else 
                                   !
                                   ! Local observer: the "distance" is the true distance to the 
                                   ! observer (more positive = farther away from the observer).
                                   ! It can be useful for analysis purposes to find the tau=1 surface 
                                   ! to some specific point in the model
                                   !
                                   camera_dstop(inu) =                                            &
                                        sqrt((camera_observer_position(1)-camera_xstop(inu))**2 + &
                                        (camera_observer_position(2)-camera_ystop(inu))**2 +      &
                                        (camera_observer_position(3)-camera_zstop(inu))**2)
                                endif
                                !
                                ! Reset the taustop to infinity
                                !
                                camera_taustop(inu) = 1d91
                                !
                             endif
                          endif
                          !
                          ! Continue integrating the optical depth
                          !
                          intensity(inu) = dummy2
                       endif
                    else
                       !
                       ! Integration of the radiative transfer equation.
                       ! Second order integration (i.e. integrating
                       ! a first order approximation of the opacity and source
                       ! functions).
                       !
                       dtau1     = 0.5d0 * ( anu_prev + anu_curr ) * ray_ds
                       theomax   = 0.5d0 * ( jnu_prev + jnu_curr ) * ray_ds
                       if(dtau1.gt.1.d-6) then
                          xp = exp(-dtau1)
                          e0 = 1.d0 - xp
                          e1 = dtau1 - e0
                          bx = e1 / dtau1
                          ax = e0 - bx
                       else
                          ax = 0.5d0 * dtau1
                          bx = 0.5d0 * dtau1
                          xp = 1.d0 - dtau1 
                       endif
                       if(dtau1.gt.1e-9) then
                          qdr = ax * snu_prev + bx * snu_curr
                       else
                          qdr = theomax
                       endif
                       qdr = min(qdr,theomax)
                       intensity(inu) = xp * intensity(inu) + qdr
                    endif
                 endif
              endif
              !
           elseif(camera_tracemode.eq.-1) then
              !
              ! Simply integrate the dust density
              !
              ! Make 'intensity' the integrated density (column density)
              !
              do ispec=1,dust_nr_species
                 intensity(1) = intensity(1) + camera_dustdens(ispec) * ray_ds
              enddo
              !
           else
              write(stdo,*) 'ERROR in camera module:'
              write(stdo,*) 'Do not know tracemode = ',camera_tracemode
              stop
           endif
           !
        else
           !
           ! First order integration. Cell-based scheme.
           !
           ! Get the src and alp values
           !
           call camera_get_src_alp(inu0,inu1,nrfreq,src,alp)
           !
           ! Now do the integration of the formal transfer equation
           !
           if(camera_tracemode.eq.1) then
              !
              ! Do formal radiative transfer
              !
              do inu=inu0,inu1
                 !
                 ! Compute the exp(-tau)
                 !
                 xp = alp(inu) * ray_ds
                 if(xp.lt.0.d0) stop 7329
                 if(xp.gt.1d-4) then
                    xp  = exp(-xp)
                    xp1 = 1.d0 - xp
                 else
                    xp1 = xp
                    xp  = exp(-xp)
                 endif
                 !
                 ! Now do the RT, including everything
                 !
                 intensity(inu) = xp * intensity(inu) + xp1 * src(inu)
              enddo
              !
           elseif(camera_tracemode.eq.-1) then
              !
              ! Simply integrate the dust density
              !
              ! Make 'intensity' the integrated density (column density)
              !
              do ispec=1,dust_nr_species
                 intensity(inu0) = intensity(inu0) + camera_dustdens(ispec) * ray_ds
              enddo
              !
           elseif(camera_tracemode.le.-2) then
              !
              ! Integration of the optacity and/or finding of the
              ! tau=1 surface
              !
              if(camera_tracemode.eq.-2) then
                 !
                 ! Integration of the optacity to obtain the total optical depth
                 ! along the ray
                 !
                 do inu=inu0,inu1
                    intensity(inu) = intensity(inu) + alp(inu) * ray_ds
                 enddo
              else
                 !
                 ! Find the tau=camera_taustop point in order to create 
                 ! a "tau=1" surface (or better: "tau=camera_tausurface").
                 !
                 do inu=inu0,inu1
                    !
                    ! Compute the delta tau at this frequency
                    !
                    dtau1 = alp(inu) * ray_ds
                    !
                    ! Now check if we reached the stopping point yet
                    !
                    dummy2 = intensity(inu) + dtau1
                    if(dummy2.ge.camera_taustop(inu)) then
                       !
                       ! We arrived at the taustop point
                       !
                       ! Self-consistency check
                       !
                       if(camera_nrrefine.gt.0) stop 9245
                       !
                       ! If we arrived here because camera_taustop(inu).le.0
                       ! then do a special treatment, otherwise continue as
                       ! normal
                       !
                       if(camera_taustop(inu).le.0.d0) then
                          camera_xstop(inu) = -1d91
                          camera_ystop(inu) = -1d91
                          camera_zstop(inu) = -1d91
                          camera_dstop(inu) = -1d91
                       else
                          !
                          ! Do a linear interpolation between tau and tau+dtau1
                          !
                          epstau = (camera_taustop(inu)-intensity(inu))/dtau1
                          if((epstau.gt.1.d0).or.(epstau.lt.0.d0)) then
                             write(stdo,*) epstau,dtau1,camera_taustop(inu)
                             stop 9223
                          endif
                          !
                          ! Find the 3-D position of the point where the linear
                          ! interpolation put the tau=1 point
                          !
                          camera_xstop(inu) = ray_prev_x + epstau * (ray_cart_x - ray_prev_x)
                          camera_ystop(inu) = ray_prev_y + epstau * (ray_cart_y - ray_prev_y)
                          camera_zstop(inu) = ray_prev_z + epstau * (ray_cart_z - ray_prev_z)
                          !
                          ! Now perform the usual rotation into the image plane, so that
                          ! the "image" is now this distace to the tau=1 plane.
                          !
                          if(.not.camera_localobserver) then
                             !
                             ! Observer at infinity: the "distance" is measured toward the 
                             ! plane going through (0,0,0) perpendicular to the line of sight, 
                             ! with positive value being toward the observer (more positive = 
                             ! closer to the observer).
                             !
                             ! A rotation in the x,y plane, rotating horizontally around
                             ! the pointing position.
                             !
                             xbk = camera_xstop(inu) - camera_pointing_position(1)
                             ybk = camera_ystop(inu) - camera_pointing_position(2)
                             !spx = camera_observer_cos_phi * xbk - camera_observer_sin_phi * ybk
                             spy = camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
                             !
                             ! Then a rotation in the y,z plane
                             !
                             ybk = spy
                             zbk = camera_zstop(inu) - camera_pointing_position(3)
                             !spy = camera_observer_cos_theta * ybk + camera_observer_sin_theta * zbk
                             spz =-camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
                             !
                             ! Take the z-value and put this into camera_dstop(inu)
                             !
                             camera_dstop(inu) = spz
                          else 
                             !
                             ! Local observer: the "distance" is the true distance to the 
                             ! observer (more positive = farther away from the observer).
                             ! It can be useful for analysis purposes to find the tau=1 surface 
                             ! to some specific point in the model
                             !
                             camera_dstop(inu) =                                            &
                                  sqrt((camera_observer_position(1)-camera_xstop(inu))**2 + &
                                  (camera_observer_position(2)-camera_ystop(inu))**2 +      &
                                  (camera_observer_position(3)-camera_zstop(inu))**2)
                          endif
                          !
                          ! Reset the taustop to infinity
                          !
                          camera_taustop(inu) = 1d91
                          !
                       endif
                    endif
                    !
                    ! Continue integrating the optical depth
                    !
                    intensity(inu) = dummy2
                 enddo
              endif
           else
              write(stdo,*) 'ERROR in camera module:'
              write(stdo,*) 'Do not know tracemode = ',camera_tracemode
              stop
           endif
        endif
     elseif(camera_incl_stars.ne.0) then
        !
        ! Special situation: we are inside a star.
        ! So set the intensity to that of the star.
        !
        do inu=inu0,inu1
           intensity(inu) = find_starlight_interpol(camera_frequencies(inu), &
                                                    camera_istar)
        enddo
        !
     endif
     !
     ! Decrease ray_dsend
     !
     ray_dsend = ray_dsend - ray_ds
     !
     ! Now make next cell the current cell
     !
     ray_index = ray_indexnext
     !
     ! Do the same for the pointers.
     ! This is grid type dependent.
     !
     if(igrid_type.lt.100) then
        !
        ! AMR type grid
        !
        if(ray_index.gt.0) then
           !
           ! We are in a cell
           !
           if(amr_tree_present) then
              amrray_cell => amr_index_to_leaf(ray_index)%link
           else
              call amr_regular_get_ixyz(ray_index,amrray_ix_curr,amrray_iy_curr,amrray_iz_curr)
           endif
        else
           nullify(amrray_cell)
           amrray_ix_curr = -1
           amrray_iy_curr = -1
           amrray_iz_curr = -1
        endif
     else
        write(stdo,*) 'ERROR: This grid type not yet implemented'
        stop
     endif
     !
  enddo
  !
  ! Return the smallest cell size we encountered
  !
  if(igrid_type.lt.100) then 
     if(igrid_coord.lt.100) then
        celldxmin = (amr_grid_xi(2,1)-amr_grid_xi(1,1)) / (2**deepestlevel)
     endif
     ! For the other coordinates celldxmin is determined on-the-fly
  else
     write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
     stop
  endif
  !
end subroutine camera_serial_raytrace


!-------------------------------------------------------------------------
!                      GET SRC AND ALP LOCALLY 
!-------------------------------------------------------------------------
subroutine camera_get_src_alp(inu0,inu1,nf,src,alp)
  implicit none
  integer :: ispec,inu,itemplate,inu0,inu1,nf,ilevel,ilevsub
  double precision :: dum,dummy,src(nf),alp(nf)
  double precision :: xav,yav,zav
  double precision :: rlen,cosp,sinp,cost,sint,vx,vy,vz
  double precision :: velgrad
  !
  ! Find the local dust densities
  !
  if(rt_incl_dust) then
     if(ray_index.ge.1) then
        camera_dustdens(:) = dustdens(:,ray_index)
        camera_dusttemp(:) = dusttemp(:,ray_index)
     else
        camera_dustdens(:) = 0.d0
        camera_dusttemp(:) = 1.d0
     endif
  endif
  !
  ! Find the local line-related quantities
  !
  if(rt_incl_lines) then
     if(ray_index.ge.1) then
        if(debug_check_all.eq.1) then
           if(.not.allocated(lines_ray_levpop)) stop 3023
           if(.not.allocated(gas_chemspec_numberdens)) stop 3024
           if((lines_mode.gt.0).and..not.allocated(lines_levelpop)) stop 3025
        endif
        !
        ! Compute the local level populations
        !
        if(lines_mode.gt.0) then
           !
           ! Get the level populations from the big stored array,
           ! meaning that the (should) have been calculated 
           ! beforehand.
           !
           ! NOTE: Only the levels from the subset that is stored
           !       are copied. The populations other levels remain
           !       undefined. 
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then           
                 !
                 !##########################################################
                 ! NOTE: For testing purposes (can later be removed)
                 !       we will put these unused level populations
                 !       to 1D90, to enforce an error if they are
                 !       accidently used.
                 !
                 lines_ray_levpop(:,ispec,1) = 1d90
                 !##########################################################
                 !
                 ! Now copy the levels from the big array to the local
                 ! array
                 !
                 do ilevsub=1,lines_nrlevels_subset(ispec)
                    ilevel = lines_levels_subset(ilevsub,ispec)
                    lines_ray_levpop(ilevel,ispec,1) =            &
                         lines_levelpop(ilevsub,ispec,ray_index)
                 enddo
              endif
           enddo
           !
        elseif(lines_mode.eq.-1) then
           !
           ! On-the-fly LTE population calculation
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 call lines_compute_ltepop_subset(lines_nrlevels(ispec),    &
                      active_nrlevels(ispec),                               &
                      active_levels(:,ispec),                               &
                      ispec,gastemp(ray_index),                             &
                      gas_chemspec_numberdens(ispec,ray_index),             &
                      lines_ray_levpop(:,ispec,1))
              endif
           enddo
        elseif(lines_mode.eq.-2) then
           !
           ! On-the-fly user-defined populations
           ! 
           ! First find the "average" position along this ray segment
           !
           xav = 0.5d0 * ( ray_cart_x + ray_prev_x )
           yav = 0.5d0 * ( ray_cart_y + ray_prev_y )
           zav = 0.5d0 * ( ray_cart_z + ray_prev_z )
           !
           ! On-the-fly user-defined population calculation
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 call userdef_compute_levelpop(ispec,                &
                      lines_nrlevels(ispec),ray_index,               &
                      xav,yav,zav,                                   &
                      gas_chemspec_numberdens(ispec,ray_index),      &
                      lines_ray_levpop(:,ispec,1))
              endif
           enddo
        elseif(lines_mode.eq.-3) then
           !
           ! On-the-fly LVG (Sobolev) populations
           !
           ! First determine the velocity gradient
           !
           call lines_compute_velgradient(ray_index,velgrad)
           !
           ! Now compute, with this velocity gradient, the
           ! LVG populations.
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 if(lines_collisions_npartners(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no collision rates are loaded for'
                    write(stdo,*) '        this molecule, ispec=',ispec
                    stop
                 endif
                 if(lines_collisions_npartners_used(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no number densities of collision partners have been'
                    write(stdo,*) '        specified for this molecule, ispec=',ispec
                    stop
                 endif
                 call lines_compute_lvgpop(ispec,lines_nrlevels(ispec),            &
                      lines_nrlines(ispec),lines_collisions_npartners_used(ispec), &
                      lines_collisions_npartmax,gastemp(ray_index),                &
                      gas_chemspec_numberdens(ispec,ray_index),                    &
                      collpartner_numberdens(:,ray_index),                         &
                      velgrad,lines_ray_levpop(:,ispec,1))
              endif
           enddo
        elseif(lines_mode.eq.-4) then
           !
           ! On-the-fly optically thin non-LTE populations
           !
           ! Now compute the populations
           ! 
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 if(lines_collisions_npartners(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no collision rates are loaded for'
                    write(stdo,*) '        this molecule, ispec=',ispec
                    stop
                 endif
                 if(lines_collisions_npartners_used(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no number densities of collision partners have been'
                    write(stdo,*) '        specified for this molecule, ispec=',ispec
                    stop
                 endif
                 call lines_compute_optthinpop(ispec,lines_nrlevels(ispec),        &
                      lines_nrlines(ispec),lines_collisions_npartners_used(ispec), &
                      lines_collisions_npartmax,gastemp(ray_index),                &
                      gas_chemspec_numberdens(ispec,ray_index),                    &
                      collpartner_numberdens(:,ray_index),                         &
                      lines_ray_levpop(:,ispec,1))
              endif
           enddo
        elseif(lines_mode.eq.-10) then
           !
           ! Entirely general userdef routine for computation of level populations
           ! NOTE: This mode is only available as on-the-fly calculation.
           !
           call userdef_general_compute_levelpop(ray_index, lines_ray_levpop(:,:,1))
        else
           write(stdo,*) 'ERROR: line_mode not known'
           stop
        endif
        !
        ! Copy the temperature and turbulence to local variable
        !
        lines_ray_temp(1)       = gastemp(ray_index)
        if(allocated(lines_microturb)) then
           lines_ray_turb(1)    = lines_microturb(ray_index)
        else
           lines_ray_turb(1)    = 1.d-30
        endif
        !
        ! If linelist molecules are present, then we must also store the
        ! number densities of the molecules
        !
        if(allocated(lines_linelist_eup)) then
           do ispec=1,lines_nr_species
              lines_ray_nrdens(ispec,1) = gas_chemspec_numberdens(ispec,ray_index)
           enddo
        endif
        !
        ! Compute the Lorentz delta parameter in case we are using the Voigt profile
        ! (Added by Thomas Peters 2011)
        !
        if(lines_profile.eq.1) call userdef_compute_lorentz_delta(ray_index)
        !
        ! Compute the doppler shift
        !
        ! ********** CHECK THE SIGN! ***********
        !
        if(igrid_coord.lt.100) then
           !
           ! Cartesian coordinates:
           !
           dummy = gasvelocity(1,ray_index) * ray_cart_dirx +   &
                   gasvelocity(2,ray_index) * ray_cart_diry +   &
                   gasvelocity(3,ray_index) * ray_cart_dirz
        elseif(igrid_coord.lt.200) then
           !
           ! Spherical coordinates: Here component 1 is the v_r component,
           ! component 2 is the v_theta component (in the same unit as
           ! v_r, i.e. cm/s) and component 3 is the v_phi component (also
           ! in the same unit: cm/s).
           !
           xav   = 0.5d0 * ( ray_cart_x + ray_prev_x )
           yav   = 0.5d0 * ( ray_cart_y + ray_prev_y )
           zav   = 0.5d0 * ( ray_cart_z + ray_prev_z )
           dummy = xav**2+yav**2
           rlen  = sqrt(dummy+zav**2)+1d-90
           dummy = sqrt(dummy)+1d-90
           cosp  = xav/dummy
           sinp  = yav/dummy
           cost  = zav/rlen
           sint  = dummy/rlen
           dummy = sint * gasvelocity(1,ray_index) +   &
                   cost * gasvelocity(2,ray_index) 
           vx    = cosp * dummy - sinp * gasvelocity(3,ray_index)
           vy    = sinp * dummy + cosp * gasvelocity(3,ray_index)
           vz    = cost * gasvelocity(1,ray_index) -   &
                   sint * gasvelocity(2,ray_index)
           dummy = vx * ray_cart_dirx +   &
                   vy * ray_cart_diry +   &
                   vz * ray_cart_dirz
        else
           stop 501
        endif
        lines_ray_doppler(1) = dummy / cc
        !
     else
        !
        ! Not in the grid, so put everything to zero
        !
        lines_ray_levpop(:,:,1) = 0.d0
        lines_ray_temp(1)       = 1d-2
        lines_ray_turb(1)       = 1d-30
        lines_ray_doppler(1)    = 1d-30
     endif
  endif
  !
  ! Compute the source terms and extinction coefficients
  !
  ! ...Reset the alp and src
  !
  alp(inu0:inu1) = 0.d0
  src(inu0:inu1) = 0.d0
  !
  ! ...Add the dust emission and extinction
  !
  if(rt_incl_dust) then
     do inu=inu0,inu1
        !
        ! Find the the dust continuum extinction coefficients
        !
        do ispec=1,dust_nr_species
           camera_alpha_a(ispec) = camera_dustdens(ispec) * camera_dustkappa_a(inu,ispec)
           camera_alpha_s(ispec) = camera_dustdens(ispec) * camera_dustkappa_s(inu,ispec)
           alp(inu) = alp(inu) + camera_alpha_a(ispec) + camera_alpha_s(ispec)
        enddo
        ! 
        ! The source function. First the thermal part.
        !
        do ispec=1,dust_nr_species
           src(inu) = src(inu) + camera_alpha_a(ispec) *                           &
                bplanck(camera_dusttemp(ispec),camera_frequencies(inu))
        enddo
        !
        ! Then the scattering part
        !
        if((scattering_mode.ge.1).and.(ray_index.ge.1)) then
           if(camera_mcscat_monochromatic) then
              src(inu) = src(inu) + mcscat_scatsrc(mcscat_current_dir,1,ray_index)
           else
              src(inu) = src(inu) + mcscat_scatsrc(mcscat_current_dir,inu,ray_index)
           endif
        endif
     enddo
  endif
  !
  ! ...Then the continuous stellar source part
  !    Note: the fourpi factor is not needed here, because the stellarsrc_templates
  !          are already in "per steradian" form.
  !
  if(incl_stellarsrc.ne.0) then
     if(ray_index.gt.0) then
        do inu=inu0,inu1
           dum = 0.d0
           do itemplate=1,stellarsrc_nrtemplates
              stellarsrc_cum(itemplate) = dum
              dum = dum + stellarsrc_dens(itemplate,ray_index) *      &
                          camera_stellarsrc_templates(inu,itemplate)
           enddo
           src(inu) = src(inu) + dum 
        enddo
     endif
  endif
  !
  ! ...Then the PAH part 
  !
  !########## STILL TO DO ###########
  ! 
  ! ...Add gas continuum sources, if applicable
  !
  if(rt_incl_gascont) then
     call gascont_addto_jnu_alpnu(ray_index,camera_nrfreq,inu0,inu1,   &
          camera_frequencies(:),src(:),alp(:))
  endif
  !
  ! ...Add the line sources and extinction, unless we use
  !    the doppler catching algorith in which case we must
  !    do the final adding of the line emission and extinction
  !    on-the-fly during the ray-tracing (and thus possibly
  !    during the sub-cell steps along the ray).
  !
  if(rt_incl_lines.and.(.not.camera_catch_doppler_line)) then
     call lines_serial_addto_jnu_alpnu(camera_nrfreq,inu0,inu1,   &
          camera_frequencies(:),src(:),alp(:))
  endif
  !
  ! ...Add the user-defined emissivity and absorptivity
  !
  if(rt_incl_userdef_srcalp) then
     call userdef_srcalp(ray_index,camera_nrfreq,inu0,inu1,     &
                         camera_frequencies(:),src(:),alp(:))
  endif
  !
  ! Now divide by the total opacity to obtain the source function
  !
  src(inu0:inu1) = src(inu0:inu1) / ( alp(inu0:inu1) + 1d-199 )
  !
end subroutine camera_get_src_alp

!-------------------------------------------------------------------------
!             COMPUTE AVERAGE INTENSITY IN A SINGLE PIXEL, 
!                WITH POSSIBLE RECURSIVE SUB-PIXELING
!-------------------------------------------------------------------------
recursive subroutine camera_compute_one_pixel(nrfreq,inu0,inu1,px,py,pdx,pdy,  &
                                              nrrefine,intensity,istar)
  implicit none
  integer :: nrfreq,istar
  double precision :: intensity(nrfreq)
  double precision :: intensdum(nrfreq)
  double precision :: intensdum2(nrfreq)
  double precision :: px,py,pdx,pdy
  double precision :: x1,y1,dx1,dy1
  double precision :: x,y,z,dirx,diry,dirz,distance
  double precision :: celldxmin,dum1,factor,rmin
  integer :: nrrefine,idum,inu0,inu1,inu,ns,istar1,is
  logical :: flag,todo,donerefine
  !
  ! Check
  !
  if(abs(pdx-pdy).gt.1.d-6*(pdx+pdy)) then
     write(stdo,*) 'ERROR in camera module: Non-square pixels not allowed.'
     write(stdo,*) '  pdx = ',pdx,', pdy = ' ,pdy
     stop
  endif
  if(nrfreq.ne.camera_nrfreq) stop 3345
  !
  ! For diagnostics only: write out (if requested) the position of this pixel
  !
  if(camera_diagnostics_subpix) then
     write(10,30) px,py,pdx,pdy
30   format(4(E17.10,1X))
  endif
  !
  ! Find the starting point and direction of this ray
  !
  call camera_set_ray(px,py,x,y,z,dirx,diry,dirz,distance)
  !
  ! Reset intensity
  ! 
  if(incl_extlum.eq.0) then
     intensity(inu0:inu1) = 0.d0
  else
     do inu=inu0,inu1
        intensity(inu) = find_extlumintens_interpol(camera_frequencies(inu))
     enddo
  endif
  !
  ! Reset flag
  !
  donerefine = .false.
  !
  ! Call the ray-tracer 
  !
  call camera_serial_raytrace(nrfreq,inu0,inu1,x,y,z,dirx,diry,dirz,distance,celldxmin,intensity)
  !
  ! Increase the counter
  !
  camera_subpixeling_npixtot = camera_subpixeling_npixtot + 1
  !
  ! Check if we need to refine our pixel
  !
  if(max(pdx,pdy).gt.celldxmin*camera_refine_criterion) then
     !
     ! Are we still allowed to refine?
     !
     if(nrrefine.gt.0) then
        !
        ! OK, we drop the above computed intensity and instead create
        ! four recursive new calls to camera_compute_one_pixel
        !
        idum = nrrefine-1
        dx1  = 0.5d0*pdx
        dy1  = 0.5d0*pdy
        intensity(inu0:inu1) = 0.d0
        !
        x1   = px-0.25d0*pdx
        y1   = py-0.25d0*pdy
        call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
        intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
        !
        x1   = px+0.25d0*pdx
        y1   = py-0.25d0*pdy
        call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
        intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
        !
        x1   = px-0.25d0*pdx
        y1   = py+0.25d0*pdy
        call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
        intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
        !
        x1   = px+0.25d0*pdx
        y1   = py+0.25d0*pdy
        call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
        intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
        !
        intensity(inu0:inu1) = intensity(inu0:inu1) * 0.25d0
        donerefine = .true.
     else
        !
        ! Need to refine, but not allowed. We must warn that our resolving
        ! power is simply not sufficient.
        !
        camera_warn_resolution = .true.
     endif
  else
     !
     ! No refinement was done, so this pixel also counts as a "fine" pixel
     ! So increase that counter (this is just for diagnostics; it's non-essential)
     !
     camera_subpixeling_npixfine = camera_subpixeling_npixfine + 1
  endif
  !
  ! Include stellar spheres
  !
  if(star_sphere.and.(.not.donerefine)) then
     !
     ! If the stars are spheres, we have to do a check if we should
     ! continue to refine
     !
     if((camera_incl_stars.eq.1).and.(istar.ne.0)) then
        if(istar.gt.0) then
           !
           ! There is precisely 1 star that contributes to this pixel
           !
           todo = .true.
           !
           ! Check if the current pixel size is already small enough.
           ! If yes, then don't continue
           !
           if(max(pdx,pdy).le.2.d0*star_r(istar)/camera_starsphere_nrpix) then
              todo = .false.
           endif
           !
           ! Check if this star is entirely inside this pixel, which
           ! would make the calculation easier. But only do this if
           ! fluxcons is not on.
           !
           if(todo.and.(nrrefine.lt.0)) then
              !
              ! Check if sphere entirely inside pixel
              !
              flag = camera_check_starsphere_in_pixel(istar,px,py,pdx,pdy,.true.)
              if(flag) then
                 !
                 ! Yes! The star lies entirely inside this pixel, and 
                 ! flux conservation is not strict, so we can now treat
                 ! this star as a point source.
                 !
                 ! Set the starting position and direction of the ray, as well as
                 ! the location in the image
                 !
                 call camera_set_ray_stars_pntsrc(istar,x,y,z,dirx,diry,dirz,px,py,distance)
                 !
                 ! Set the intensity to the stellar spectrum
                 ! 
                 do inu=inu0,inu1
                    intensdum(inu) =                                          &
                         find_starlight_interpol(camera_frequencies(inu),istar)
                 enddo
                 !
                 ! Do ray tracing
                 !
                 call camera_serial_raytrace(camera_nrfreq,inu0,inu1,        &
                                             x,y,z,dirx,diry,dirz,distance,  &
                                             dum1,intensdum)
                 !
                 ! Compute the ratio starsurface / pixelsurface
                 !
                 if(camera_localobserver) then
                    stop 7390
                    !                factor = pi*(star_r(istar)/distance)**2 / (pdx*pdy)
                 else
                    factor = pi*star_r(istar)**2 / (pdx*pdy)
                 endif
                 !
                 ! Now modify the intensity of the pixel to include the star
                 !
                 intensity(inu0:inu1) =                          &
                      (1.d0-factor) * intensity(inu0:inu1) +     &  
                      factor * intensdum(inu0:inu1)
                 !
                 ! Signal that we are done
                 !
                 todo = .false.
              endif
           endif
           !
           ! If we are not yet done, we need to find which of the 2x2 
           ! sub-pixels have to be refined
           !
           if(todo) then
              !
              ! Reset some stuff
              !
              factor = 1.d0
              intensdum2(inu0:inu1) = intensity(inu0:inu1)
              dx1  = 0.5d0*pdx
              dy1  = 0.5d0*pdy
              intensity(inu0:inu1) = 0.d0
              !
              ! Pixel 1,1
              !
              x1   = px-0.25d0*pdx
              y1   = py-0.25d0*pdy
              flag = camera_check_starsphere_in_pixel(istar,x1,y1,dx1,dy1,.false.)
              if(flag) then
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 factor = factor - 0.25d0
              endif
              !
              ! Pixel 2,1
              !
              x1   = px+0.25d0*pdx
              y1   = py-0.25d0*pdy
              flag = camera_check_starsphere_in_pixel(istar,x1,y1,dx1,dy1,.false.)
              if(flag) then
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 factor = factor - 0.25d0
              endif
              !
              ! Pixel 1,2
              !
              x1   = px-0.25d0*pdx
              y1   = py+0.25d0*pdy
              flag = camera_check_starsphere_in_pixel(istar,x1,y1,dx1,dy1,.false.)
              if(flag) then
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 factor = factor - 0.25d0
              endif
              !
              ! Pixel 2,2
              !
              x1   = px+0.25d0*pdx
              y1   = py+0.25d0*pdy
              flag = camera_check_starsphere_in_pixel(istar,x1,y1,dx1,dy1,.false.)
              if(flag) then
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 factor = factor - 0.25d0
              endif
              !
              intensity(inu0:inu1) = factor * intensdum2(inu0:inu1) +    &
                                     intensity(inu0:inu1) * 0.25d0
              !
           endif
        elseif(istar.le.-1) then
           !
           ! We don't know how many stars. We have to count.
           !
           ns = 0
           istar1 = 0
           rmin = 1d99
           do is=1,nstars
              flag = camera_check_starsphere_in_pixel(is,px,py,pdx,pdy,.false.)
              if(flag) then
                 ns = ns + 1
                 if(star_r(is).lt.rmin) then
                    istar1 = is
                    rmin = star_r(is)
                 endif
              endif
           enddo
!!#####################################
!           write(stdo,*) '%% ',istar1,ns
!!#####################################
           !
           ! Check if we still have to refine even if we have one or multiple
           ! stars
           !
           if(max(pdx,pdy).gt.2.d0*rmin/camera_starsphere_nrpix) then
              !
              ! Yes, we have to refine
              !
              if(ns.eq.1) then
                 !
                 ! If there is one star, then redo this pixel with istar=istar1
                 !
                 idum = nrrefine
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,px,py,pdx,pdy,idum,intensdum,istar1)
                 intensity(inu0:inu1) = intensdum(inu0:inu1)
              elseif(ns.gt.1) then
                 !
                 ! More than one star. Refine.
                 !
                 ! Reset some stuff
                 !
                 dx1  = 0.5d0*pdx
                 dy1  = 0.5d0*pdy
                 intensity(inu0:inu1) = 0.d0
                 !
                 ! Pixel 1,1
                 !
                 x1   = px-0.25d0*pdx
                 y1   = py-0.25d0*pdy
                 istar1 = -1
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar1)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 !
                 ! Pixel 2,1
                 !
                 x1   = px+0.25d0*pdx
                 y1   = py-0.25d0*pdy
                 istar1 = -1
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar1)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 !
                 ! Pixel 1,2
                 !
                 x1   = px-0.25d0*pdx
                 y1   = py+0.25d0*pdy
                 istar1 = -1
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar1)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 !
                 ! Pixel 2,2
                 !
                 x1   = px+0.25d0*pdx
                 y1   = py+0.25d0*pdy
                 istar1 = -1
                 call camera_compute_one_pixel(nrfreq,inu0,inu1,x1,y1,dx1,dy1,idum,intensdum,istar1)
                 intensity(inu0:inu1) = intensity(inu0:inu1) + intensdum(inu0:inu1)
                 !
                 ! Finalize
                 !
                 intensity(inu0:inu1) = intensity(inu0:inu1) * 0.25d0
                 !
              endif
           endif
        endif
     endif
  endif
  !
end subroutine camera_compute_one_pixel


!-------------------------------------------------------------------------
!                   CHECK IF A STAR SPHERE IS INSIDE PIXEL
!-------------------------------------------------------------------------
function camera_check_starsphere_in_pixel(istar,px,py,dx,dy,inside)
  implicit none
  logical :: camera_check_starsphere_in_pixel,inside
  integer :: istar
  double precision :: x,y,z,dirx,diry,dirz,px,py,distance,pxs,pys,dx,dy,rs
  rs = star_r(istar)
  call camera_set_ray_stars_pntsrc(istar,x,y,z,dirx,diry,dirz,pxs,pys,distance)
  if(inside) then
     !
     ! Check if stellar disc is entirely inside this pixel (ignore the
     ! roundness)
     !
     if((pxs.gt.px-0.5d0*dx+rs).and.(pxs.lt.px+0.5d0*dx-rs).and. &
          (pys.gt.py-0.5d0*dy+rs).and.(pys.lt.py+0.5d0*dy-rs)) then
        camera_check_starsphere_in_pixel = .true.
     else
        camera_check_starsphere_in_pixel = .false.
     endif
  else
     !
     ! Check if stellar disc can contribute to this pixel (ignore the
     ! roundness)
     !
     if((pxs.gt.px-0.5d0*dx-rs).and.(pxs.lt.px+0.5d0*dx+rs).and. &
          (pys.gt.py-0.5d0*dy-rs).and.(pys.lt.py+0.5d0*dy+rs)) then
        camera_check_starsphere_in_pixel = .true.
     else
        camera_check_starsphere_in_pixel = .false.
     endif
  endif
  return
end function camera_check_starsphere_in_pixel



!-------------------------------------------------------------------------
!                         MAKE A RECTANGULAR IMAGE
!
! If img=0, then do just one image. If img>0, then use vantage point
!    settings and pointing settings from the multiple-images settings
!    array, entry number img. This allows you to make an entire movie
!    of viewing the object under different vantage points all at once.
!    Though for each of these images (each may have multiple wavelengths,
!    se above), you must call camera_make_rect_image separately, and
!    save the image separately. 
!
!-------------------------------------------------------------------------
subroutine camera_make_rect_image(img,tausurf)
  use amr_module
  implicit none
  double precision :: x,y,z,dxx,dyy,dzz,px,py
  integer :: inu,img,ix,iy,idxx,idxy,istar,ierr,inu0,inu1,ierror,ispec
  double precision :: pdx,pdy,dz,d2,d3,factor,celldxmin
  double precision :: dirx,diry,dirz,distance,xbk,ybk,zbk
  logical :: domc,dotausurf
  logical, optional :: tausurf
  character*80 :: strint
  integer :: iact
  logical :: redo
  !
  ! If "tausurf" is set, then the purpose of this subroutine
  ! changes from being an imager to being a "tau=1 surface finder".
  ! Default is .false.
  !
  dotausurf = .false.
  if(present(tausurf)) then
     dotausurf = tausurf
  endif
  !
  ! If camera_diagnostics_subpix.eq..true. then we will make
  ! a dump of the subpixeling diagnostics, so that the user can see
  ! from the file "subpixeling_diagnostics.out" where the (sub-)pixels
  ! are put.
  !
  if(camera_diagnostics_subpix) then
     open(unit=10,file='subpixeling_diagnostics.out')
  endif
  !
  ! Do a self-consistency test: for spherical coordinates with 
  ! a symmetry (i.e. 2-D axial symmetry or 1-D spherical symmetry)
  ! anisotropic scattering cannot be included. This is because for
  ! e.g. 2-D axial symmetry each cell is in fact a "ring" around the
  ! symmetry axis. The various locations along this ring have different
  ! angle toward the observer (unless the observer looks pole-on). 
  ! Only with isotropic scattering this problem does not show up.
  ! One solution is to convert a 2-D model into a 3-D model before
  ! making the images. 
  !
  if(igrid_type.lt.100) then
     if((igrid_coord.ge.100).and.(amr_zdim.eq.0).and.  &
          (scattering_mode.ge.2)) then
        write(stdo,*) 'ERROR: Anisotropic scattering only allowed in 3-D models,'
        write(stdo,*) '       or in 2-D models that are converted into 3-D before'
        write(stdo,*) '       the images or spectra are made.'
        stop
     endif
  endif
  !
  ! Local observer mode is incompatible with 1-D plane-parallel
  ! or 2-D pencil-parallel modes
  !
  if(camera_localobserver) then
     if(igrid_coord.eq.10) then 
        write(stdo,*) 'ERROR: In 1-D plane-parallel mode, no local observer is allowed.'
        stop
     endif
     if(igrid_coord.eq.20) then
        write(stdo,*) 'ERROR: In 2-D pencil-parallel mode, no local observer is allowed.'
        stop
     endif
  endif
  !
  ! If we have 1-D plane-parallel or 2-D pencil-parallel modes, we
  ! switch off the sub-pixeling. For 1-D this is not necessary anyway.
  ! For 2-D it might still be necessary, but we will leave it to the
  ! user to do it by hand - we'll warn.
  !
  if(igrid_coord.eq.10) then
     camera_nrrefine = -1
  endif
  if(igrid_coord.eq.20) then
     if(camera_nrrefine.gt.0) then
        write(stdo,*) 'WARNING: When using 2-D pencil-parallel coordinates,'
        write(stdo,*) '         RADMC-3D does not provide an automatic'
        write(stdo,*) '         subpixeling method (yet). In order to '
        write(stdo,*) '         ensure flux conservation you must take'
        write(stdo,*) '         care yourself to have enough resolution'
        write(stdo,*) '         to resolve all scales in the y-z plane.'
        write(stdo,*) '         In other words, we now switch to nofluxcons.'
     endif
     camera_nrrefine = -1
  endif
  !
  ! If we have 1-D plane-parallel mode, then we always do just a 1x1
  ! pixel image (it's useless to do more pixels)
  !
  if(igrid_coord.eq.10) then
     camera_image_nx = 1
     camera_image_ny = 1
  endif
  !
  ! If we have 2-D pencil-parallel mode, then we always do just a 1xN
  ! pixel image (it's useless to do more pixels)
  !
  if(igrid_coord.eq.20) then
     camera_image_nx = 1
  endif
  !
  ! Since we have two very different modes of observation (one with the 
  ! observer at infinity, in which the image sizes are specified in cm,
  ! and one with the observer local, in which the image sizes are specified
  ! in radians), we must do some checks for self-consistency, because
  ! the user may easily have forgotten to specify something.
  !
  if(img.eq.0) then
     if(camera_localobserver) then
        if(camera_image_halfsize_x.gt.10.d0) then
           write(stdo,*) 'ERROR in camera module: You have chosen the local observer perspective,'
           write(stdo,*) '      but the image size is still specified in cm instead of radian.'
           stop
        endif
        if((abs(camera_observer_position(1)).ge.1d90).or. &
             (abs(camera_observer_position(2)).ge.1d90).or. &
             (abs(camera_observer_position(3)).ge.1d90)) then
           write(stdo,*) 'ERROR in camera module: You have chosen the local observer perspective,'
           write(stdo,*) '      but you have not yet specified the 3-D position of the observer.'
           stop
        endif
     else
        if((igrid_coord.ne.10).and.(igrid_coord.ne.20)) then
           if(camera_image_halfsize_x.lt.10.d0) then
              write(stdo,*) 'ERROR in camera module: You have chosen the observer at infinity perspective,'
              write(stdo,*) '      but the image size is still specified in radian instead of cm (or AU or pc).'
              stop
           endif
        endif
     endif
  endif
  !
  ! Also let the Monte Carlo module know what mode we have (local observer
  ! or not)
  !
  mcscat_localobserver = camera_localobserver
  if(mcscat_localobserver) then
     mcscat_localobs_pos(1:3) = camera_observer_position(1:3)
  endif
  !
  ! If the dust emission is included, then make sure the dust data,
  ! density and temperature are read. If yes, do not read again.
  !
  if(rt_incl_dust) then
     call read_dustdata(1)
     call read_dust_density(1)
     call read_dust_temperature(1)
  endif
  !
  ! If line emission is included, then make sure the line data are
  ! read. If yes, then do not read it again.
  !
  if(rt_incl_lines) then
     call read_lines_all(1)
  endif
  !
  ! If gas continuum is included, then make sure the gas continuum
  ! data are read. If yes, then do not read it again.
  !
  if(rt_incl_gascont) then
     call gascont_init(1)
  endif
  !
  ! If lines are active, and if the level populations are to be calculated
  ! beforehand and stored in the big array, then compute them now, if not
  ! already done. 
  !
  ! Note: it only stores those levels which are selected in the "subset"
  ! (see lines_module.f90). The idea of the subset is that to save memory
  ! you may not want to store all level populations.  Example: A molecule
  ! may have 30 relevant levels. If you need to store all populations
  ! globally, then you need 30 x 8 bytes x nrcells of memory. For large
  ! grids that could be very much. If, however, you wish to only model one
  ! of the lines, then only 2 level populations have to be stored at each
  ! point (the upper and the lower level belonging to that line). If you
  ! select these 2 levels as your "subset" then only these 2 levels will be
  ! stored in the global lines_levelpop() array, saving a lot of memory. For
  ! the LVG method all levels are needed *locally* to compute the
  ! populations, but once these populations are computed, only the 2
  ! relevant ones are then stored in the *global* array lines_levelpop().
  !
  ! You can select this subset manually (in the lines.inp file) or
  ! automatically (by calling the subroutine
  ! lines_automatic_subset_selection() in the lines_module.f90).  The latter
  ! is done just above here.
  !
  if(rt_incl_lines) then
     if((lines_mode.ge.1).and.(lines_mode.le.9)) then
        !
        ! Level subset selection
        !
        if(lines_autosubset) then
           !
           ! If requested, do an automatic subset selection of the
           ! molecular levels
           ! 
           call lines_automatic_subset_selection(camera_nrfreq,     &
                camera_frequencies,1,camera_nrfreq,lines_maxshift,  &
                redo)
           !
           ! Now force a recomputation of the populations, unless
           ! "redo" is .false., meaning that we have the same
           ! levels as before (and thus the population as computed
           ! before must be still correct). Note that if there
           ! exists no lines_levelpop() array, then redo will also
           ! be .true.
           !
           if(redo) then
              iact=2
           else
              iact=1
           endif
           call lines_compute_and_store_local_populations(iact)
        else
           !
           ! Subset is not automatically selected. So either it has
           ! been selected manually or not at all (in which case the
           ! "subset" is the full set of levels).
           !
           write(stdo,*) 'Will store level populations for all levels or ',   &
                'for a manually (in lines.inp) selected subset of levels.'
           !
           ! Now compute the populations only if they have not yet been
           ! computed. If the lines_popul() array is present, we know that
           ! it contains the correct populations, because the subset
           ! is fixed by the user (or is the complete set). 
           !
           call lines_compute_and_store_local_populations(1)
           !
        endif
     endif
  endif
  !
  ! Do the initializing of the camera
  !
  call camera_init()
  !
  ! If tau surface mode, then allocate the following arrays, too.
  ! These give back the 3-D positions of the points on the tau surface.
  ! The projected position is put into the camera_rect_image() array.
  !
  if(dotausurf) then
     if(allocated(camera_tausurface_z)) deallocate(camera_tausurface_z)
     if(allocated(camera_tausurface_y)) deallocate(camera_tausurface_y)
     if(allocated(camera_tausurface_x)) deallocate(camera_tausurface_x)
     allocate(camera_tausurface_x(1:camera_image_nx,1:camera_image_ny,1:camera_nrfreq))
     allocate(camera_tausurface_y(1:camera_image_nx,1:camera_image_ny,1:camera_nrfreq))
     allocate(camera_tausurface_z(1:camera_image_nx,1:camera_image_ny,1:camera_nrfreq))
  endif
  !
  ! Warnings
  !
  if(rt_incl_lines.and.(lines_mode.lt.0).and.(scattering_mode.ne.0)) then
     write(stdo,*) 'WARNING: Using dust scattering AND line transfer with '
     write(stdo,*) '         on-the-fly level population determination can'
     write(stdo,*) '         make RADMC-3D very slow, because dust scattering'
     write(stdo,*) '         means that the transfer must be done freq-by-freq,'
     write(stdo,*) '         and thus the populations must be recalculated'
     write(stdo,*) '         at each freq, which is slowing down the code.'
     write(stdo,*) '         Various possible solutions:'
     write(stdo,*) '          1. If possible, switch off dust scattering'
     write(stdo,*) '             (for instance for far-IR or mm wavelengths)'
     write(stdo,*) '             by setting scattering_mode_max=0 in radmc3d.inp'
     write(stdo,*) '          2. Use not-on-the-fly populations (lines_mode>0)'
     write(stdo,*) '             [at the moment this mode is still in prep]'
  endif
  !
  ! Check
  !
  if(camera_nrfreq.lt.1) then
     write(stdo,*) 'ERROR in camera module: camera frequency array'
     write(stdo,*) '      not set when calling camera_make_rect_image()'
     stop
  endif
  if((camera_image_nx.le.0).or.(camera_image_ny.le.0)) then
     write(stdo,*) 'ERROR in camera module: must have >=1 values for'
     write(stdo,*) '    camera_image_nx and camera_image_ny'
     stop
  endif
  if((camera_refine_criterion.le.0.d0).and.(camera_nrrefine.gt.0)) then
     write(stdo,*) 'ERROR in camera module: must set the camera_refine_criterion'
     write(stdo,*) '      when making flux-conserving images.'
     stop
  endif
  if(img.gt.0) then
     if(.not.allocated(cameras_pt_pos)) then
        write(stdo,*) 'ERROR in camera module: multiple image arrays not'
        write(stdo,*) '      yet allocated.'
        stop
     endif
     if(img.gt.cameras_nr_images) then
        write(stdo,*) 'ERROR in camera module: asked to make image nr ',img,&
             ' but that is beyond cameras_nr_images.'
        stop
     endif
  endif
  !
  ! If using multiple images, then we copy all the info to the relevant
  ! arrays here.
  !
  if(img.gt.0) then
     !
     ! Get the data from the movie arrays, and perform some basic
     ! checks.
     !
     camera_pointing_position(1:3)   = cameras_pt_pos(1:3,img)    
     camera_image_halfsize_x         = cameras_img_hs_x(img)    
     camera_image_halfsize_y         = cameras_img_hs_y(img)    
     camera_zoomcenter_x             = cameras_zmc_x(img)
     camera_zoomcenter_y             = cameras_zmc_y(img)
     camera_pointing_degr_posang     = cameras_pt_degr_pa(img)  
     if(camera_localobserver) then
        camera_observer_position(1:3)   = cameras_obs_pos(1:3,img)   
        if((camera_image_halfsize_x.gt.10.d0).or. &
           (camera_image_halfsize_y.gt.10.d0)) then
           write(stdo,*) 'ERROR in camera module: You have chosen the local observer perspective,'
           write(stdo,*) '      but the image size is still specified in cm instead of radian.'
           stop
        endif
     else
        camera_observer_degr_theta      = cameras_obs_degr_th(img) 
        camera_observer_degr_phi        = cameras_obs_degr_ph(img) 
        if((camera_image_halfsize_x.lt.10.d0).or. &
           (camera_image_halfsize_y.lt.10.d0)) then
           write(stdo,*) 'ERROR in camera module: You have chosen the observer at infinity perspective,'
           write(stdo,*) '      but the image size is still specified in radian instead of cm (or AU or pc).'
           stop
        endif
     endif
  endif
  !
  ! Check
  !
  if((camera_image_halfsize_x.le.0.d0).or.(camera_image_halfsize_y.le.0.d0)) then
     write(stdo,*) 'ERROR in camera module: zero image scale'
     stop
  endif
  !
  ! For now the flux-conserving ray-tracing is not yet implemented for
  ! local-observer perspective
  !
  if(camera_localobserver) then
     if(camera_nrrefine.gt.0) then
        write(stdo,*) 'ERROR: For the local observer viewing mode the '
        write(stdo,*) '       flux-conserving ray-tracing is not yet available.'
        stop
     endif
  endif
  !
  ! Pre-compute cosines and sines
  !
  if(camera_localobserver) then
     !
     ! For the local perspective view the angles must first be calculated
     !
     camera_pointing_cos_posang  = cos(camera_pointing_degr_posang*pi/180.)
     camera_pointing_sin_posang  = sin(camera_pointing_degr_posang*pi/180.)
     dxx = camera_pointing_position(1) - camera_observer_position(1)
     dyy = camera_pointing_position(2) - camera_observer_position(2)
     dzz = camera_pointing_position(3) - camera_observer_position(3)
     d3 = sqrt( dxx**2 + dyy**2 + dzz**2 )
     d2 = sqrt( dxx**2 + dyy**2 )
     if(d3.eq.0.d0) then
        write(stdo,*) 'ERROR in camera module: observer and pointing positions'
        write(stdo,*) '      are identical. Cannot computing pointing direction.'
        stop
     endif
     camera_observer_sin_theta   = d2/d3
     camera_observer_cos_theta   = sqrt(1.d0-camera_observer_sin_theta**2)
     if(camera_observer_position(3).lt.camera_pointing_position(3)) then
        camera_observer_cos_theta   = -camera_observer_cos_theta
     endif
     if(d2.gt.0.d0) then
        camera_observer_sin_phi     = dxx/d2
        camera_observer_cos_phi     = dyy/d2
     else
        camera_observer_sin_phi     = 0.d0
        camera_observer_cos_phi     = 1.d0
     endif
  else
     !
     ! For the camera-at-infinity view the angles are given and the
     ! cos and sin can be calculated easily
     !
     camera_pointing_cos_posang  = cos(camera_pointing_degr_posang*pi/180.)
     camera_pointing_sin_posang  = sin(camera_pointing_degr_posang*pi/180.)
     camera_observer_cos_theta   = cos(camera_observer_degr_theta*pi/180.)
     camera_observer_sin_theta   = sin(camera_observer_degr_theta*pi/180.)
     camera_observer_cos_phi     = cos(camera_observer_degr_phi*pi/180.)
     camera_observer_sin_phi     = sin(camera_observer_degr_phi*pi/180.)
  endif
  !
  ! For now we do not allow local observer if scattering is switched on
  !
  if((scattering_mode.gt.0).and.camera_localobserver) then
     write(stdo,*) 'ERROR: Monte Carlo scattering not yet implemented with local observer...'
     stop
  endif
  !
  ! For now we do not allow multiple vantage points
  !
  if(mcscat_nrdirs.gt.1) then
     write(stdo,*) 'ERROR: Multiple vantage points is not yet allowed in this version!'
     stop
  endif
  !
  ! Check if we need to (re)do the one-frequency scattering Monte Carlo simulation
  !
  domc = .false.
  if(scattering_mode.ge.1) then
     !
     ! For now anisotropic scattering is not allowed in local observer mode
     !
     if(camera_localobserver) then
        write(stdo,*) 'ERROR: At the moment anisotropic scattering is not'
        write(stdo,*) '       allowed for local observer mode.'
        stop
     endif
     !
     ! Compute the direction vector of the observer. Start with (0,0,1).
     ! Actually this is only necessary for scattering_mode.gt.1, but alas.
     !
     dirx = 0.d0
     diry = 0.d0
     dirz = 1.d0
     ybk  = diry
     zbk  = dirz
     diry = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
     dirz = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
     xbk  = dirx
     ybk  = diry
     dirx = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
     diry =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     if(abs(dirx**2+diry**2+dirz**2-1.d0).gt.1d-6) then
        write(stdo,*) 'ERROR in camera module: direction vector not OK.'
        write(stdo,*) dirx,diry,dirz
        stop
     endif
     !
     ! Some simple tests to see if a new MC calculation is necessary
     !
     if((.not.allocated(mcscat_dirs)).and.(scattering_mode.gt.1)) domc=.true.
     if(allocated(mcscat_dirs).and.(scattering_mode.eq.1)) then
        write(stdo,*) 'ERROR: if isotropic scattering is set, then do not set the mcscat_dirs(:,:) array...'
        stop
     endif
     if(mcscat_nrdirs.ne.1) domc=.true.
     if(.not.camera_scatsrc_allfreq) domc=.true.
     !
     ! Some more subtle tests to see if a new MC calculation is necessary
     !
     if(camera_scatsrc_allfreq) then
        if(mc_nrfreq.ne.camera_nrfreq) then
           domc=.true.
        else
           do inu=1,camera_nrfreq
              if(abs((camera_frequencies(inu)-mc_frequencies(inu))/   &
                     (camera_frequencies(inu)+mc_frequencies(inu))).gt.1d-10) domc=.true.
           enddo
        endif
        if(allocated(mcscat_dirs)) then
           mcscat_current_dir = 1
           if((abs(mcscat_dirs(1,mcscat_current_dir)-dirx).gt.1d-6).or.   &
              (abs(mcscat_dirs(2,mcscat_current_dir)-diry).gt.1d-6).or.   &
              (abs(mcscat_dirs(3,mcscat_current_dir)-dirz).gt.1d-6)) then
              domc=.true.
           endif
        endif
     endif
  endif
  !
  ! Prepare the direction for the scattering source function
  ! For now allow only one single direction (i.e. one single vantage point)
  !
  if(domc.and.(scattering_mode.gt.1)) then
     !
     ! Make the scattering direction the same as the direction of
     ! viewing 
     !
     if(allocated(mcscat_dirs)) deallocate(mcscat_dirs)
     allocate(mcscat_dirs(1:3,1:1),STAT=ierr)
     if(ierr.gt.0) then
        write(stdo,*) 'ERROR: Could not allocate mcscat_dirs()'
        stop
     endif
     mcscat_nrdirs = 1
     mcscat_current_dir = 1
     mcscat_dirs(1,1:1) = dirx
     mcscat_dirs(2,1:1) = diry
     mcscat_dirs(3,1:1) = dirz
     !
     ! For now anisotropic scattering is not available for 2-D axisymmetric
     ! models or 1-D spherically symmetric models, because those would require 
     ! many scattering directions in the mcscat_dirs. That should be easiliy
     ! possible to build in (that is one of the main reasons why I introduced
     ! the multi-angle mcscat_dirs in the first place), but for now it is not.
     !
     ! WILL COME SOON, I HOPE!
     !
     if((igrid_coord.ge.100).and.(igrid_coord.le.199)) then
        if(igrid_mirror.ne.0) then
           write(stdo,*) 'ERROR: Mirror symmetry not compatible with anisotropic scattering.'
           stop
        endif
        if(igrid_type.lt.100) then
           if((amr_xdim.ne.1).or.(amr_ydim.ne.1).or.(amr_zdim.ne.1)) then
              write(stdo,*) 'ERROR: For the moment anisotropic scattering in spherical coordinates'
              write(stdo,*) '       is only allowed in 3-D models.'
              stop
           endif
        else
           write(stdo,*) 'INTERNAL ERROR: Grid type not known'
           stop
        endif
     endif
  endif
  !
  ! Compute pixel size.
  !
  ! NOTE: For the observer-at-infinity mode this is in centimeters.
  !       For the local-observer mode this is in radian.
  !
  pdx = 2*camera_image_halfsize_x / (1.d0*camera_image_nx)
  pdy = 2*camera_image_halfsize_y / (1.d0*camera_image_ny)
  !
  ! If the tausurface mode is on, then allocate some arrays
  !
  if(dotausurf) then
     if(allocated(camera_dstop)) deallocate(camera_dstop)
     if(allocated(camera_zstop)) deallocate(camera_zstop)
     if(allocated(camera_ystop)) deallocate(camera_ystop)
     if(allocated(camera_xstop)) deallocate(camera_xstop)
     if(allocated(camera_taustop)) deallocate(camera_taustop)
     allocate(camera_taustop(camera_nrfreq))
     allocate(camera_xstop(camera_nrfreq))
     allocate(camera_ystop(camera_nrfreq))
     allocate(camera_zstop(camera_nrfreq))
     allocate(camera_dstop(camera_nrfreq))
  endif
  !
  ! Reset flag
  !
  camera_warn_resolution = .false.
  !
  ! Message
  !
  write(stdo,*) 'Rendering image(s)...'
  !
  ! Now make the image. We can do so in two different ways. The first is to
  ! do all frequencies (wavelengths) at once. If scattering is included this
  ! requires the precomputing of the scattering source function at all
  ! wavelengths.  This can have a too large memory requirement.  The second
  ! is to do it frequency-by-frequency, and doing a MC calculation for each
  ! of them separately. That saves a lot of computer memory, so it is the
  ! preferred method for large simulations when making spectra. But it may
  ! be slower. The way to choose one or the other is with the switch called
  ! camera_scatsrc_allfreq.
  !
  if(((scattering_mode.eq.0).or.camera_scatsrc_allfreq).and.   &
      (.not.camera_secondorder)) then
     !
     ! Multi-wavelength Method 1:
     !
     ! Do all frequencies at once for each pixel. This is the preferred
     ! method for cases without scattering. With scattering this requires a
     ! lot of memory to store the scattering source function for all
     ! frequencies, so it might not be the best for that.
     !
     ! Note that at present this multi-frequency mode is not available
     ! for second order integration of the transfer equation (i.e. with
     ! the corner points). 
     !
     inu0 = 1
     inu1 = camera_nrfreq
     !
     ! Tell the ray tracer that if scattering is done, then it is done
     ! frequency dispersed
     !
     camera_mcscat_monochromatic = .false.
     !
     ! If necessary, then do the scattering source functions at all
     ! frequencies beforehand.  WARNING: This can be a very large array!
     !
     if(domc) then
        if(allocated(mc_frequencies)) deallocate(mc_frequencies)
        mc_nrfreq=camera_nrfreq
        allocate(mc_frequencies(1:mc_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate mc_frequencies(:) array'
           stop
        endif
        mc_frequencies(:) = camera_frequencies(:)
        !
        ! Message
        !
        write(stdo,*) 'Doing scattering Monte Carlo simulation...'
        call flush(stdo)
        !
        ! Do Monte Carlo simulation
        !
        call do_monte_carlo_scattering(rt_mcparams,ierror,do_resetseed,&
                                       scatsrc=.true.)
        write(stdo,*) 'Average number of scattering events per photon package = ', &
                      ieventcounttot/(1.d0*rt_mcparams%nphot_scat)
     endif
     !
     ! Pre-compute which lines and which levels for line transfer may
     ! contribute to these wavelengths. Note that this only has to be
     ! pre-computed if the serial ray tracing is used, but we do it 
     ! nevertheless, as it will not hurt.
     !
     if(rt_incl_lines) then
        ! ------------------------------------------------------------------
        ! Attila Juhasz
        ! lines_find_active_lines_leves -> if all energy levels are known - leiden
        ! format for line data
        ! lines_find_active_lines_linelist -> for linelist mode
        ! ------------------------------------------------------------------
        !   call lines_find_active_lines_levels(camera_nrfreq,             &
        !                   camera_frequencies,inu0,inu1,lines_maxshift)
        !
        if(lines_maxnrlevels.gt.0) then
           call lines_find_active_lines_levels(camera_nrfreq,             &
                camera_frequencies,inu0,inu1,lines_maxshift)
        else
           call lines_find_active_lines_linelist(camera_nrfreq,             &
                camera_frequencies,inu0,inu1,lines_maxshift)
        endif
        ! ------------------------------------------------------------------
     endif
     !
     ! Message
     !
     if(inu0.eq.inu1) then
        write(stdo,*) 'Ray-tracing image for lambda = ', &
                1d4*cc/camera_frequencies(inu0),' micron...'
     else
        call integer_to_string(abs(inu1-inu0)+1,strint)
        write(stdo,*) 'Ray-tracing images: all '//trim(strint)//' wavelength at once...'
     endif
     !
     ! Now make the image at all wavelengths simultaneously
     !
     call camera_make_rect_image_sub(inu0,inu1)
     !
  else
     !
     ! Multi-wavelength Method 2:
     !
     ! If scattering is included, we need the scattering source function. Since this 
     ! array is easily too big to be stored for all wavelengths. In this method we
     ! compute the scattering source function for each frequency separately, and then
     ! make the corresponding image, and then go to the next frequency.
     !
     ! Tell the ray tracer that if scattering is done, then it is done
     ! one frequency at a time, i.e. the scattering source function is
     ! just one frequency bin.
     !
     camera_mcscat_monochromatic = .true.
     !
     ! Since we recompute the scattering source function for each wavelength, the domc
     ! is by default true. Note that if you make an image at a single wavelength you 
     ! may want to be able to make use of a previously computed source function, if for
     ! instance you make your next image at the same vantage point, the same wavelength
     ! but at a different zoom factor or so. In that case, please set the flag
     ! camera_scatsrc_allfreq to .true.. 
     !
     if(domc) then
        if(allocated(mc_frequencies)) deallocate(mc_frequencies)
        mc_nrfreq=1
        allocate(mc_frequencies(1:1),STAT=ierr)
     endif
     !
     ! Do a loop over all camera frequencies
     !
     do inu0=1,camera_nrfreq
        !
        ! Set this by default
        !
        inu1=inu0
        !
        ! If we must do Monte Carlo, then do this
        !
        if(domc) then
           !
           ! Set the wavelength for the Monte Carlo scattering simulation
           !
           mc_frequencies(1) = camera_frequencies(inu0)
           !
           ! Message
           !
           write(stdo,*) 'Doing scattering Monte Carlo simulation for lambda = ', &
                1d4*cc/mc_frequencies(1),' micron...'
           call flush(stdo)
           !
           ! Call the single wavelength Monte Carlo module
           !
           call do_monte_carlo_scattering(rt_mcparams,ierror,do_resetseed,&
                                          scatsrc=.true.)
           write(stdo,*) 'Average number of scattering events per photon package = ', &
                      ieventcounttot/(1.d0*rt_mcparams%nphot_scat)
        endif
        !
        ! Pre-compute which lines and which levels for line transfer may
        ! contribute to these wavelengths. Note that this only has to be
        ! pre-computed if the serial ray tracing is used, but we do it 
        ! nevertheless, as it will not hurt.
        !
        if(rt_incl_lines) then
           !
           ! Check which lines to include
           !
           ! ------------------------------------------------------------------
           ! Attila Juhasz
           ! lines_find_active_lines_leves -> if all energy levels are known - leiden
           ! format for line data
           ! lines_find_active_lines_linelist -> for linelist mode
           ! ------------------------------------------------------------------
           !   call lines_find_active_lines_levels(camera_nrfreq,             &
           !                   camera_frequencies,inu0,inu1,lines_maxshift)
           !
            if(lines_maxnrlevels.gt.0) then
                call lines_find_active_lines_levels(camera_nrfreq,             &
                     camera_frequencies,inu0,inu1,lines_maxshift)
            else
                call lines_find_active_lines_linelist(camera_nrfreq,             &
                     camera_frequencies,inu0,inu1,lines_maxshift)
            endif

           ! ------------------------------------------------------------------
           !
           ! If camera_catch_doppler_line.eq..true., then allocate the
           ! corner-based line quantities
           !
           if(camera_catch_doppler_line) then
              if(allocated(camera_vertex_line_jnubase)) then
                 deallocate(camera_local_line_anubase_curr)
                 deallocate(camera_local_line_jnubase_curr)
                 deallocate(camera_local_line_anubase_prev)
                 deallocate(camera_local_line_jnubase_prev)
                 deallocate(camera_local_line_anubase_end)
                 deallocate(camera_local_line_jnubase_end)
                 deallocate(camera_cell_line_anubase)
                 deallocate(camera_cell_line_jnubase)
                 deallocate(camera_vertex_line_anubase)
                 deallocate(camera_vertex_line_jnubase)
              endif
              camera_vertex_lines_nractivetot = 0
              do ispec=1,lines_nr_species
                 camera_vertex_lines_nractivetot = camera_vertex_lines_nractivetot + &
                      active_nrlines(ispec)
              enddo
              allocate(camera_vertex_line_jnubase(1:camera_vertex_lines_nractivetot,1:amr_nr_vertices_max))
              allocate(camera_vertex_line_anubase(1:camera_vertex_lines_nractivetot,1:amr_nr_vertices_max))
              allocate(camera_cell_line_jnubase(1:camera_vertex_lines_nractivetot,1:amr_nr_vertices_max))
              allocate(camera_cell_line_anubase(1:camera_vertex_lines_nractivetot,1:amr_nr_vertices_max))
              allocate(camera_local_line_jnubase_curr(1:camera_vertex_lines_nractivetot))
              allocate(camera_local_line_anubase_curr(1:camera_vertex_lines_nractivetot))
              allocate(camera_local_line_jnubase_prev(1:camera_vertex_lines_nractivetot))
              allocate(camera_local_line_anubase_prev(1:camera_vertex_lines_nractivetot))
              allocate(camera_local_line_jnubase_end(1:camera_vertex_lines_nractivetot))
              allocate(camera_local_line_anubase_end(1:camera_vertex_lines_nractivetot))
           endif
        endif
        !
        ! Message
        !
        write(stdo,*) 'Ray-tracing image for lambda = ', &
                1d4*cc/camera_frequencies(inu0),' micron...'
        !
        ! If we do second order integration of the transfer equation,
        ! we must compute the emissivities at the corner points of the
        ! cells (the vertex grid). NOTE: The direction vector will be
        ! set automatically by the called subroutine.
        !
        if(camera_secondorder) then
           call camera_compute_snualphanu_at_vertices(inu0)
        endif
        !
        ! Now make the image for this wavelength only
        !
        call camera_make_rect_image_sub(inu0,inu1)
        !
     enddo
     !
  endif
  !
  ! Check if refinement was sufficient
  !
  if(camera_warn_resolution) then
     write(stdo,*) 'WARNING: The currently produced image is not '
     write(stdo,*) '         guaranteed to have the correct flux '
     write(stdo,*) '         because it does not use the flux'
     write(stdo,*) '         conserving recursive pixeling.'
     write(stdo,*) '    Tip: Use fluxcons argument in command line.'
  endif
  !
  ! Close (if necessary) the "subpixeling_diagnostics.out" file
  ! and switch this diagnostics off, so that it won't do this each
  ! image in case you render a spectrum or SED.
  !
  if(camera_diagnostics_subpix) then
     close(10)
     camera_diagnostics_subpix = .false.
  endif
  !
  ! If the tausurface mode is on, then deallocate some arrays
  !
  if(dotausurf) then
     if(allocated(camera_dstop)) deallocate(camera_dstop)
     if(allocated(camera_zstop)) deallocate(camera_zstop)
     if(allocated(camera_ystop)) deallocate(camera_ystop)
     if(allocated(camera_xstop)) deallocate(camera_xstop)
     if(allocated(camera_taustop)) deallocate(camera_taustop)
  endif
  !
  !--------------------------------------------------------------
  !        A sub-subroutine for making the image
  !--------------------------------------------------------------
  !
  contains
  subroutine camera_make_rect_image_sub(inu00,inu11)
    implicit none
    integer :: inu00,inu11
    integer :: inuu
    integer :: backup_nrrefine,backup_tracemode
    logical :: warn_tausurf_problem
    !
    ! Reset some non-essential counters
    !
    camera_subpixeling_npixfine = 0
    camera_subpixeling_npixtot  = 0
    !
    ! Here we decide whether to make a "normal" image or
    ! whether we find the tau=1 surface 
    !
    if(.not.dotausurf) then
       !
       ! Make a "normal" image, i.e. compute the intensity of all pixels
       !
       ! *** NEAR FUTURE: PUT OPENMP DIRECTIVES HERE (START) ***
       !
       do iy=1,camera_image_ny
          do ix=1,camera_image_nx
             !
             ! Set the ray variables
             !
             px = camera_zoomcenter_x + ((ix-1)-(0.5d0*(camera_image_nx-1)))*pdx
             py = camera_zoomcenter_y + ((iy-1)-(0.5d0*(camera_image_ny-1)))*pdy
             !
             ! Now compute the intensity of this pixel
             !
             ! Note that if nrrefine>0 this routine will call itself
             ! recursively until a desired spatial resolution is acquired
             ! so as to guarantee that all flux is picked up, but this
             ! recursion is limited to nrrefine depth levels.
             !
             call camera_compute_one_pixel(camera_nrfreq,inu00,inu11,px,py,pdx,pdy,    &
                                           camera_nrrefine,camera_intensity,-1)
             !
             ! Put the result into the image array
             !
             camera_rect_image(ix,iy,inu00:inu11) = camera_intensity(inu00:inu11)
             !
          enddo
       enddo
       !
       ! *** NEAR FUTURE: PUT OPENMP DIRECTIVES HERE (FINISH) ***
       !
    else
       !
       ! Find the tau=1 surface (or any tau=tausurf surface)
       !
       ! Set the nrrefine to 0, and set the camera_tracemode to -3
       !
       backup_nrrefine  = camera_nrrefine
       backup_tracemode = camera_tracemode
       camera_nrrefine  = 0
       camera_tracemode = -3
       !
       warn_tausurf_problem = .false.
       !
       do iy=1,camera_image_ny
          do ix=1,camera_image_nx
             !
             ! Set the ray variables
             !
             px = camera_zoomcenter_x + ((ix-1)-(0.5d0*(camera_image_nx-1)))*pdx
             py = camera_zoomcenter_y + ((iy-1)-(0.5d0*(camera_image_ny-1)))*pdy
             !
             ! Reset the camera_xyzsstop
             !
             camera_taustop(inu00:inu11) = 1d91
             camera_xstop(inu00:inu11)   = -1d91
             camera_ystop(inu00:inu11)   = -1d91
             camera_zstop(inu00:inu11)   = -1d91
             camera_dstop(inu00:inu11)   = -1d91
             !
             ! Now do a first ray trace, to find the total optical depth
             !
             call camera_compute_one_pixel(camera_nrfreq,inu00,inu11,px,py,pdx,pdy,    &
                                           camera_nrrefine,camera_intensity,-1)
             !
             ! Check if anywhere the optical depth goes beyond 1d14*camera_tausurface
             !
             do inuu=inu00,inu11
                if(camera_intensity(inuu).gt.1d14*camera_tausurface) then
                   warn_tausurf_problem = .true.
                endif
             enddo
             !
             ! Compute the taustop
             !
             camera_taustop(inu00:inu11) = camera_intensity(inu00:inu11) - camera_tausurface
             !
             ! Reset xyzdstop (should not be necessary, but just for safety)
             !
             camera_xstop(inu00:inu11)   = -1d91
             camera_ystop(inu00:inu11)   = -1d91
             camera_zstop(inu00:inu11)   = -1d91
             camera_dstop(inu00:inu11)   = -1d91
             !
             ! Now do the second ray trace, to find the tau=tausurface surface
             !
             call camera_compute_one_pixel(camera_nrfreq,inu00,inu11,px,py,pdx,pdy,    &
                                           camera_nrrefine,camera_intensity,-1)
             !
             ! Put the resulting dstop into the image array
             !
             camera_rect_image(ix,iy,inu00:inu11) = camera_dstop(inu00:inu11)
             !
             ! If the big arrays for the xyz stop are available, then also store
             ! these
             !
             if(allocated(camera_tausurface_x)) then
                camera_tausurface_x(ix,iy,inu00:inu11) = camera_xstop(inu00:inu11)
                camera_tausurface_y(ix,iy,inu00:inu11) = camera_ystop(inu00:inu11)
                camera_tausurface_z(ix,iy,inu00:inu11) = camera_zstop(inu00:inu11)
             endif
             !
          enddo
       enddo
       !
       ! Print a warning if the tau exceeded the limit somewhere
       !
       if(warn_tausurf_problem) then
          write(stdo,*) 'WARNING: The optical depth exceeded 1d14*tausurface, so the tau surface determination may go wrong.'
       endif
       !
       ! Reset the original values of the tracemode and nrrefine
       !
       camera_nrrefine  = backup_nrrefine
       camera_tracemode = backup_tracemode
    endif
    !
    ! Add the discrete star sources. 
    !
    ! NOTE: Only for the mode in which stars are treated as point-sources.
    !
    if((camera_incl_stars.ne.0).and.(.not.star_sphere).and.(.not.dotausurf)) then
       !
       ! Do a check
       !
       if(.not.allocated(star_spec).and.(nstars.gt.0)) then
          write(stdo,*) 'WARNING in camera module: Stars not allocated.'
          stop
       endif
       !
       ! Now add all stars
       !
       do istar=1,nstars
          !
          ! Set the starting position and direction of the ray, as well as
          ! the location in the image
          !
          call camera_set_ray_stars_pntsrc(istar,x,y,z,dirx,diry,dirz,px,py,distance)
          !
          ! Check if this star is located in the image
          !
          if((px.ge.camera_zoomcenter_x-camera_image_halfsize_x).and.  &
             (px.le.camera_zoomcenter_x+camera_image_halfsize_x).and.  &
             (py.ge.camera_zoomcenter_y-camera_image_halfsize_y).and.  &
             (py.le.camera_zoomcenter_y+camera_image_halfsize_y)) then
             !
             ! Find the indices of the pixel in which the star is
             !
             idxx = floor((px-(camera_zoomcenter_x-camera_image_halfsize_x))/pdx)+1
             idxy = floor((py-(camera_zoomcenter_y-camera_image_halfsize_y))/pdy)+1
             if((idxx.lt.1).or.(idxx.gt.camera_image_nx).or.&
                (idxy.lt.1).or.(idxy.gt.camera_image_ny)) stop 8309
             !
             ! Set the intensity to the stellar spectrum
             ! 
             do inu=inu00,inu11
                camera_intensity(inu) =                                  &
                     find_starlight_interpol(camera_frequencies(inu),istar)
             enddo
             !
             ! Do ray tracing
             !
             call camera_serial_raytrace(camera_nrfreq,inu00,inu11,x,y,z,dirx,diry,dirz,distance,  &
                                  celldxmin,camera_intensity)
             !
             ! Compute the ratio starsurface / pixelsurface
             !
             if(camera_localobserver) then
                factor = pi*(star_r(istar)/distance)**2 / (pdx*pdy)
             else
                factor = pi*star_r(istar)**2 / (pdx*pdy)
             endif
             !
             ! Check if the point source assumption is violated. 
             !
!             if(factor.gt.1.d0) then
!                write(stdo,*) 'ERROR in camera module: The image resolution is so'
!                write(stdo,*) '      high that the stellar surface is no longer'
!                write(stdo,*) '      small enough to be considered a point source.'
!                write(stdo,*) '  NOTE: In the (hopefully near) future a mode will'
!                write(stdo,*) '        be implemented to include non-point stars.'
!                stop
!             endif
             !
             ! Now modify the intensity of the pixel to include the star
             !
             camera_rect_image(idxx,idxy,inu00:inu11) =                           &
                   (1.d0-factor) * camera_rect_image(idxx,idxy,inu00:inu11) +     &  
                   factor * camera_intensity(inu00:inu11)
          endif
       enddo
    endif
  end subroutine camera_make_rect_image_sub
  !
end subroutine camera_make_rect_image


!-------------------------------------------------------------------------
!                              WRITE THE IMAGE
!-------------------------------------------------------------------------
subroutine camera_write_image(img,ifill,noclip)
  implicit none
  character*80 :: filename,base,ext
  integer :: ix,iy,img,iinu
  integer :: ifill
  logical, optional :: noclip
  logical :: donoclip
  !
  ! Interpret the noclip
  !
  donoclip = .false.
  if(present(noclip)) then
     donoclip = noclip
  endif
  !
  ! Open file
  !
  if(stdo.eq.6) then
     !
     ! Current radmc run is not a child process 
     ! So write output to a file  (otherwise to standard out)
     !
     if(img.le.0) then
        if(writeimage_unformatted) then
           open(unit=fflo,file='image.out',form='unformatted')
        else
           open(unit=fflo,file='image.out')
        endif
     else
        base='image_'
        ext ='.out'
        if(ifill.le.0) then
           call make_indexed_filename(base,img,ext,filename)
        else
           call make_indexed_filename_fill(base,img,ext,filename,ifill)
        endif
        if(writeimage_unformatted) then
           open(unit=fflo,file=filename,form='unformatted')
        else
           open(unit=fflo,file=filename)
        endif
     endif
  else
     !
     ! Current radmc run is a child process, so the image must
     ! be written to standard out. Check if fflo is indeed 6.
     !
     if(fflo.ne.6) then
        write(stdo,*) 'ERROR: fflo.ne.6 while stdo.ne.6'
        stop
     endif
  endif
  if(writeimage_unformatted) then
     !
     ! Unformatted output (faster, more compact)
     !
     if(camera_localobserver) then
        write(fflo) 2           ! Format number: size units in radian (angular)
     else
        write(fflo) 1           ! Format number: size units in cm (spatial size)
     endif
     write(fflo) camera_image_nx,camera_image_ny
     write(fflo) camera_nrfreq
     write(fflo) 2.d0*camera_image_halfsize_x/(1.d0*camera_image_nx), &
          2.d0*camera_image_halfsize_y/(1.d0*camera_image_ny)
     write(fflo) (1d4*cc/camera_frequencies(iinu),iinu=1,camera_nrfreq)
     do iinu=1,camera_nrfreq
        write(fflo) ((camera_rect_image(ix,iy,iinu),  &
                     ix=1,camera_image_nx),iy=1,camera_image_ny)
     enddo
  else
     !
     ! Standard: write it formatted way
     !
     if(camera_localobserver) then
        write(fflo,*) 2           ! Format number: size units in radian (angular)
     else
        write(fflo,*) 1           ! Format number: size units in cm (spatial size)
     endif
     write(fflo,*) camera_image_nx,camera_image_ny
     write(fflo,*) camera_nrfreq
     write(fflo,*) 2.d0*camera_image_halfsize_x/(1.d0*camera_image_nx), &
          2.d0*camera_image_halfsize_y/(1.d0*camera_image_ny)
     write(fflo,320) (1d4*cc/camera_frequencies(iinu),iinu=1,camera_nrfreq)
320  format(E21.14)
     write(fflo,*) ' '
     do iinu=1,camera_nrfreq
        do iy=1,camera_image_ny
           do ix=1,camera_image_nx
              if((camera_rect_image(ix,iy,iinu).gt.1d-97).or.donoclip) then
                 write(fflo,333) camera_rect_image(ix,iy,iinu)
!333              format(E13.6)
333              format(E21.14)
              else
                 write(fflo,*) 0.d0
              endif
           enddo
        enddo
        write(fflo,*) ' '
     enddo
  endif
  if(stdo.eq.6) then
     !
     ! Output went to file, so close that file
     !
     close(1)
  else
     !
     ! Output went to standard output, so flush this
     !
     call flush(fflo)
  endif
  !
end subroutine camera_write_image


!-------------------------------------------------------------------------
!                  MAKE A SPECTRUM USING THE IMAGES
!
! This routine calls the make image routine at all frequencies and
! integrates these to make the spectrum. It can do it the simple
! way (just making rectangular images and integrating over them), or
! with aperture information (approximated as a circular mask with
! a wavelength-dependent size given in the aperture_info.inp file).
! Note, however, that the use of the aperture may make the routine
! slower in some cases.
!
! Be sure to have set the following variables to the appropriate values:
!  camera_incl_stars                Put to 1 for full SED with starlight
!  camera_pointing_position(1:3)    As with images
!  camera_image_halfsize_x,y        This is the size of the image 
!                                   in which the fluxes are computed for
!                                   the spectrum
!  camera_zoomcenter_x,y            As with images
!  camera_pointing_degr_*           As with images
!  camera_refine_criterion          As with images
!  camera_observer_degr_theta       As with images
!  camera_observer_degr_phi         As with images
!  camera_use_aperture_info         =.false. --> Integrates square image
!                                   =.true.  --> Uses aperture
!  
!-------------------------------------------------------------------------
subroutine camera_make_spectrum()
  use constants_module
  implicit none
  integer :: inu,backup_nrrefine,backup_nx,backup_ny,ierr,backup_nrfreq
  integer :: backup_lines_mode
  double precision :: pdx,pdy,factor,r_colarea,r,eps
  double precision, allocatable :: backup_freq(:),backup_spectrum(:)
  double precision, allocatable :: mask(:,:)
  integer :: ix,iy,apinu
  integer :: iact
  logical :: redo
  !
  ! Check
  !
  if(igrid_coord.eq.10) then
     write(stdo,*) 'ERROR: In 1-D plane-parallel mode the spectrum mode is not'
     write(stdo,*) '       available, because for an infinitely large slab'
     write(stdo,*) '       it anyway does not have meaning. Instead you should'
     write(stdo,*) '       make multi-wavelength 1x1-pixel images to get the'
     write(stdo,*) '       frequency-dependent intensity.'
     stop
  endif
  if(igrid_coord.eq.20) then
     write(stdo,*) 'ERROR: In 2-D pencil-parallel mode the spectrum mode is not'
     write(stdo,*) '       available, because for an infinitely long slab'
     write(stdo,*) '       it anyway does not have meaning. Instead you should'
     write(stdo,*) '       make multi-wavelength 1xN-pixel images to get the'
     write(stdo,*) '       frequency-dependent intensity.'
     stop
  endif
  if((camera_nrfreq.le.0).or.(.not.allocated(camera_frequencies))) then
     write(stdo,*) 'ERROR: Somehow the frequencies for the spectrum are not set.'
     stop
  endif
  if(camera_localobserver) then
     write(stdo,*) 'ABORTING: For the moment we do not allow the making of a spectrum'
     write(stdo,*) '          in the local observer mode. Use observer at infinity mode,'
     write(stdo,*) '          for instance by specifying the incl and phi on the '
     write(stdo,*) '          command line and setting the image size with sizeau or sizepc.'
     stop
  endif
  if((((scattering_mode.ne.0).and.(.not.camera_scatsrc_allfreq)).or.   &
      camera_secondorder).and.(lines_mode.lt.-1).and.rt_incl_lines) then
     write(stdo,*)    '===> Warning: Combination of modes that may make RADMC-3D very slow... Here is why:'
     if((scattering_mode.ne.0).and.(.not.camera_scatsrc_allfreq)) then
        write(stdo,*) '     You are including dust scattering (doing it freq-by-freq), which means that'
     else
        if(camera_catch_doppler_line) then
           write(stdo,*) '     You are using the doppler catching mode, which requires second order '
           write(stdo,*) '     integration of the RT equation, which (for memory reasons) means that'
        else
           write(stdo,*) '     You are using second order integration of the RT equation,'
           write(stdo,*) '     which (for memory reasons) means that:'
        endif
     endif
     write(stdo,*) '     RADMC-3D must make one image for each frequency at a time to compute each '
     write(stdo,*) '     point of the spectrum. At the same time you use on-the-fly computation of'
     write(stdo,*) '     the line level populations, which are not stored. If you have 100 frequencies'
     write(stdo,*) '     in your spectrum, this means that 100 times an image is made, and a 100 times'
     write(stdo,*) '     the populations are RECALCULATED. If the population calculation is heavy, this'
     write(stdo,*) '     means that you waste a huge amount of time. You may want to calculate these'
     write(stdo,*) '     populations ONCE and store them, which can be done using a positive lines_mode'
     write(stdo,*) '     number (see manual). This may require a lot of memory if you have a molecule'
     write(stdo,*) '     with many levels. To overcome this, you can tell RADMC-3D to store only a subset'
     write(stdo,*) '     of the computed level populations: Only those related to the line you wish to'
     write(stdo,*) '     model. See manual.'
  endif
  !
  ! Make a backup of the camera_nrrefine, and set camera_nrrefine to infinity
  !
  backup_nrrefine = camera_nrrefine 
  camera_nrrefine = 10000
  !
  ! Set some other values
  !
  camera_localobserver = .false.      ! Spectra only for observer at infinity for now
  camera_tracemode = 1
  camera_nrrefine = 100
  !
  ! Check if the aperture file is present and read if needed
  !
  if(camera_use_aperture_info) then
     !
     ! Read the file aperture_info.inp
     ! 
     call camera_read_aperture_info()
     if(.not.allocated(camera_aperture_freq)) stop 8341
     if(.not.allocated(camera_aperture_radius_collectarea)) stop 8342
     !
     ! Check if the current wavelengths are all within range
     !
     if((max(camera_frequencies(1),camera_frequencies(camera_nrfreq)).gt.           &
         max(camera_aperture_freq(1),camera_aperture_freq(camera_aperture_nf))).or. &
        (min(camera_frequencies(1),camera_frequencies(camera_nrfreq)).lt.           &
         min(camera_aperture_freq(1),camera_aperture_freq(camera_aperture_nf)))) then
        write(stdo,*) 'ERROR while making spectrum with aperture information:'
        write(stdo,*) '      The aperture information does not cover all wavelengths'
        write(stdo,*) '      that are to be used for the spectrum.'
        stop
     endif
  endif
  !
  ! If the dust emission is included, then make sure the dust data,
  ! density and temperature are read. If yes, do not read again.
  !
  if(rt_incl_dust) then
     call read_dustdata(1)
     call read_dust_density(1)
     call read_dust_temperature(1)
  endif
  !
  ! If line emission is included, then make sure the line data are
  ! read. If yes, then do not read it again.
  !
  if(rt_incl_lines) then
     call read_lines_all(1)
  endif
  !
  ! Now switch between different methods of making the spectrum
  !
  if(.not.camera_use_aperture_info) then
     !
     ! Simplest form of spectrum: a 1x1 pixel image with recursive
     ! pixel rendering
     !
     backup_nx = camera_image_nx
     backup_ny = camera_image_ny
!     camera_image_nx = 1
!     camera_image_ny = 1
!     !
!     ! Init or re-init the camera
!     !
!     call camera_init()
!     !
!     ! Set some values
!     !
!     pdx = 2*camera_image_halfsize_x / (1.d0*camera_image_nx)
!     pdy = 2*camera_image_halfsize_y / (1.d0*camera_image_ny)
!     !
!     ! Make the 1x1 pixel images
!     !
!     call camera_make_rect_image(0)
!     !
!     ! Now make the spectrum at 1 parsec distance
!     !
!     factor = pdx*pdy/(parsec**2)
!     do inu=1,camera_nrfreq
!        camera_spectrum(inu) = camera_rect_image(1,1,inu) * factor
!     enddo
!
!    (To prevent singularities in the calculations of the rays for 
!    spherical coordinates with incl=0, we make, from the start, a
!    2x2 image instead of a 1x1 image; but this is essentially identical
!    since the sub-pixeling would anyway refine the image)
!
     camera_image_nx = 2
     camera_image_ny = 2
     !
     ! Init or re-init the camera
     !
     call camera_init()
     !
     ! Set some values
     !
     pdx = 2*camera_image_halfsize_x / (1.d0*camera_image_nx)
     pdy = 2*camera_image_halfsize_y / (1.d0*camera_image_ny)
     !
     ! Make the 2x2 pixel images
     !
     call camera_make_rect_image(0)
     !
     ! Now make the spectrum at 1 parsec distance
     !
     factor = pdx*pdy/(parsec**2)
     do inu=1,camera_nrfreq
        camera_spectrum(inu) =                     &
             camera_rect_image(1,1,inu) * factor + &
             camera_rect_image(1,2,inu) * factor + &
             camera_rect_image(2,1,inu) * factor + &
             camera_rect_image(2,2,inu) * factor
     enddo
     !
     ! Restore the camera_nrrefine and the image nx,ny to their original value
     ! 
     camera_nrrefine = backup_nrrefine 
     camera_image_nx = backup_nx
     camera_image_ny = backup_ny
  else
     !
     ! Use the aperture information to make the spectrum
     !
     ! Use the nx,ny as in the images, but check that they are not too small.
     !
     if((camera_image_nx.lt.10).or.(camera_image_ny.lt.10)) then
        write(stdo,*) 'ERROR in camera module while making spectrum:'
        write(stdo,*) '      Taking npix less than 10 in either direction '
        write(stdo,*) '      is bound to yield wrong results.'
        write(stdo,*) '      nx,ny = ',camera_image_nx,camera_image_ny
        stop
     endif
     !
     ! Now make a separate image for each wavelength, where the size of the image is
     ! tuned to the size of the circular collecting area. 
     !
     ! We must first "fool" the imaging routine in thinking that we have only 1 frequency.
     ! So we must make a backup of the camera_frequencies() array and also allocate
     ! memory for the spectrum.
     !
     ! For 1 <= lines_mode <= 9 we must also tell the imaging routine that the
     ! level populations have already been calculated, and calculate them here.
     ! Otherwise the imaging routine will recalculate the populations every time
     ! again, which is very time-consuming for e.g. LVG or opt. thin. populations.
     !
     allocate(backup_freq(1:camera_nrfreq),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in spectrum: could not allocate backup_freq() array'
        stop
     endif
     allocate(backup_spectrum(1:camera_nrfreq),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in spectrum: could not allocate backup_spectrum() array'
        stop
     endif
     backup_spectrum(:) = -1d33     ! Purposely put to unphysical value, for check
     backup_freq(:) = camera_frequencies(:)
     backup_nrfreq  = camera_nrfreq
     backup_lines_mode = lines_mode
     !
     ! So if 1 <= lines_mode <= 9 we calculate the populations beforehand and
     ! put lines_mode to lines_mode=50, which means that the imaging routine
     ! will simply use the level populations that are present.
     !
     if(rt_incl_lines.and.(lines_mode.ge.1).and.(lines_mode.le.9)) then
        !
        ! Level subset selection and computation of the populations
        !
        if(lines_autosubset) then
           !
           ! If requested, do an automatic subset selection of the
           ! molecular levels
           ! 
           call lines_automatic_subset_selection(camera_nrfreq,     &
                camera_frequencies,1,camera_nrfreq,lines_maxshift,  &
                redo)
           !
           ! Now force a recomputation of the populations, unless
           ! "redo" is .false., meaning that we have the same
           ! levels as before (and thus the population as computed
           ! before must be still correct). Note that if there
           ! exists no lines_levelpop() array, then redo will also
           ! be .true.
           !
           if(redo) then
              iact=2
           else
              iact=1
           endif
           call lines_compute_and_store_local_populations(iact)
        else
           !
           ! Subset is not automatically selected. So either it has
           ! been selected manually or not at all (in which case the
           ! "subset" is the full set of levels).
           !
           write(stdo,*) 'Will store level populations for all levels or ',   &
                'for a manually (in lines.inp) selected subset of levels.'
           !
           ! Compute the populations if they have not yet been
           ! computed beforehand
           !
           call lines_compute_and_store_local_populations(1)
           !
        endif
        !
        ! Set lines_mode to 50 (a backup has been made, so it will be
        ! restored)
        !
        lines_mode = 50
     endif
     !
     ! From now on we will do things 1 frequency at a time
     !
     camera_nrfreq  = 1
     !
     ! Make a circular mask
     !
     allocate(mask(camera_image_nx,camera_image_ny),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in spectrum: could not allocate mask() array'
        stop
     endif
     do iy=1,camera_image_ny
        do ix=1,camera_image_nx
           r = sqrt( ((ix-0.5d0-0.5d0*camera_image_nx)/(0.5d0*camera_image_nx))**2 + &
                     ((iy-0.5d0-0.5d0*camera_image_ny)/(0.5d0*camera_image_ny))**2 )
           if(r.le.1.d0) then
              mask(ix,iy) = 1.d0
           else
              mask(ix,iy) = 0.d0
           endif
        enddo
     enddo
     !
     ! Re-init the camera, now making it think that we have just 1 frequency
     !
     call camera_init()
     !
     ! Now a loop over all frequencies
     !
     do inu=1,backup_nrfreq
        !
        ! Set frequency for this image
        !
        camera_frequencies(1) = backup_freq(inu)
        !
        ! Now interpolate the aperture size from the tables
        ! Use logarithmic interpolation, which preserves
        ! the powerlaw character of the beam size as a
        ! function of wavelength.
        !
        call hunt(camera_aperture_freq,camera_aperture_nf, &
                  camera_frequencies(1),apinu)
        if((apinu.lt.1).or.(apinu.ge.camera_aperture_nf)) then
           write(stdo,*) 'INTERNAL ERROR in spectrum: aperture index out of range...'
           stop 6398
        endif
        eps = (log(camera_frequencies(1))-log(camera_aperture_freq(apinu))) / &
              (log(camera_aperture_freq(apinu+1))-log(camera_aperture_freq(apinu)))
        if((eps.lt.0.d0).or.(eps.gt.1.d0)) stop 6399
        r_colarea = exp((1.d0-eps) * log(camera_aperture_radius_collectarea(apinu)) + &
                    eps * log(camera_aperture_radius_collectarea(apinu+1)))
        !
        ! Convert this radius from arcsec into cm for this image, using the
        ! distance specified by the observer
        !
        r_colarea = r_colarea * AU * camera_observer_distance_pc
        !
        ! Set the pixel sizes
        !
        pdx = 2*r_colarea / (1.d0*camera_image_nx)
        pdy = 2*r_colarea / (1.d0*camera_image_ny)
        !
        ! Make the image
        !
        call camera_make_rect_image(0)
        !
        ! Now compute the flux at 1 parsec distance, using the mask
        !
        factor = pdx*pdy/(parsec**2)
        backup_spectrum(inu) = 0.d0
        do iy=1,camera_image_ny
           do ix=1,camera_image_nx
              backup_spectrum(inu) = backup_spectrum(inu) + &
                   mask(ix,iy)*camera_rect_image(ix,iy,1) * factor
           enddo
        enddo
     enddo
     !
     ! Reset the camera frequencies
     !
     camera_nrfreq = backup_nrfreq
     camera_frequencies(:) = backup_freq(:)
     !
     ! Restore lines_mode
     !
     lines_mode = backup_lines_mode
     !
     ! Re-init camera, to get the right nr of frequencies 
     ! in the arrays back again.
     !
     call camera_init()
     !
     ! Now reinstall the frequencies and copy the spectrum
     !
     camera_frequencies(:) = backup_freq(:)
     camera_spectrum(:) = backup_spectrum(:)
     !
  endif
  !
  ! Deallocate stuff
  !
  if(allocated(backup_freq)) deallocate(backup_freq)
  if(allocated(backup_spectrum)) deallocate(backup_spectrum)
  if(allocated(mask)) deallocate(mask)
  !
end subroutine camera_make_spectrum



!--------------------------------------------------------------------------
!                         WRITE THE SPECTRUM 
!--------------------------------------------------------------------------
subroutine camera_write_spectrum()
  use constants_module
  implicit none
  integer :: inu
  double precision :: lambda
  !
  ! Open file
  !
  if(stdo.eq.6) then
     !
     ! Current radmc run is not a child process 
     ! So write output to a file  (otherwise to standard out)
     !
     open(unit=fflo,file='spectrum.out')
  else
     !
     ! Current radmc run is a child process, so the spectrum must
     ! be written to standard out. Check if fflo is indeed 6.
     !
     if(fflo.ne.6) then
        write(stdo,*) 'ERROR: fflo.ne.6 while stdo.ne.6'
        stop
     endif
  endif
  !
  ! Now write
  !
  write(fflo,*) 1                             ! Format number
  write(fflo,*) camera_nrfreq
  write(fflo,*) ' '
  do inu=1,camera_nrfreq
     lambda = 1d4*cc/camera_frequencies(inu)
     if(camera_spectrum(inu).gt.1.d-97) then
        write(fflo,10) lambda,camera_spectrum(inu)
     else
        write(fflo,10) lambda,0.d0
     endif
  enddo
10 format(E21.13,1X,E21.13)
  !
  ! Now close file or flush output
  !
  if(stdo.eq.6) then
     !
     ! Output went to file, so close that file
     !
     close(1)
  else
     !
     ! Output went to standard output, so flush this
     !
     call flush(fflo)
  endif
  !
end subroutine camera_write_spectrum



!--------------------------------------------------------------------------
!                      READ THE APERTURE INFORMATION
!--------------------------------------------------------------------------
subroutine camera_read_aperture_info()
  implicit none
  logical :: fex
  integer :: iformat,ierr,inu
  !
  ! Check that the distance is set
  !
  if(camera_observer_distance_pc.eq.0.d0) then
     write(stdo,*) 'ERROR: When using the aperture information, please'
     write(stdo,*) '       also specify the distance with dpc on the command line.'
     stop
  endif
  !
  ! First check if the file is present
  !
  inquire(file='aperture_info.inp',exist=fex)
  if(.not.fex) then
     write(stdo,*) 'ERROR: No file aperture_info.inp found.'
     stop
  endif
  !
  ! If some aperture stuff is already allocated, then deallocate
  !
  if(allocated(camera_aperture_freq)) deallocate(camera_aperture_freq)
  if(allocated(camera_aperture_radius_collectarea)) deallocate(camera_aperture_radius_collectarea)
  !
  ! Now open the file
  !
  open(unit=1,file='aperture_info.inp')
  read(1,*) iformat
  if(iformat.eq.1) then
     !
     ! With iformat==1 the radius of the collecting area as a function of
     ! wavelength is given.
     !
     read(1,*) camera_aperture_nf
     allocate(camera_aperture_freq(1:camera_aperture_nf),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in aperture information reading: could not allocate array'
        stop
     endif
     allocate(camera_aperture_radius_collectarea(1:camera_aperture_nf),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in aperture information reading: could not allocate array'
        stop
     endif
     do inu=1,camera_aperture_nf
        read(1,*) camera_aperture_freq(inu),camera_aperture_radius_collectarea(inu)
        camera_aperture_freq(inu) = 1d4*cc/camera_aperture_freq(inu) ! Convert to Hz
     enddo
     !
     ! Check that the frequencies are monotonic
     !
     if(camera_aperture_freq(camera_aperture_nf).gt.camera_aperture_freq(1)) then
        !
        ! Check monotonicity
        !
        do inu=2,camera_aperture_nf
           if(camera_aperture_freq(inu).le.camera_aperture_freq(inu-1)) then
              write(stdo,*) 'ERROR in aperture_info.inp file:'
              write(stdo,*) '      wavelengths not monotonic.'
              stop
           endif
        enddo
        !
        ! Just for safety, expand the range a tiny bit
        !
        camera_aperture_freq(1)                   = 0.999999d0 * camera_aperture_freq(1)
        camera_aperture_freq(camera_aperture_nf)  = 1.000001d0 * camera_aperture_freq(camera_aperture_nf)
     else
        !
        ! Check monotonicity
        !
        do inu=2,camera_aperture_nf
           if(camera_aperture_freq(inu).ge.camera_aperture_freq(inu-1)) then
              write(stdo,*) 'ERROR in aperture_info.inp file:'
              write(stdo,*) '      wavelengths not monotonic.'
              stop
           endif
        enddo
        !
        ! Just for safety, expand the range a tiny bit
        !
        camera_aperture_freq(1)                   = 1.000001d0 * camera_aperture_freq(1)
        camera_aperture_freq(camera_aperture_nf)  = 0.999999d0 * camera_aperture_freq(camera_aperture_nf)
     endif
  else
     write(stdo,*) 'ERROR: For now ',iformat,' is not known in aperture_info.inp'
     stop
  endif
  close(1)
  !
end subroutine camera_read_aperture_info



!----------------------------------------------------------------------------
!                  RAY TRACING IN GPU-PARALLELIZABLE WAY
!
! This routine is meant to emulate the way a GPU can parallelize the ray-
! tracing in frequency. 
!----------------------------------------------------------------------------
subroutine camera_ray1d_raytrace(nrfreq,x,y,z,dx,dy,dz,distance,celldxmin,&
                                   intensity)
  implicit none
  double precision :: x,y,z,dx,dy,dz,dum,distance,celldxmin
  double precision :: freq,alpnu,jnu,alpha_a,dummy,expt,temp,src
  integer :: ispec,nrfreq,deepestlevel
  integer :: inu,levelnext,is
  logical :: arrived
  double precision :: intensity(nrfreq)
  double precision :: rlen,cosp,sinp,cost,sint,vx,vy,vz
  double precision :: xav,yav,zav
  type(amr_branch), pointer :: a
  integer :: ixx,iyy,izz,bc_idir,bc_ilr
  !
  ! Do checks
  !
  write(stdo,*) 'ERROR: GPU emulator not yet ready'
  stop
  if(nrfreq.ne.camera_nrfreq) stop 3345
  if(camera_tracemode.ne.1) then
     write(stdo,*) 'ERROR in camera module: for lines only tracemode=1 allowed.'
     stop
  endif
  if(debug_check_all.eq.1) then
     if(rt_incl_lines) then
        if(.not.allocated(gastemp)) then
           write(stdo,*) 'ERROR in line ray tracing: gas temperature not set.'
           stop
        endif
        if(.not.allocated(lines_ray_levpop)) then
           write(stdo,*) 'ERROR in camera_ray1d_raytrace: lines module not ready:'
           write(stdo,*) '  lines_ray_levpop not allocated.'
           stop
        endif
        if(lines_mode.ge.1) then
           if(.not.allocated(lines_levelpop)) then
              write(stdo,*) 'ERROR in camera_ray1d_raytrace: lines module not ready:'
              write(stdo,*) '  lines_levelpop not allocated.'
              stop
           endif
        endif
     endif
     dum = sqrt(dx**2 + dy**2 + dz**2)
     if(abs(dum-1.d0).gt.1d-6) stop 3487
     if(.not.allocated(camera_frequencies)) then
        write(stdo,*) 'ERROR: Wanting to ray-trace using general method, but'
        write(stdo,*) '       camera_frequencies is not allocated.'
        stop
     endif
  endif
  !
  ! Set starting value of celldxmin, i.e. the size of the smallest cell
  ! we will pass through. Note that for cartesian coordinates, since cell 
  ! sizes will then always be cubes of the base size times (1/2)^level
  ! we can use just the deepest level to compute celldxmin.
  !
  ! NOTE: In the AMR module we now changed the level-numbering: Now
  !       the base grid level is level=0 (used to be 1). Hence we must
  !       now set deepestlevel = 0 here, and the multiplication factor
  !       is now (1/2)^level instead of (1/2)^(level-1).
  !
  celldxmin     = 1d99
  deepestlevel  = 0
  !
  ! Set the starting point and the direction
  !
  ray_cart_x    = x
  ray_cart_y    = y
  ray_cart_z    = z
  ray_cart_dirx = dx
  ray_cart_diry = dy
  ray_cart_dirz = dz
  !
  ! Put the distance from the starting point to the observer here
  ! If observer at infinity, simply put distance=1d99
  !
  ray_dsend     = distance 
  !
  ! Check if we start inside or outside the domain
  !
  if(igrid_type.lt.100) then 
     !
     ! We use a regular or AMR grid
     !
     if(igrid_coord.lt.100) then
        !
        ! We use cartesian coordinates
        !
        if(amr_tree_present) then
           !
           ! We use the AMR tree
           !
           call amr_findcell(x,y,z,a)
           if(associated(a)) then
              !
              ! We start inside the domain
              !
              ray_index     = a%leafindex
              ray_indexnext = 0
           else
              !
              ! We start outside the domain, so all indices must be 0
              !
              ray_index     = 0
              ray_indexnext = 0
           endif
        else
           !
           ! We have a regular grid
           !
           call amr_findbasecell(x,y,z,ixx,iyy,izz)
           if(ixx.gt.0) then
              ray_index     = ixx+((iyy-1)+(izz-1)*amr_grid_ny)*amr_grid_nx
              ray_indexnext = 0
           else
              ray_index     = 0
              ray_indexnext = 0
           endif
        endif
        !
     elseif(igrid_coord.lt.200) then
        !
        ! We use spherical coordinates
        !
        write(stdo,*) 'FINDING CELL NOT YET READY FOR SPHERICAL COORDINATES'
        stop
        !
     else
        write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
        stop
     endif
  else
     write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
     stop
  endif
  !
  ! Reset arrays for the 1-D line tracing
  !
  camera_ray_jnu(:,:)   = 0.d0
  camera_ray_alpnu(:,:) = 0.d0
  ray_ns                = 0
  is = 1
  !
  ! Now trace
  !
  arrived = .false.
  do while(.not.arrived)
     !
     ! Check if the ray is not too long
     !
     if(is.gt.ray_nsmax) then
        write(stdo,*) 'ERROR in camera/lines: ray length exceeds maximum.'
        stop
     endif
     !
     ! Back up the current position
     !
     ray_prev_x     = ray_cart_x
     ray_prev_y     = ray_cart_y
     ray_prev_z     = ray_cart_z
     !
     ! Find new position
     !
     if(igrid_type.lt.100) then 
        !
        ! We use a regular or AMR grid
        !
        if(igrid_coord.lt.100) then
           !
           ! We use cartesian coordinates
           !
           call amrray_find_next_location_cart(ray_dsend,            &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived,              &
                levelnext=levelnext)
           !
           ! Check if this cell is the smallest so far
           !
           deepestlevel = max(deepestlevel,levelnext)
           !
           ! If we went through one of the boundaries of the
           ! domain, we might need to implement the thermal
           ! boundary, if present
           !
           if(incl_thermbc.ne.0) then
              if((ray_index.eq.0).and.(ray_indexnext.ne.0)) then
                 !
                 ! First determine which boundary we just went
                 ! through 
                 !
                 if(amrray_icross.ge.0) stop 3902  ! Self-consistency check
                 if(amrray_icross.eq.-1) then
                    bc_idir = 1
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-2) then
                    bc_idir = 1
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-3) then
                    bc_idir = 2
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-4) then
                    bc_idir = 2
                    bc_ilr  = 2
                 elseif(amrray_icross.eq.-5) then
                    bc_idir = 3
                    bc_ilr  = 1
                 elseif(amrray_icross.eq.-6) then
                    bc_idir = 3
                    bc_ilr  = 2
                 else
                    stop 4944
                 endif
                 !
                 ! Now set the intensity to the thermal
                 ! emission of the boundary, if the
                 ! thermal boundary is switched on
                 !
                 if(thermal_bc_active(bc_ilr,bc_idir)) then
                    do inu=1,camera_nrfreq
                       intensity(inu) = bplanck(thermal_bc_temp(bc_ilr,bc_idir), &
                                                camera_frequencies(inu))
                    enddo
                 endif
              endif
           endif
           !
        elseif(igrid_coord.lt.200) then
           !
           ! We use spherical coordinates
           !
           call amrray_find_next_location_spher(ray_dsend,           &
                ray_cart_x,ray_cart_y,ray_cart_z,                    &
                ray_cart_dirx,ray_cart_diry,ray_cart_dirz,           &
                ray_index,ray_indexnext,ray_ds,arrived)
           !
           ! Check if this cell is the smallest so far
           !
           ! **** NOT YET READY. Use the camera_amr_zoomfact array
           ! **** multiplied by the local radius to get the correct
           ! **** celldxmin
           write(stdo,*) 'FINDING CELLDXMIN NOT YET READY FOR SPHERICAL COORDINATES'
           stop
           !
        else
           write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
           stop
        endif
        !
        ! Set the camera_istar, which is used for treating finite-size
        ! stars (instead of point source stars)
        !
        camera_istar = amrray_ispherehit
        !   
     else
        write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
        stop
     endif
     !
     ! Check if we are in a cell or not
     !
     if(debug_check_all.eq.1) then
        if(ray_index.gt.nrcellsmax) stop 6271
     endif
     if(ray_index.ge.1) then
        !
        ! Set the length 
        !
        ray_ds            = sqrt( (ray_cart_x-ray_prev_x)**2 + (ray_cart_y-ray_prev_y)**2 + (ray_cart_z-ray_prev_z)**2 )
        camera_ray_ds(is) = ray_ds
        !
        ! Set the dust quantities
        !
        if(rt_incl_dust) then
           if(allocated(dustdens)) then
              camera_dustdens(:) = dustdens(:,ray_index)
              camera_dusttemp(:) = dusttemp(:,ray_index)
           else
              camera_dustdens(:) = 0.d0
              camera_dusttemp(:) = 0.d0
           endif
        endif
        !
        ! Find the local dust j_nu and alpha_nu
        !
        !  ############# FUTURE: ADD SCATTERING ALPHA AND SOURCE AS WELL ###########
        !
        if(rt_incl_dust) then
           temp    = 100.d0     ! Arbitrary temperature...
           do inu=1,camera_nrfreq
              freq  = camera_frequencies(inu)
              alpnu = 0.d0
              jnu   = 0.d0
              do ispec=1,dust_nr_species
                 alpha_a = camera_dustdens(ispec) * find_dust_kappa_interpol(freq,ispec,temp,1,0,0)
                 !      alpha_s = camera_dustdens(ispec) * find_dust_kappa_interpol(freq,ispec,temp,0,1,0)
                 alpnu   = alpnu + alpha_a ! + alpha_s
                 jnu     = jnu + alpha_a * bplanck(camera_dusttemp(ispec),camera_frequencies(inu))
              enddo
              camera_ray_jnu(inu,is)   = jnu
              camera_ray_alpnu(inu,is) = alpnu
           enddo
        else
           camera_ray_jnu(:,is)     = 0.d0
           camera_ray_alpnu(:,is) = 0.d0
        endif
        !
        ! Find the local line j_nu and alpha_nu
        !
        if(rt_incl_lines) then
           !
           ! Set the doppler shift
           !
           if(igrid_coord.lt.100) then
              !
              ! Cartesian coordinates:
              !
              dummy = gasvelocity(1,ray_index) * ray_cart_dirx +   &
                      gasvelocity(2,ray_index) * ray_cart_diry +   &
                      gasvelocity(3,ray_index) * ray_cart_dirz
           elseif(igrid_coord.lt.200) then
              !
              ! Spherical coordinates: Here component 1 is the v_r component,
              ! component 2 is the v_theta component (in the same unit as
              ! v_r, i.e. cm/s) and component 3 is the v_phi component (also
              ! in the same unit: cm/s).
              !
              xav   = 0.5d0 * ( ray_cart_x + ray_prev_x )
              yav   = 0.5d0 * ( ray_cart_y + ray_prev_y )
              zav   = 0.5d0 * ( ray_cart_z + ray_prev_z )
              dummy = xav**2+yav**2
              rlen  = sqrt(dummy+zav**2)+1d-90
              dummy = sqrt(dummy)+1d-90
              cosp  = xav/dummy
              sinp  = yav/dummy
              cost  = zav/rlen
              sint  = dummy/rlen
              dummy = sint * gasvelocity(1,ray_index) +   &
                      cost * gasvelocity(2,ray_index) 
              vx    = cosp * dummy - sinp * gasvelocity(3,ray_index)
              vy    = sinp * dummy + cosp * gasvelocity(3,ray_index)
              vz    = cost * gasvelocity(1,ray_index) -   &
                      sint * gasvelocity(2,ray_index)
              dummy = vx * ray_cart_dirx +   &
                      vy * ray_cart_diry +   &
                      vz * ray_cart_dirz
           else
              stop 501
           endif
           lines_ray_doppler(is) = dummy / cc
           !
           ! Set the gas temperature
           !
           lines_ray_temp(is)    = gastemp(ray_index)
           !
           ! Set the microturbulence
           !
           if(allocated(lines_microturb)) then
              lines_ray_turb(is) = lines_microturb(ray_index)
           else
              lines_ray_turb(is) = 0.d0
           endif
           !
           ! Set the level populations
           !
           ! NOTE: THE STUFF BELOW IS STILL NOT UP TO DATE
           !
           if(lines_mode.ge.1) then
              !
              ! We use the level populations from the global array
              ! **** THE LINE BELOW IS WRONG ****
              lines_ray_levpop(:,:,is) = lines_levelpop(:,:,ray_index)
           else
              !
              ! We compute the level populations on-the-fly
              !
              do ispec=1,lines_nr_species
                 call lines_compute_ltepop_subset(lines_nrlevels(ispec),    &
                        active_nrlevels(ispec),                             &
                        active_levels(:,ispec),                             &
                        ispec,lines_ray_temp(is),                           &
                        gas_chemspec_numberdens(ispec,ray_index),           &
                        lines_ray_levpop(:,ispec,is))
              enddo
           endif
           !
        endif
     else
        !
        ! Reset all to zero
        !
        ray_ds                = 0.d0
        camera_ray_ds(is)     = 0.d0
        if(rt_incl_lines) then
           lines_ray_doppler(is) = 0.d0
           lines_ray_temp(is)    = 0.d0
           lines_ray_turb(is)    = 0.d0
           lines_ray_levpop(:,:,is) = 0.d0
        endif
        !
     endif
     !
     ! Increase is
     !
     is = is + 1
     !
     ! Decrease ray_dsend
     !
     ray_dsend = ray_dsend - ray_ds
     !
     ! Now make next cell the current cell
     !
     ray_index = ray_indexnext
     !
     ! Do the same for the pointers.
     ! This is grid type dependent.
     !
     if(igrid_type.lt.100) then
        !
        ! AMR type grid
        !
        if(ray_index.gt.0) then
           !
           ! We are in a cell
           !
           if(amr_tree_present) then
              amrray_cell => amr_index_to_leaf(ray_index)%link
           else
              call amr_regular_get_ixyz(ray_index,amrray_ix_curr,amrray_iy_curr,amrray_iz_curr)
           endif
        else
           nullify(amrray_cell)
           amrray_ix_curr = -1
           amrray_iy_curr = -1
           amrray_iz_curr = -1
        endif
     else
        write(stdo,*) 'ERROR: This grid type not yet implemented'
        stop
     endif
     !
  enddo
  !
  ! Set the length of the 1-D ray
  !
  ray_ns = is - 1
  !
  ! Return the smallest cell size we encountered
  !
  if(igrid_type.lt.100) then 
     if(igrid_coord.lt.100) then
        celldxmin = (amr_grid_xi(2,1)-amr_grid_xi(1,1)) / (2**deepestlevel)
     elseif(igrid_coord.lt.200) then
        write(stdo,*) 'For spherical coordinates, the celldxmin not yet ready'
        stop
     else
        write(stdo,*) 'ERROR: Cylindrical coordinates not yet implemented'
        stop
     endif
  else
     write(stdo,*) 'SORRY: Delaunay or Voronoi grids not yet implemented'
     stop
  endif
  !
  ! Now install the line sources in the 1-D ray
  !
  if(rt_incl_lines) then
!!!     call lines_addto_jnu_alpnu()
     stop 3333
  endif
  !
  ! Now call the ray tracer of the lines module
  !
  intensity(:) = 0.d0
  !
  ! Do a loop over frequencies
  !
  ! ############## FUTURE: FOR NVIDIA GRAPHICS CARD: PARALLELIZE IN NU HERE #########
  !
  do inu=1,camera_nrfreq
     !
     ! Loop along the ray
     !
     do is=1,ray_ns
        !
        ! Each segment is assumed to be at constant j_nu and alpha_nu.
        !
        ! ############ FUTURE: HERE WE MUST CATCH THIN-LINE DOPPLER 'QUANTUM LEAPS' ############
        !
        expt   = exp(-camera_ray_ds(is)*camera_ray_alpnu(inu,is))
        src    = camera_ray_jnu(inu,is)/camera_ray_alpnu(inu,is)
        intensity(inu) = intensity(inu) * expt + (1.d0-expt) * src
     enddo
  enddo
  !
  ! Done
  !
end subroutine camera_ray1d_raytrace



!----------------------------------------------------------------------------
!                SET THE FREQUENCIES FOR THE CAMERA
!----------------------------------------------------------------------------
subroutine set_camera_frequencies(nfr,freq1,freq2)
  implicit none
  logical :: flag
  integer :: inu,ierr,i,iformat
  integer, intent(in), optional :: nfr
  double precision, intent(in), optional :: freq1,freq2
  logical :: fex1,fex2
  !
  ! Do some checks
  !
  if((lines_user_widthkms.gt.0.d0).or.(lines_user_iline.gt.0)) then
     !
     ! Apparently the user wants to make a spectral line spectrum
     !
     ! Check that at least both widthkms and iline are set
     !
     if((lines_user_widthkms.le.0).and.(lines_user_nrfreq.gt.1)) then
        write(stdo,*) 'ERROR: If you specify iline, you must also specify widthkms'
        write(stdo,*) '       (unless you make a single wavelength image).'
        write(stdo,*) '       Currently nr of freq = ',lines_user_nrfreq
        stop
     endif
     if(lines_user_iline.le.0) then
        write(stdo,*) 'ERROR: If you specify widthkms, you must also specify iline'
        stop
     endif
     !
     ! Some things can be defaults, if they are not yet set
     ! 
     if(lines_user_ispec.le.0) then
        lines_user_ispec = 1
        if(lines_nr_species.gt.1) then
           write(stdo,*) 'For line spectra we, by default, choose molecular species 1.'
           write(stdo,*) '   Specify with imolspec on the command line if you want another molecule.'
        endif
     endif
     if(lines_user_nrfreq.le.0) then
        lines_user_nrfreq = 100
        write(stdo,*) 'For line spectra we, by default, take nr of frequencies to 100.'
        write(stdo,*) '   Specify with linenlam or linenfreq if you want another nr.'
     endif
  endif
  !
  ! Deallocate frequencies
  !
  if(allocated(camera_frequencies)) deallocate(camera_frequencies)
  !
  ! Check whether to use the above nfr, freq1 and freq2 to make the
  ! frequencies
  !
  if(present(nfr)) then
     !
     ! A regularly spaced grid between freq1 and freq2
     !
     if((.not.present(freq1)).or.(.not.present(freq2))) then
        write(stdo,*) 'ERROR in set_camera_frequencies: either set all or no optional args.'
        stop
     endif
     camera_nrfreq = nfr
     allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
     if(ierr.ne.0) then
        write(stdo,*) 'ERROR in lines module: Could not allocate '
        write(stdo,*) '      camera_frequencies(:).'
        stop 
     endif
     if(camera_nrfreq.gt.1) then
        do inu=1,camera_nrfreq
           camera_frequencies(inu) = freq1 * (freq2/freq1)**   &
                ((inu-1.d0)/(camera_nrfreq-1.d0))
        enddo
     else
        camera_frequencies(1) = freq1
     endif
  else
     !
     ! Now make decisions where to get our frequency grid from
     !
     if((camera_lambdamic.gt.0.d0).and.(camera_lambdamic1.gt.0.d0).and. &
          (camera_range_nlam.gt.1)) then
        !
        ! Make a simple regular grid between lambdamic and lambdamic1
        ! using logarithmic spacing
        !
        camera_nrfreq = camera_range_nlam
        allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in camera module: Could not allocate '
           write(stdo,*) '      camera_frequencies(:).'
           stop 
        endif
        do inu=1,camera_nrfreq
           camera_frequencies(inu) = 1d4*cc / ( camera_lambdamic *   &
                (camera_lambdamic1/camera_lambdamic)**               &
                ((inu-1.d0)/(camera_nrfreq-1.d0)) )
        enddo
        !
     elseif(camera_lambdamic.gt.0.d0) then
        !
        ! Use exactly this wavelength, given on the command line
        !
        camera_nrfreq = 1
        allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in camera module: Could not allocate '
           write(stdo,*) '      camera_frequencies(:).'
           stop 
        endif
        camera_frequencies(1) = 1d4*cc/camera_lambdamic
        !
     elseif(camera_theinu.gt.0) then
        !
        ! Use one of the global frequency array frequencies, selected
        ! on the command line
        !
        if(.not.allocated(freq_nu)) stop
        camera_nrfreq = 1
        allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in camera module: Could not allocate '
           write(stdo,*) '      camera_frequencies(:).'
           stop 
        endif
        camera_frequencies(1) = freq_nu(camera_theinu)
        !
     elseif((lines_user_nrfreq.ge.1).and. &
        (lines_user_ispec.gt.0).and.(lines_user_iline.gt.0).and.        &
        rt_incl_lines) then
        !
        ! Use the line user-parameters to set the frequency grid 
        !
        ! Make sure the line data is read, otherwise we cannot specify
        ! the frequencies based on the line positions. 
        !
        call read_lines_all(1)
        !
        ! Check stuff
        !
        if(.not.allocated(lines_nu0)) then
           write(stdo,*) 'ERROR in camera module: while creating line frequencies array'
           write(stdo,*) 'line module not yet initialized.'
           stop
        endif
        if(lines_user_ispec.gt.lines_nr_species) then
           write(stdo,*) 'ERROR in line module: lines_user_ispec out of range'
           stop
        endif
        if(lines_user_iline.gt.lines_nrlines(lines_user_ispec)) then
           write(stdo,*) 'ERROR in line module: lines_user_iline out of range'
           stop
        endif
        !
        ! Allocate the camera frequencies array
        !
        camera_nrfreq = lines_user_nrfreq
        allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR in lines module: Could not allocate '
           write(stdo,*) '      camera_frequencies(:).'
           stop 
        endif
        !
        ! Make a window around a particular line
        !
        ! * CONVENTION * : Positive velocity = reddening (moving away from obs)
        !
        if(camera_nrfreq.gt.1) then
           do inu=1,camera_nrfreq
              camera_frequencies(inu) = lines_nu0(lines_user_iline,lines_user_ispec) *  &
                   (1.d0 - lines_user_kms0*1d5/cc -                                     &
                    (2*(inu-1.d0)/(lines_user_nrfreq-1.d0)-1.d0)*                       &
                    lines_user_widthkms*1d5/cc)
           enddo
        else
           camera_frequencies(1) = lines_nu0(lines_user_iline,lines_user_ispec) * &
                (1.d0-lines_user_kms0*1d5/cc)
        endif
     elseif(camera_loadlambdas) then
        !
        ! Read frequencies / wavelengths from a file
        !
        inquire(file='camera_frequency.inp',exist=fex1)
        inquire(file='camera_wavelength_micron.inp',exist=fex2)
        if(fex1.and.fex2) then
           write(stdo,*) 'ERROR in camera module: found camera_frequency.inp AND camera_wavelength_micron.inp'
           stop
        endif
        if(fex1) then
           open(unit=1,file='camera_frequency.inp')
           read(1,*) camera_nrfreq
           allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
           if(ierr.ne.0) then
              write(stdo,*) 'ERROR in camera module: Could not allocate '
              write(stdo,*) '      camera_frequencies(:).'
              stop 
           endif
           do inu=1,camera_nrfreq
              read(1,*) camera_frequencies(inu)
           enddo
           close(1)
        elseif(fex2) then
           open(unit=1,file='camera_wavelength_micron.inp')
           read(1,*) camera_nrfreq
           allocate(camera_frequencies(camera_nrfreq),STAT=ierr)
           if(ierr.ne.0) then
              write(stdo,*) 'ERROR in camera module: Could not allocate '
              write(stdo,*) '      camera_frequencies(:).'
              stop 
           endif
           do inu=1,camera_nrfreq
              read(1,*) camera_frequencies(inu)
              camera_frequencies(inu) = 1d4*cc/camera_frequencies(inu)
           enddo
           close(1)
        else
           write(stdo,*) 'ERROR: Cannot find camera_wavelength_micron.inp or'
           write(stdo,*) '       camera_frequency.inp file.'
           stop
        endif
     elseif(camera_loadcolor) then
        !
        ! Use a subset of wavelengths from the global frequency array
        !
        if(.not.allocated(freq_nu)) stop
        if(rt_incl_lines) then
           write(stdo,*) 'ERROR: Cannot use SED colors and do lines at the same time'
           stop
        endif
        if(camera_theinu.ge.1) then
           write(stdo,*) 'ERROR: Cannot use SED colors and specify a single frequency at the same time'
           stop
        endif
        inquire(file='color_inus.inp',exist=fex1)
        if(.not.fex1) then
           write(stdo,*) 'ERROR: Could not find file color_inus.inp'
           stop
        endif
        write(stdo,*) 'Reading color_inus.inp file...'
        call flush(stdo)
        open(unit=1,file='color_inus.inp')
        read(1,*) iformat
        read(1,*) camera_nrfreq
        if(camera_nrfreq.le.0) then
           write(stdo,*) 'ERROR: File color_inus.inp corrupt.'
           stop
        endif
        allocate(camera_frequencies(1:camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate camera_frequencies array'
           stop
        endif
        do i=1,camera_nrfreq
           read(1,*) inu
           if((inu.lt.1).or.(inu.gt.freq_nr)) then
              write(stdo,*) 'ERROR: in color_inus.inp inu out of range.'
              stop
           endif
           camera_frequencies(i) = freq_nu(inu)
        enddo
        close(1)
     elseif(camera_setfreq_global) then
        !
        ! Use the global frequency array
        !
        if(.not.allocated(freq_nu)) stop
        camera_nrfreq = freq_nr
        allocate(camera_frequencies(1:camera_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate camera_frequencies array'
           stop
        endif
        do inu=1,camera_nrfreq
           camera_frequencies(inu) = freq_nu(inu)
        enddo
     else
        write(stdo,*) 'ERROR: No method specified for determining the wavelength grid'
        write(stdo,*) '       of the camera module. Use, on the command line, one of the following:'
        write(stdo,*) '          inu or ilambda followed by an integer'
        write(stdo,*) '          nuhz or lambda followed by a number'
        write(stdo,*) '          color or loadcolor to let RADMC-3D read color_inus.inp'
        write(stdo,*) '          loadlambda to let RADMC-3D read camera_wavelength_micron.inp'
        write(stdo,*) '          globalfreq or globallam to use the global wavelength array.'
        stop
     endif
  endif
end subroutine set_camera_frequencies


!==========================================================================
!        SUBROUTINES FOR SECOND ORDER INTEGRATION ON VERTEX GRID
!==========================================================================

!--------------------------------------------------------------------------
!            COMPUTE SOURCE FUNCTION AND ALPHA AT VERTICES
!
! This subroutine is used for the second-order integration of the transfer
! equation. For this we need to know the sources at the corner points.
!
! NOTE: If camera_interpol_jnu is set, then snu = jnu rather than the
!       source function jnu/anu. 
!--------------------------------------------------------------------------
subroutine camera_compute_snualphanu_at_vertices(inu)
  implicit none
  integer :: inu,icell,ivt,ix,iy,iz,cnt,icnt
  integer :: ixx,iyy,izz
  double precision :: nu,snu,anu,dpl,trb,tmp
  double precision :: dirx,diry,dirz,xbk,ybk,zbk
  double precision, allocatable :: src(:),alp(:)
  double precision, allocatable :: jnubase(:),anubase(:)
  type(amr_branch), pointer :: b
  !
  ! Check if frequency index is in range
  !
  if(inu.gt.camera_nrfreq) then
     write(stdo,*) 'ERROR when computing snu and alphanu at vertices:'
     write(stdo,*) '      inu out of range.'
     stop
  endif
  !
  ! Allocate temporary arrays
  !
  allocate(src(1:camera_nrfreq))
  allocate(alp(1:camera_nrfreq))
  if(camera_catch_doppler_line) then
     allocate(jnubase(1:camera_vertex_lines_nractivetot))
     allocate(anubase(1:camera_vertex_lines_nractivetot))
  endif
  !
  ! Get frequency
  !
  nu = camera_frequencies(inu)
  !
  ! Set the direction vector, if observer is at infinity
  !
  if(.not.camera_localobserver) then
     dirx = 0.d0
     diry = 0.d0
     dirz = 1.d0
     ybk  = diry
     zbk  = dirz
     diry = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
     dirz = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
     xbk  = dirx
     ybk  = diry
     dirx = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
     diry =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     if(abs(dirx**2+diry**2+dirz**2-1.d0).gt.1d-6) then
        write(stdo,*) 'ERROR in camera module: direction vector not OK.'
        write(stdo,*) dirx,diry,dirz
        stop
     endif
     ray_cart_dirx = dirx
     ray_cart_diry = diry
     ray_cart_dirz = dirz
  endif
  !
  ! Fill the cell center values of j_nu and alpha_nu
  !
  do icell=1,nrcells
     !
     ! Find the position of the cell
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is present, so we get our information from there
        !
        ! Get the current cell index
        !
        ray_index = cellindex(icell)
        !
        ! Set the coordinates
        !
        b => amr_index_to_leaf(ray_index)%link
        if(.not.associated(b)) stop 3130
        ray_cart_x = amr_finegrid_xc(b%ixyzf(1),1,b%level)
        ray_cart_y = amr_finegrid_xc(b%ixyzf(2),2,b%level)
        ray_cart_z = amr_finegrid_xc(b%ixyzf(3),3,b%level)
     else
        !
        ! We have a regular grid, so no AMR tree available
        !
        ray_index = icell
        call amr_regular_get_ixyz(ray_index,ixx,iyy,izz)
        ray_cart_x = amr_finegrid_xc(ixx,1,0)
        ray_cart_y = amr_finegrid_xc(iyy,2,0)
        ray_cart_z = amr_finegrid_xc(izz,3,0)
     endif
     !
     ! Copy also to prev
     !
     ray_prev_x = ray_cart_x
     ray_prev_y = ray_cart_y
     ray_prev_z = ray_cart_z
     !
     ! If local observer, set the direction vector. For an observer
     ! at infinity this direction will be the same for all cells, and
     ! will thus be set beforehand, at the start of this routine.
     !
     if(camera_localobserver) then
        write(stdo,*) 'Local observer for now not possible in second order integration'
        stop
     endif
     !
     ! Compute the j_nu and alpha_nu, as well as (if requested)
     ! the line level populations and other line-related stuff.
     ! Note that if 'camera_catch_doppler_line' is switched on,
     ! we only calculate what is needed to compute the line 
     ! j_nu and alpha_nu; we do not add these to the src and alp
     ! yet. If this option is not set, then we DO add them to
     ! the src and alp.
     !
     call camera_get_src_alp(inu,inu,camera_nrfreq,src,alp)  
     !
     ! Fill the anu and jnu
     !
     camera_cell_anu(ray_index) = alp(inu)
     if(camera_interpol_jnu) then
        camera_cell_snu(ray_index) = src(inu)*alp(inu)
     else
        camera_cell_snu(ray_index) = src(inu)
     endif
     !
     ! If camera_catch_doppler_line option is set, then fill the line stuff
     !
     if(camera_catch_doppler_line) then
        !
        ! The line-of-sight Doppler shift at this point
        !
        camera_cell_doppler(ray_index) = lines_ray_doppler(1)
!        !
!        ! The local microturbulence at this point
!        !
!        camera_cell_turb(ray_index) = lines_ray_turb(1)
!        !
!        ! The local temperature at this point
!        ! NOTE: THIS IS NOT REALLY NECESSARY. DOUBLE STORAGE...
!        ! 
!        camera_cell_temp(ray_index) = gastemp(ray_index)
        !
        ! The line jnubase functions and the basic part of the
        ! line absorption coefficient (without the line profile).
        ! Both depend on the level populations.
        !
        call lines_jnu_anu_base(camera_vertex_lines_nractivetot,&
             jnubase,anubase)
        do icnt=1,camera_vertex_lines_nractivetot
           camera_cell_line_jnubase(icnt,ray_index) = jnubase(icnt)
           camera_cell_line_anubase(icnt,ray_index) = anubase(icnt)
        enddo
     endif
  enddo
  !
  ! Now average these onto the corner cells
  !
  do ivt=1,amr_nr_vertices
     snu=0.d0
     anu=0.d0
     cnt=0
     if(camera_catch_doppler_line) then
        dpl        = 0.d0
        trb        = 0.d0
        tmp        = 0.d0
        jnubase(:) = 0.d0
        anubase(:) = 0.d0
     endif
     if(.not.amr_tree_present) then
        call amr_regular_get_vertex_ixyz(ivt,ixx,iyy,izz)
     endif
     do iz=1,1+amr_zdim
        do iy=1,1+amr_ydim
           do ix=1,1+amr_xdim
              !
              ! Find the cell connecting to the vertex in this
              ! octant
              !
              if(amr_tree_present) then
                 !
                 ! If there is an AMR tree, then it is easy
                 !
                 if(associated(amr_vertex_cells(ix,iy,iz,ivt)%link)) then
                    ray_index = amr_vertex_cells(ix,iy,iz,ivt)%link%leafindex
                 else
                    ray_index = 0
                 endif
              else
                 !
                 ! If the grid is regular, then we must do some more work
                 !
                 amrray_ix_curr = ixx+ix-2
                 amrray_iy_curr = iyy+iy-2
                 amrray_iz_curr = izz+iz-2
                 !
                 ! Special treatment if phi is cyclic
                 !
                 if(amr_cyclic_xyz(3)) then
                    if(amrray_iz_curr.lt.1) amrray_iz_curr=amr_grid_nz
                    if(amrray_iz_curr.gt.amr_grid_nz) amrray_iz_curr=1
                 endif
                 !
                 ! Special treatment of theta if near the midplane
                 !
                 if(amrray_iy_curr.gt.amr_grid_ny) then
                    if(amrray_mirror_equator) then
                       amrray_iy_curr = amr_grid_ny
                    endif
                 endif
                 !
                 ! For the rest simply check if we are on the grid or not
                 !
                 if((amrray_ix_curr.ge.1).and.(amrray_ix_curr.le.amr_grid_nx).and. &
                    (amrray_iy_curr.ge.1).and.(amrray_iy_curr.le.amr_grid_ny).and. &
                    (amrray_iz_curr.ge.1).and.(amrray_iz_curr.le.amr_grid_nz)) then
                    ray_index = amrray_ix_curr+((amrray_iy_curr-1)+(amrray_iz_curr-1)*amr_grid_ny)*amr_grid_nx
                 else
                    ray_index =  0
                 endif
              endif
              if(ray_index.gt.0) then
                 snu = snu + camera_cell_snu(ray_index)
                 anu = anu + camera_cell_anu(ray_index)
                 if(camera_catch_doppler_line) then
                    dpl = dpl + camera_cell_doppler(ray_index)
                    if(allocated(lines_microturb)) trb = trb + lines_microturb(ray_index)
                    tmp = tmp + gastemp(ray_index)
                    do icnt=1,camera_vertex_lines_nractivetot
                       jnubase(icnt) = jnubase(icnt) + camera_cell_line_jnubase(icnt,ray_index)
                       anubase(icnt) = anubase(icnt) + camera_cell_line_anubase(icnt,ray_index)
                    enddo
                 endif
                 cnt = cnt + 1
              endif
           enddo
        enddo
     enddo
     if(cnt.ge.1) then
        snu = snu / cnt
        anu = anu / cnt
        if(camera_catch_doppler_line) then
           dpl     = dpl / cnt
           trb     = trb / cnt
           tmp     = tmp / cnt
           jnubase = jnubase / cnt
           anubase = anubase / cnt
        endif
     endif
     camera_vertex_snu(ivt) = snu
     camera_vertex_anu(ivt) = anu
     if(camera_catch_doppler_line) then
        camera_vertex_doppler(ivt) = dpl
        camera_vertex_turb(ivt)    = trb
        camera_vertex_temp(ivt)    = tmp
        do icnt=1,camera_vertex_lines_nractivetot
           camera_vertex_line_jnubase(icnt,ivt) = jnubase(icnt)
           camera_vertex_line_anubase(icnt,ivt) = anubase(icnt)
        enddo
     endif
  enddo
  !
  deallocate(src,alp)
  if(allocated(jnubase)) deallocate(jnubase,anubase)
  !
end subroutine camera_compute_snualphanu_at_vertices


!-------------------------------------------------------------------------
!             DO LINEAR INTERPOLATION OF ALPHA AND SOURCE
!
! This subroutine does the bilinear interpolation of the s_nu and alpha_nu
! (snu and anu) at the cell interfaces, based on the values at the cell
! corners (vertices).
!
! NOTE: If camera_interpol_jnu is set, then snu = jnu rather than the
!       source function jnu/anu. 
!-------------------------------------------------------------------------
subroutine camera_find_srcalp_interpol(x,y,z,snu,anu)
  implicit none
  double precision :: x,y,z,snu,anu
  integer :: index,icross,ivt00,ivt01,ivt10,ivt11
  type(amr_branch), pointer :: a
  double precision :: epsx,epsx1,epsy,epsy1,epsz,epsz1
  double precision :: x0,x1,y0,y1,z0,z1
  integer :: ixx,iyy,izz,nnx,nny
  !
  ! Small computation
  !
  nnx = amr_grid_nx + amr_xdim
  nny = amr_grid_ny + amr_ydim
  !
  ! Check which cell to use
  !
  if(amrray_icross.gt.0) then
     index = ray_index
     icross = amrray_icross
  elseif(amrray_icross.lt.0) then
     index = ray_indexnext
     icross = -amrray_icross
  else
     snu = 0.d0
     anu = 1d-99
     return
  endif
  !
  ! Check
  !
  if(index.le.0) then
     write(stdo,*) 'ERROR in srcalp interpol: index le 0'
     stop
  endif
  if(amr_tree_present) then
     if(.not.associated(amr_index_to_leaf(index)%link)) then
        write(stdo,*) 'ERROR in srcalp interpol: link to leaf not associated'
        stop
     endif
  endif
  !
  ! Find the pointer to the cell
  !
  if(amr_tree_present) then
     !
     ! We have an AMR tree, so get the information from the tree
     !
     a => amr_index_to_leaf(index)%link
  else
     call amr_regular_get_ixyz(index,ixx,iyy,izz)
  endif
  !
  ! Check out which interface
  !
  select case(icross)
  case(1)
     !
     ! Left x-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(a%ixyzf(2),2,a%level)
           y1 = amr_finegrid_xi(a%ixyzf(2)+1,2,a%level)
           epsy  = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2091
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(a%ixyzf(3),3,a%level)
           z1 = amr_finegrid_xi(a%ixyzf(3)+1,3,a%level)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = amr_cell_corners(1,1,1,index)
        ivt10 = amr_cell_corners(1,1+amr_ydim,1,index)
        ivt01 = amr_cell_corners(1,1,1+amr_zdim,index)
        ivt11 = amr_cell_corners(1,1+amr_ydim,1+amr_zdim,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(iyy,2,0)
           y1 = amr_finegrid_xi(iyy+1,2,0)
           epsy  = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2091
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(izz,3,0)
           z1 = amr_finegrid_xi(izz+1,3,0)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = ixx+((iyy-1)+(izz-1)*nny)*nnx
        ivt10 = ixx+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
        ivt01 = ixx+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
        ivt11 = ixx+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsz1*(epsy1*camera_vertex_snu(ivt00)+epsy*camera_vertex_snu(ivt10)) + &
              epsz*(epsy1*camera_vertex_snu(ivt01)+epsy*camera_vertex_snu(ivt11))
     anu   = epsz1*(epsy1*camera_vertex_anu(ivt00)+epsy*camera_vertex_anu(ivt10)) + &
              epsz*(epsy1*camera_vertex_anu(ivt01)+epsy*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsz1*(epsy1*camera_vertex_doppler(ivt00)+epsy*camera_vertex_doppler(ivt10)) + &
             epsz *(epsy1*camera_vertex_doppler(ivt01)+epsy*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsz1*(epsy1*camera_vertex_turb(ivt00)+epsy*camera_vertex_turb(ivt10)) + &
             epsz *(epsy1*camera_vertex_turb(ivt01)+epsy*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsz1*(epsy1*camera_vertex_temp(ivt00)+epsy*camera_vertex_temp(ivt10)) + &
             epsz *(epsy1*camera_vertex_temp(ivt01)+epsy*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsz1*(epsy1*camera_vertex_line_jnubase(:,ivt00)+epsy*camera_vertex_line_jnubase(:,ivt10)) + &
             epsz *(epsy1*camera_vertex_line_jnubase(:,ivt01)+epsy*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsz1*(epsy1*camera_vertex_line_anubase(:,ivt00)+epsy*camera_vertex_line_anubase(:,ivt10)) + &
             epsz *(epsy1*camera_vertex_line_anubase(:,ivt01)+epsy*camera_vertex_line_anubase(:,ivt11))
     endif
  case(2)
     !
     ! Right x-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(a%ixyzf(2),2,a%level)
           y1 = amr_finegrid_xi(a%ixyzf(2)+1,2,a%level)
           epsy  = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2091
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(a%ixyzf(3),3,a%level)
           z1 = amr_finegrid_xi(a%ixyzf(3)+1,3,a%level)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = amr_cell_corners(2,1,1,index)
        ivt10 = amr_cell_corners(2,1+amr_ydim,1,index)
        ivt01 = amr_cell_corners(2,1,1+amr_zdim,index)
        ivt11 = amr_cell_corners(2,1+amr_ydim,1+amr_zdim,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(iyy,2,0)
           y1 = amr_finegrid_xi(iyy+1,2,0)
           epsy  = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2091
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(izz,3,0)
           z1 = amr_finegrid_xi(izz+1,3,0)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = ixx+amr_xdim+((iyy-1)+(izz-1)*nny)*nnx
        ivt10 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
        ivt01 = ixx+amr_xdim+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
        ivt11 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsz1*(epsy1*camera_vertex_snu(ivt00)+epsy*camera_vertex_snu(ivt10)) + &
              epsz*(epsy1*camera_vertex_snu(ivt01)+epsy*camera_vertex_snu(ivt11))
     anu   = epsz1*(epsy1*camera_vertex_anu(ivt00)+epsy*camera_vertex_anu(ivt10)) + &
              epsz*(epsy1*camera_vertex_anu(ivt01)+epsy*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsz1*(epsy1*camera_vertex_doppler(ivt00)+epsy*camera_vertex_doppler(ivt10)) + &
             epsz *(epsy1*camera_vertex_doppler(ivt01)+epsy*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsz1*(epsy1*camera_vertex_turb(ivt00)+epsy*camera_vertex_turb(ivt10)) + &
             epsz *(epsy1*camera_vertex_turb(ivt01)+epsy*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsz1*(epsy1*camera_vertex_temp(ivt00)+epsy*camera_vertex_temp(ivt10)) + &
             epsz *(epsy1*camera_vertex_temp(ivt01)+epsy*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsz1*(epsy1*camera_vertex_line_jnubase(:,ivt00)+epsy*camera_vertex_line_jnubase(:,ivt10)) + &
             epsz *(epsy1*camera_vertex_line_jnubase(:,ivt01)+epsy*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsz1*(epsy1*camera_vertex_line_anubase(:,ivt00)+epsy*camera_vertex_line_anubase(:,ivt10)) + &
             epsz *(epsy1*camera_vertex_line_anubase(:,ivt01)+epsy*camera_vertex_line_anubase(:,ivt11))
     endif
  case(3)
     !
     ! Left y-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(a%ixyzf(1),1,a%level)
           x1 = amr_finegrid_xi(a%ixyzf(1)+1,1,a%level)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(a%ixyzf(3),3,a%level)
           z1 = amr_finegrid_xi(a%ixyzf(3)+1,3,a%level)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = amr_cell_corners(1,1,1,index)
        ivt10 = amr_cell_corners(1+amr_xdim,1,1,index)
        ivt01 = amr_cell_corners(1,1,1+amr_zdim,index)
        ivt11 = amr_cell_corners(1+amr_xdim,1,1+amr_zdim,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(ixx,1,0)
           x1 = amr_finegrid_xi(ixx+1,1,0)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(izz,3,0)
           z1 = amr_finegrid_xi(izz+1,3,0)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = ixx+((iyy-1)+(izz-1)*nny)*nnx
        ivt10 = ixx+amr_xdim+((iyy-1)+(izz-1)*nny)*nnx
        ivt01 = ixx+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
        ivt11 = ixx+amr_xdim+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsz1*(epsx1*camera_vertex_snu(ivt00)+epsx*camera_vertex_snu(ivt10)) + &
              epsz*(epsx1*camera_vertex_snu(ivt01)+epsx*camera_vertex_snu(ivt11))
     anu   = epsz1*(epsx1*camera_vertex_anu(ivt00)+epsx*camera_vertex_anu(ivt10)) + &
              epsz*(epsx1*camera_vertex_anu(ivt01)+epsx*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsz1*(epsx1*camera_vertex_doppler(ivt00)+epsx*camera_vertex_doppler(ivt10)) + &
             epsz *(epsx1*camera_vertex_doppler(ivt01)+epsx*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsz1*(epsx1*camera_vertex_turb(ivt00)+epsx*camera_vertex_turb(ivt10)) + &
             epsz *(epsx1*camera_vertex_turb(ivt01)+epsx*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsz1*(epsx1*camera_vertex_temp(ivt00)+epsx*camera_vertex_temp(ivt10)) + &
             epsz *(epsx1*camera_vertex_temp(ivt01)+epsx*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsz1*(epsx1*camera_vertex_line_jnubase(:,ivt00)+epsx*camera_vertex_line_jnubase(:,ivt10)) + &
             epsz *(epsx1*camera_vertex_line_jnubase(:,ivt01)+epsx*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsz1*(epsx1*camera_vertex_line_anubase(:,ivt00)+epsx*camera_vertex_line_anubase(:,ivt10)) + &
             epsz *(epsx1*camera_vertex_line_anubase(:,ivt01)+epsx*camera_vertex_line_anubase(:,ivt11))
     endif
  case(4)
     !
     ! Right y-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(a%ixyzf(1),1,a%level)
           x1 = amr_finegrid_xi(a%ixyzf(1)+1,1,a%level)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(a%ixyzf(3),3,a%level)
           z1 = amr_finegrid_xi(a%ixyzf(3)+1,3,a%level)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = amr_cell_corners(1,2,1,index)
        ivt10 = amr_cell_corners(1+amr_xdim,2,1,index)
        ivt01 = amr_cell_corners(1,2,1+amr_zdim,index)
        ivt11 = amr_cell_corners(1+amr_xdim,2,1+amr_zdim,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(ixx,1,0)
           x1 = amr_finegrid_xi(ixx+1,1,0)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_zdim.eq.1) then
           z0 = amr_finegrid_xi(izz,3,0)
           z1 = amr_finegrid_xi(izz+1,3,0)
           epsz = (z-z0)/(z1-z0)
           epsz1 = 1.d0-epsz
           if((epsz.lt.-1d-6).or.(epsz.gt.1.000001d0)) stop 2092
        else
           epsz  = 0.d0
           epsz1 = 1.d0
        endif
        ivt00 = ixx+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
        ivt10 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
        ivt01 = ixx+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
        ivt11 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsz1*(epsx1*camera_vertex_snu(ivt00)+epsx*camera_vertex_snu(ivt10)) + &
              epsz*(epsx1*camera_vertex_snu(ivt01)+epsx*camera_vertex_snu(ivt11))
     anu   = epsz1*(epsx1*camera_vertex_anu(ivt00)+epsx*camera_vertex_anu(ivt10)) + &
              epsz*(epsx1*camera_vertex_anu(ivt01)+epsx*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsz1*(epsx1*camera_vertex_doppler(ivt00)+epsx*camera_vertex_doppler(ivt10)) + &
             epsz *(epsx1*camera_vertex_doppler(ivt01)+epsx*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsz1*(epsx1*camera_vertex_turb(ivt00)+epsx*camera_vertex_turb(ivt10)) + &
             epsz *(epsx1*camera_vertex_turb(ivt01)+epsx*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsz1*(epsx1*camera_vertex_temp(ivt00)+epsx*camera_vertex_temp(ivt10)) + &
             epsz *(epsx1*camera_vertex_temp(ivt01)+epsx*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsz1*(epsx1*camera_vertex_line_jnubase(:,ivt00)+epsx*camera_vertex_line_jnubase(:,ivt10)) + &
             epsz *(epsx1*camera_vertex_line_jnubase(:,ivt01)+epsx*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsz1*(epsx1*camera_vertex_line_anubase(:,ivt00)+epsx*camera_vertex_line_anubase(:,ivt10)) + &
             epsz *(epsx1*camera_vertex_line_anubase(:,ivt01)+epsx*camera_vertex_line_anubase(:,ivt11))
     endif
  case(5)
     !
     ! Left z-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(a%ixyzf(1),1,a%level)
           x1 = amr_finegrid_xi(a%ixyzf(1)+1,1,a%level)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(a%ixyzf(2),2,a%level)
           y1 = amr_finegrid_xi(a%ixyzf(2)+1,2,a%level)
           epsy = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2092
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        ivt00 = amr_cell_corners(1,1,1,index)
        ivt10 = amr_cell_corners(1+amr_xdim,1,1,index)
        ivt01 = amr_cell_corners(1,1+amr_ydim,1,index)
        ivt11 = amr_cell_corners(1+amr_xdim,1+amr_ydim,1,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(ixx,1,0)
           x1 = amr_finegrid_xi(ixx+1,1,0)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(iyy,2,0)
           y1 = amr_finegrid_xi(iyy+1,2,0)
           epsy = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2092
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        ivt00 = ixx+((iyy-1)+(izz-1)*nny)*nnx
        ivt10 = ixx+amr_xdim+((iyy-1)+(izz-1)*nny)*nnx
        ivt01 = ixx+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
        ivt11 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsy1*(epsx1*camera_vertex_snu(ivt00)+epsx*camera_vertex_snu(ivt10)) + &
              epsy*(epsx1*camera_vertex_snu(ivt01)+epsx*camera_vertex_snu(ivt11))
     anu   = epsy1*(epsx1*camera_vertex_anu(ivt00)+epsx*camera_vertex_anu(ivt10)) + &
              epsy*(epsx1*camera_vertex_anu(ivt01)+epsx*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsy1*(epsx1*camera_vertex_doppler(ivt00)+epsx*camera_vertex_doppler(ivt10)) + &
             epsy *(epsx1*camera_vertex_doppler(ivt01)+epsx*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsy1*(epsx1*camera_vertex_turb(ivt00)+epsx*camera_vertex_turb(ivt10)) + &
             epsy *(epsx1*camera_vertex_turb(ivt01)+epsx*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsy1*(epsx1*camera_vertex_temp(ivt00)+epsx*camera_vertex_temp(ivt10)) + &
             epsy *(epsx1*camera_vertex_temp(ivt01)+epsx*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsy1*(epsx1*camera_vertex_line_jnubase(:,ivt00)+epsx*camera_vertex_line_jnubase(:,ivt10)) + &
             epsy *(epsx1*camera_vertex_line_jnubase(:,ivt01)+epsx*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsy1*(epsx1*camera_vertex_line_anubase(:,ivt00)+epsx*camera_vertex_line_anubase(:,ivt10)) + &
             epsy *(epsx1*camera_vertex_line_anubase(:,ivt01)+epsx*camera_vertex_line_anubase(:,ivt11))
     endif
  case(6)
     !
     ! Right z-wall...
     !
     ! Find the vertex indices and the linear interpolation epsilons
     !
     if(amr_tree_present) then
        !
        ! The AMR tree is available, so get the information from there
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(a%ixyzf(1),1,a%level)
           x1 = amr_finegrid_xi(a%ixyzf(1)+1,1,a%level)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(a%ixyzf(2),2,a%level)
           y1 = amr_finegrid_xi(a%ixyzf(2)+1,2,a%level)
           epsy = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2092
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        ivt00 = amr_cell_corners(1,1,2,index)
        ivt10 = amr_cell_corners(1+amr_xdim,1,2,index)
        ivt01 = amr_cell_corners(1,1+amr_ydim,2,index)
        ivt11 = amr_cell_corners(1+amr_xdim,1+amr_ydim,2,index)
     else
        !
        ! Regular grid, so construct the information from ixx,iyy,izz
        !
        if(amr_xdim.eq.1) then
           x0 = amr_finegrid_xi(ixx,1,0)
           x1 = amr_finegrid_xi(ixx+1,1,0)
           epsx  = (x-x0)/(x1-x0)
           epsx1 = 1.d0-epsx
           if((epsx.lt.-1d-6).or.(epsx.gt.1.000001d0)) stop 2091
        else
           epsx  = 0.d0
           epsx1 = 1.d0
        endif
        if(amr_ydim.eq.1) then
           y0 = amr_finegrid_xi(iyy,2,0)
           y1 = amr_finegrid_xi(iyy+1,2,0)
           epsy = (y-y0)/(y1-y0)
           epsy1 = 1.d0-epsy
           if((epsy.lt.-1d-6).or.(epsy.gt.1.000001d0)) stop 2092
        else
           epsy  = 0.d0
           epsy1 = 1.d0
        endif
        ivt00 = ixx+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
        ivt10 = ixx+amr_xdim+((iyy-1)+(izz-1+amr_zdim)*nny)*nnx
        ivt01 = ixx+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
        ivt11 = ixx+amr_xdim+((iyy-1+amr_ydim)+(izz-1+amr_zdim)*nny)*nnx
     endif
     !
     ! Do the interpolations
     !
     snu   = epsy1*(epsx1*camera_vertex_snu(ivt00)+epsx*camera_vertex_snu(ivt10)) + &
              epsy*(epsx1*camera_vertex_snu(ivt01)+epsx*camera_vertex_snu(ivt11))
     anu   = epsy1*(epsx1*camera_vertex_anu(ivt00)+epsx*camera_vertex_anu(ivt10)) + &
              epsy*(epsx1*camera_vertex_anu(ivt01)+epsx*camera_vertex_anu(ivt11))
     if(camera_catch_doppler_line) then
        camera_local_doppler_curr = &
             epsy1*(epsx1*camera_vertex_doppler(ivt00)+epsx*camera_vertex_doppler(ivt10)) + &
             epsy *(epsx1*camera_vertex_doppler(ivt01)+epsx*camera_vertex_doppler(ivt11))
        camera_local_turb_curr = &
             epsy1*(epsx1*camera_vertex_turb(ivt00)+epsx*camera_vertex_turb(ivt10)) + &
             epsy *(epsx1*camera_vertex_turb(ivt01)+epsx*camera_vertex_turb(ivt11))
        camera_local_temp_curr = &
             epsy1*(epsx1*camera_vertex_temp(ivt00)+epsx*camera_vertex_temp(ivt10)) + &
             epsy *(epsx1*camera_vertex_temp(ivt01)+epsx*camera_vertex_temp(ivt11))
        camera_local_line_jnubase_curr(:) = &
             epsy1*(epsx1*camera_vertex_line_jnubase(:,ivt00)+epsx*camera_vertex_line_jnubase(:,ivt10)) + &
             epsy *(epsx1*camera_vertex_line_jnubase(:,ivt01)+epsx*camera_vertex_line_jnubase(:,ivt11))
        camera_local_line_anubase_curr(:) = &
             epsy1*(epsx1*camera_vertex_line_anubase(:,ivt00)+epsx*camera_vertex_line_anubase(:,ivt10)) + &
             epsy *(epsx1*camera_vertex_line_anubase(:,ivt01)+epsx*camera_vertex_line_anubase(:,ivt11))
     endif
  end select
end subroutine camera_find_srcalp_interpol


!==========================================================================
!             SOME BACKWARD COMPATIBILITY ROUTINES WITH RADMC
!==========================================================================

!--------------------------------------------------------------------------
!                  SET UP PIXEL POSITIONS FOR CIRCULAR IMAGES
!
! Since we now have better methods than the circular images, we do not
! really need these. But since RADMC uses them, and since we want to 
! be able to reproduce old RADMC results with RADMC-3D, it might be useful.
! Note that it only works for spherical coordinates.
!--------------------------------------------------------------------------
subroutine setup_pixels_circular(irmin,irmax,nrphiinf,nrext,dbdr,imethod,nrref)
  use amr_module
  implicit none
  integer :: irmin,irmax,nrphiinf,nrext,dbdr
  integer :: imethod,nrref
  integer :: ix,ir,iins,i
  integer :: iradius,nrrextra
  doubleprecision :: epsrrr,refdum1,refdum2,telesc_eps,dr
  parameter(telesc_eps=1d-10)
  parameter(epsrrr=1.0d3*telesc_eps)
  !
  ! Check if we are indeed in spherical coordinates
  !
  if(amrray_icoord.ne.1) then 
     write(stdo,*) 'ERROR: You can only use the circular images if you'
     write(stdo,*) '       use spherical coordinates.'
     stop
  endif
  if(.not.allocated(star_r)) then
     write(stdo,*) 'ERROR in circular image generator: For circular images at '
     write(stdo,*) '      a central star is required (even if its luminosity is 0).'
     stop
  endif
  if((star_pos(1,1).ne.0.d0).or.(star_pos(2,1).ne.0.d0).or.(star_pos(3,1).ne.0.d0)) then
     write(stdo,*) 'ERROR in circular image generator: For circular images the '
     write(stdo,*) '      first star must be the central star (i.e. pos=0,0,0).'
     stop
  endif
  !
  ! Default
  !
  nrrextra = nrext
  if(nrrextra.lt.0) nrrextra = -nrrextra
  !
  ! Decide on which way the radial rays are arranged
  !     
  if(imethod.eq.0) then
     !
     ! The original method from RADICAL
     !
     if(nrext.gt.0) then
        nrrextra = nrext
        imethod = -1
     elseif(nrext.lt.0) then
        nrrextra = -nrext
        imethod = -2
     else
        write(*,*) 'ERROR: Must have non-zero nrrextra'
        stop
     endif
  endif
  !
  ! Allocate arrays
  !
  cim_nr  = (irmax-irmin+1)*dbdr+nrrextra+nrref
  cim_np  = nrphiinf
  if(allocated(cim_rc)) deallocate(cim_rc)
  if(allocated(cim_ri)) deallocate(cim_ri)
  if(allocated(cim_pc)) deallocate(cim_pc)
  if(allocated(cim_pi)) deallocate(cim_pi)
  allocate(cim_rc(0:cim_nr),cim_ri(0:cim_nr+1))  ! Note: cell 0 = star
  allocate(cim_pc(1:cim_np),cim_pi(1:cim_np+1))
  !
  ! Do some checks
  !
  if(irmax.le.irmin) then
     write(stdo,*) 'ERROR Telescope: irmax.le.irmin'
     stop
  endif
  if((nrrextra.gt.0).and.(irmin.ne.1)) then
     write(stdo,*) 'ERROR: In setup_pixels_circular():'
     write(stdo,*) '       In You request extra rays at the ', &
          'inner edge while you dont have irmin=1.'
     write(stdo,*) '       This is contradictory. Aborting...'
     stop
  endif
  if(nstars.ne.1) then
     write(stdo,*) 'ERROR: In circular images: must have nstar==1.'
     stop
  endif
  !
  ! Now let us make the set of rays for the image at
  ! infinity... This consists of a series of concentric rings
  ! of pixels around the center of the object. Each pixel of 
  ! the image corresponds to one ray defined here. This
  ! circular image has coordinates r (in [cm], is the radius
  ! of each ringin the circular image) and phi (in [rad], is
  ! the angle along each concentric ring). 
  !
  ! First make the phi grid, which is the easiest
  !
  do i=1,cim_np
     cim_pc(i) = twopi * (i-0.5d0)/cim_np
  enddo
  do i=1,cim_np+1
     cim_pc(i) = twopi * (i-1.d0)/cim_np
  enddo
  !
  ! Now make the r-grid, which must be adapted to the r-grid of the
  ! model itself. NOTE: It will only adapt itself to the REGULAR radial
  ! grid, not to the AMR grid. 
  !
  ! The 0th pixel is the star
  !
  ir              = 1
  iradius         = 0
  cim_ri(0)       = 0.d0
  cim_rc(0)       = 0.d0
  cim_ri(1)       = star_r(1)
  !     
  ! Next make the extra rays, which are the rays with impact parametes
  ! smaller than the radius of the inner edge of the (R,Theta) spatial
  ! grid of the object itself. 
  !
  ir              = 2
  iradius         = 1
  !
  ! These are the original RADICAL-style ways to arrange these
  ! rays. But be careful that there are drawbacks for near-face-on
  ! spectra for very optically thick disks with vertical inner walls.
  !
  if(imethod.eq.-2) then 
     !
     ! The old RADICAL way
     !      
     do ix=1,nrrextra 
        if(iradius.gt.cim_nr) then
           write(stdo,*) 'EROR in circular images: exceeded cim_nr'
           stop
        endif
        if(amr_grid_xi(1,1).le.star_r(1)) then
           write(stdo,*) 'ERROR in circular images: Grid inner edge is not'
           write(stdo,*) '      larger than stellar radius.'
           stop
        endif
        cim_rc(iradius) = ( ix * ( amr_grid_xi(1,1) - &
                          star_r(1) ) / (nrrextra+1.d0) ) + star_r(1)
        iradius = iradius + 1
     enddo
  elseif(imethod.eq.1) then
     !
     ! The newer (RAYTRACE) methods for arranging the rays.
     ! The idea here is to make sure that the inner edge is better
     ! sampled. Also, at some point, I want to introduce here the
     ! method that allows the star to be sampled in a better way,
     ! in case the inner edge of the disk/envelope is not so far 
     ! away from the stellar surface.
     !
     if(nrref.le.0) then
        write(stdo,*) 'ERROR: If new method for ray-arrangement'
        write(stdo,*) '  is chosen, then nrref must be set>0, too'
        stop
     endif
     refdum1 = 0.5d0
     refdum2 = 0.d0
     do ix=1,nrrextra+nrref 
        if(iradius.gt.cim_nr) then
           write(stdo,*) 'BUG in setup_pixels_circular()'
           write(stdo,*) '    iradius exceeds max.'
           stop 155
        endif
        !
        ! A method by which a minimal grid fineness is guaranteed,
        ! but also refinement near the inner edge is done.
        ! 
        if(amr_grid_xi(1,1).le.star_r(1)) then
           write(stdo,*) 'ERROR in circular images: Grid inner edge is not'
           write(stdo,*) '      larger than stellar radius.'
           stop
        endif
        if(ix.le.nrrextra) then
           cim_rc(iradius) = ( ix * ( amr_grid_xi(1,1) - &
                             star_r(1) ) / (nrrextra+1.d0) ) + star_r(1)
        else
           refdum2 = refdum2 + refdum1
           refdum1 = refdum1/2
           cim_rc(iradius) = ( ( nrrextra + refdum2 )            &
                             * ( amr_grid_xi(1,1) - star_r(1) ) /    &
                             (nrrextra+1.d0) ) + star_r(1)
        endif
        iradius = iradius + 1
     enddo
  else
     stop 33720
  endif
  !
  ! Now make the usual rays, which are the rays with impact parameters
  ! larger than the inner edge, but smaller than the outer edge of the
  ! spatial object. These rays will contain most of the information of 
  ! the image, once the telescope imaging routine has done it's job.
  !
  irmax = max(irmax,amr_grid_nx)
  do ix=irmin,irmax-1
     if(iradius.gt.cim_nr) then
        write(stdo,*) 'BUG setup_pixels_circular(): exceed limits'
        stop 155
     endif
     cim_rc(iradius) = amr_grid_xi(ix,1) * (1.d0 + epsrrr )
     iradius = iradius + 1
     !
     ! If I am not at the last point, and if the dbdr
     ! gt 1 then I will add a few points in between this
     ! R_i and R_i+1.... And I will simply count the amount
     ! of added points....
     !
     if((ix.lt.irmax).and.(dbdr.gt.1)) then
        dr = ( amr_grid_xi(ix+1,1) - amr_grid_xi(ix,1) ) / &
             ( 1.d0*dbdr )
        do iins = 1,dbdr-1
           if(iradius.gt.cim_nr) then
              write(stdo,*) 'BUG in circular images: exceeded limit.'
              stop 155
           endif
           cim_rc(iradius) = amr_grid_xi(ix,1) + iins*dr
           iradius = iradius + 1
        enddo
     endif
  enddo
  if(iradius-1.gt.cim_nr) stop 4
  cim_nr  = iradius - 1
  !
  ! Compute the pixel cell walls 
  !
  do ix=2,cim_nr
     cim_ri(ix) = 0.5d0*(cim_rc(ix)+cim_rc(ix-1))
  enddo
  if(cim_ri(1).ge.cim_rc(1)) stop 4091
  cim_ri(cim_nr+1) = 2*cim_ri(cim_nr) - cim_ri(cim_nr-1) 
  !
  ! Set some auxiliary stuff
  !
  camera_image_halfsize_x = cim_ri(cim_nr+1)
  camera_image_halfsize_y = cim_ri(cim_nr+1)
  camera_image_nx = 0
  camera_image_ny = 0
  !
end subroutine setup_pixels_circular

!-------------------------------------------------------------------------
!                         MAKE A CIRCULAR IMAGE
!                  (BACKWARD COMPATIBILITY WITH RADMC)
!-------------------------------------------------------------------------
subroutine camera_make_circ_image()
  use amr_module
  implicit none
  double precision :: x,y,z,dxx,dyy,dzz,px,py
  integer :: inu,ix,iy,idxx,idxy,istar,ierr,inu0,inu1,ierror
  double precision :: pdx,pdy,dz,d2,d3,factor,celldxmin
  double precision :: dirx,diry,dirz,distance,xbk,ybk,zbk
  logical :: domc
  !
  ! Some checks
  !
  if(camera_localobserver) then
     write(stdo,*) 'ERROR: In circular images no local observer perspective allowed'
     stop
  endif
  if(.not.allocated(cim_ri)) then
     write(stdo,*) 'ERROR in circular image generator: please set up pixels first.'
     stop
  endif
  !
  ! Do a self-consistency test: for spherical coordinates with 
  ! a symmetry (i.e. 2-D axial symmetry or 1-D spherical symmetry)
  ! anisotropic scattering cannot be included. This is because for
  ! e.g. 2-D axial symmetry each cell is in fact a "ring" around the
  ! symmetry axis. The various locations along this ring have different
  ! angle toward the observer (unless the observer looks pole-on). 
  ! Only with isotropic scattering this problem does not show up.
  ! One solution is to convert a 2-D model into a 3-D model before
  ! making the images. 
  !
  if(igrid_type.lt.100) then
     if((igrid_coord.ge.100).and.(amr_zdim.eq.0).and.  &
          (scattering_mode.ge.2)) then
        write(stdo,*) 'ERROR: Anisotropic scattering only allowed in 3-D models,'
        write(stdo,*) '       or in 2-D models that are converted into 3-D before'
        write(stdo,*) '       the images or spectra are made.'
        stop
     endif
  endif
  !
  ! Let the Monte Carlo module know what mode we have (local observer
  ! or not)
  !
  mcscat_localobserver = .false.
  camera_localobserver = .false.
  !
  ! If the dust emission is included, then make sure the dust data,
  ! density and temperature are read. If yes, do not read again.
  !
  if(rt_incl_dust) then
     call read_dustdata(1)
     call read_dust_density(1)
     call read_dust_temperature(1)
  endif
  !
  ! If line emission is included, then make sure the line data are
  ! read. If yes, then do not read it again.
  !
  if(rt_incl_lines) then
     call read_lines_all(1)
  endif
  !
  ! If gas continuum is included, then make sure the gas continuum
  ! data are read. If yes, then do not read it again.
  !
  if(rt_incl_gascont) then
     call gascont_init(1)
  endif
  !
  ! If lines are active, and if the level populations are to be calculated
  ! beforehand and stored in the big array, then compute them now, if not
  ! already done.
  !
  if(rt_incl_lines) then
     if((lines_mode.ge.1).and.(lines_mode.le.9)) then
        call lines_compute_and_store_local_populations(2)
     endif
  endif
  !
  ! Do the initializing of the camera
  !
  call camera_init()
  !
  ! Check
  !
  if(camera_nrfreq.lt.1) then
     write(stdo,*) 'ERROR in camera module: camera frequency array'
     write(stdo,*) '      not set when calling camera_make_rect_image()'
     stop
  endif
  !
  ! Check
  !
  if((camera_image_halfsize_x.le.0.d0).or.(camera_image_halfsize_y.le.0.d0)) then
     write(stdo,*) 'ERROR in camera module: zero image scale'
     stop
  endif
  !
  ! For the camera-at-infinity view the angles are given and the
  ! cos and sin can be calculated easily
  !
  camera_pointing_cos_posang  = cos(camera_pointing_degr_posang*pi/180.)
  camera_pointing_sin_posang  = sin(camera_pointing_degr_posang*pi/180.)
  camera_observer_cos_theta   = cos(camera_observer_degr_theta*pi/180.)
  camera_observer_sin_theta   = sin(camera_observer_degr_theta*pi/180.)
  camera_observer_cos_phi     = cos(camera_observer_degr_phi*pi/180.)
  camera_observer_sin_phi     = sin(camera_observer_degr_phi*pi/180.)
  !
  ! For now we do not allow multiple vantage points
  !
  if(mcscat_nrdirs.gt.1) then
     write(stdo,*) 'ERROR: Multiple vantage points is not yet allowed in this version!'
     stop
  endif
  !
  ! Check if we need to (re)do the one-frequency scattering Monte Carlo simulation
  !
  domc = .false.
  if(scattering_mode.ge.1) then
     !
     ! Compute the direction vector of the observer. Start with (0,0,1).
     ! Actually this is only necessary for scattering_mode.gt.1, but alas.
     !
     dirx = 0.d0
     diry = 0.d0
     dirz = 1.d0
     ybk  = diry
     zbk  = dirz
     diry = camera_observer_cos_theta * ybk - camera_observer_sin_theta * zbk
     dirz = camera_observer_sin_theta * ybk + camera_observer_cos_theta * zbk
     xbk  = dirx
     ybk  = diry
     dirx = camera_observer_cos_phi * xbk + camera_observer_sin_phi * ybk
     diry =-camera_observer_sin_phi * xbk + camera_observer_cos_phi * ybk
     if(abs(dirx**2+diry**2+dirz**2-1.d0).gt.1d-6) then
        write(stdo,*) 'ERROR in camera module: direction vector not OK.'
        write(stdo,*) dirx,diry,dirz
        stop
     endif
     !
     ! Some simple tests to see if a new MC calculation is necessary
     !
     if((.not.allocated(mcscat_dirs)).and.(scattering_mode.gt.1)) domc=.true.
     if(allocated(mcscat_dirs).and.(scattering_mode.eq.1)) then
        write(stdo,*) 'ERROR: if isotropic scattering is set, then do not set the mcscat_dirs(:,:) array...'
        stop
     endif
     if(mcscat_nrdirs.ne.1) domc=.true.
     if(.not.camera_scatsrc_allfreq) domc=.true.
     !
     ! Some more subtle tests to see if a new MC calculation is necessary
     !
     if(camera_scatsrc_allfreq) then
        if(mc_nrfreq.ne.camera_nrfreq) then
           domc=.true.
        else
           do inu=1,camera_nrfreq
              if(abs((camera_frequencies(inu)-mc_frequencies(inu))/   &
                     (camera_frequencies(inu)+mc_frequencies(inu))).gt.1d-10) domc=.true.
           enddo
        endif
        if(allocated(mcscat_dirs)) then
           mcscat_current_dir = 1
           if((abs(mcscat_dirs(1,mcscat_current_dir)-dirx).gt.1d-6).or.   &
              (abs(mcscat_dirs(2,mcscat_current_dir)-diry).gt.1d-6).or.   &
              (abs(mcscat_dirs(3,mcscat_current_dir)-dirz).gt.1d-6)) then
              domc=.true.
           endif
        endif
     endif
  endif
  !
  ! Prepare the direction for the scattering source function
  ! For now allow only one single direction (i.e. one single vantage point)
  !
  if(domc.and.(scattering_mode.gt.1)) then
     if(allocated(mcscat_dirs)) deallocate(mcscat_dirs)
     allocate(mcscat_dirs(1:3,1:1),STAT=ierr)
     if(ierr.gt.0) then
        write(stdo,*) 'ERROR: Could not allocate mcscat_dirs()'
        stop
     endif
     mcscat_nrdirs = 1
     mcscat_current_dir = 1
     mcscat_dirs(1,1:1) = dirx
     mcscat_dirs(2,1:1) = diry
     mcscat_dirs(3,1:1) = dirz
  endif
  !
  ! Compute pixel size.
  !
  ! NOTE: For the observer-at-infinity mode this is in centimeters.
  !       For the local-observer mode this is in radian.
  !
  pdx = 2*camera_image_halfsize_x / (1.d0*camera_image_nx)
  pdy = 2*camera_image_halfsize_y / (1.d0*camera_image_ny)
  !
  ! Reset flag
  !
  camera_warn_resolution = .false.
  !
  ! Message
  !
  write(stdo,*) 'Rendering image(s)...'
  !
  ! Now make the image. We can do so in two different ways. The first is to
  ! do all frequencies (wavelengths) at once. If scattering is included this
  ! requires the precomputing of the scattering source function at all
  ! wavelengths.  This can have a too large memory requirement.  The second
  ! is to do it frequency-by-frequency, and doing a MC calculation for each
  ! of them separately. That saves a lot of computer memory, so it is the
  ! preferred method for large simulations when making spectra. But it may
  ! be slower. The way to choose one or the other is with the switch called
  ! camera_scatsrc_allfreq.
  !
  if(((scattering_mode.eq.0).or.camera_scatsrc_allfreq).and.   &
      (.not.camera_secondorder)) then
     !
     ! Multi-wavelength Method 1:
     !
     ! Do all frequencies at once for each pixel. This is the preferred
     ! method for cases without scattering. With scattering this requires a
     ! lot of memory to store the scattering source function for all
     ! frequencies, so it might not be the best for that.
     !
     ! Note that at present this multi-frequency mode is not available
     ! for second order integration of the transfer equation (i.e. with
     ! the corner points). 
     !
     inu0 = 1
     inu1 = camera_nrfreq
     !
     ! Tell the ray tracer that if scattering is done, then it is done
     ! frequency dispersed
     !
     camera_mcscat_monochromatic = .false.
     !
     ! If necessary, then do the scattering source functions at all
     ! frequencies beforehand.  WARNING: This can be a very large array!
     !
     if(domc) then
        if(allocated(mc_frequencies)) deallocate(mc_frequencies)
        mc_nrfreq=camera_nrfreq
        allocate(mc_frequencies(1:mc_nrfreq),STAT=ierr)
        if(ierr.ne.0) then
           write(stdo,*) 'ERROR: Could not allocate mc_frequencies(:) array'
           stop
        endif
        mc_frequencies(:) = camera_frequencies(:)
        !
        ! Message
        !
        write(stdo,*) 'Doing scattering Monte Carlo simulation...'
        call flush(stdo)
        !
        ! Do Monte Carlo simulation
        !
        call do_monte_carlo_scattering(rt_mcparams,ierror,do_resetseed,&
                                       scatsrc=.true.)
        write(stdo,*) 'Average number of scattering events per photon package = ', &
                      ieventcounttot/(1.d0*rt_mcparams%nphot_scat)
     endif
     !
     ! Pre-compute which lines and which levels for line transfer may
     ! contribute to these wavelengths. Note that this only has to be
     ! pre-computed if the serial ray tracing is used, but we do it 
     ! nevertheless, as it will not hurt.
     !
     if(rt_incl_lines) then
     ! ------------------------------------------------------------------
     ! Attila Juhasz
     ! lines_find_active_lines_leves -> if all energy levels are known - leiden
     ! format for line data
     ! lines_find_active_lines_linelist -> for linelist mode
     ! ------------------------------------------------------------------
     !   call lines_find_active_lines_levels(camera_nrfreq,             &
     !                   camera_frequencies,inu0,inu1,lines_maxshift)
     !
        if(lines_maxnrlevels.gt.0) then
           call lines_find_active_lines_levels(camera_nrfreq,             &
                camera_frequencies,inu0,inu1,lines_maxshift)
        else
           call lines_find_active_lines_linelist(camera_nrfreq,             &
                camera_frequencies,inu0,inu1,lines_maxshift)
        endif

     ! ------------------------------------------------------------------
     endif
     !
     ! Now make the image at all wavelengths simultaneously
     !
     call camera_make_circ_image_sub(inu0,inu1)
     !
  else
     !
     ! Multi-wavelength Method 2:
     !
     ! Scattering is included, so we need the scattering source function. Since this 
     ! array is easily too big to be stored for all wavelengths. In this method we
     ! compute the scattering source function for each frequency separately, and then
     ! make the corresponding image, and then go to the next frequency.
     !
     ! Tell the ray tracer that if scattering is done, then it is done
     ! one frequency at a time, i.e. the scattering source function is
     ! just one frequency bin.
     !
     camera_mcscat_monochromatic = .true.
     !
     ! Since we recompute the scattering source function for each wavelength, the domc
     ! is by default true. Note that if you make an image at a single wavelength you 
     ! may want to be able to make use of a previously computed source function, if for
     ! instance you make your next image at the same vantage point, the same wavelength
     ! but at a different zoom factor or so. In that case, please set the flag
     ! camera_scatsrc_allfreq to .true.. 
     !
     if(domc) then
        if(allocated(mc_frequencies)) deallocate(mc_frequencies)
        mc_nrfreq=1
        allocate(mc_frequencies(1:1),STAT=ierr)
     endif
     !
     ! Do a loop over all camera frequencies
     !
     do inu0=1,camera_nrfreq
        inu1=inu0
        !
        ! Set the wavelength for the Monte Carlo scattering simulation
        !
        mc_frequencies(1) = camera_frequencies(inu0)
        !
        ! Message
        !
        write(stdo,*) 'Doing scattering Monte Carlo simulation for lambda = ', &
                1d4*cc/mc_frequencies(1),' micron...'
        call flush(stdo)
        !
        ! Call the single wavelength Monte Carlo module
        !
        call do_monte_carlo_scattering(rt_mcparams,ierror,do_resetseed,&
                                       scatsrc=.true.)
        write(stdo,*) 'Average number of scattering events per photon package = ', &
                      ieventcounttot/(1.d0*rt_mcparams%nphot_scat)
        !
        ! Pre-compute which lines and which levels for line transfer may
        ! contribute to these wavelengths. Note that this only has to be
        ! pre-computed if the serial ray tracing is used, but we do it 
        ! nevertheless, as it will not hurt.
        !
        if(rt_incl_lines) then
           ! ------------------------------------------------------------------
           ! Attila Juhasz
           ! lines_find_active_lines_leves -> if all energy levels are known - leiden
           ! format for line data
           ! lines_find_active_lines_linelist -> for linelist mode
           ! ------------------------------------------------------------------
           !   call lines_find_active_lines_levels(camera_nrfreq,             &
           !                   camera_frequencies,inu0,inu1,lines_maxshift)
           !
            if(lines_maxnrlevels.gt.0) then
                call lines_find_active_lines_levels(camera_nrfreq,             &
                     camera_frequencies,inu0,inu1,lines_maxshift)
            else
                call lines_find_active_lines_linelist(camera_nrfreq,             &
                     camera_frequencies,inu0,inu1,lines_maxshift)
            endif

           ! ------------------------------------------------------------------

        endif
        !
        ! Now make the image for this wavelength only
        !
        write(stdo,*) 'Making image for lambda = ', &
                1d4*cc/mc_frequencies(1),' micron...'
        call camera_make_circ_image_sub(inu0,inu1)
        !
     enddo
     !
  endif
  !
  ! Check if refinement was sufficient
  !
  if(camera_warn_resolution) then
     write(stdo,*) 'WARNING: The currently produced image is not '
     write(stdo,*) '         guaranteed to have the correct flux '
     write(stdo,*) '         because it does not use the flux'
     write(stdo,*) '         conserving recursive pixeling.'
     write(stdo,*) '    Tip: Use fluxcons argument in command line.'
  endif
  !
  !--------------------------------------------------------------
  !        A sub-subroutine for making the image
  !--------------------------------------------------------------
  !
  contains
  subroutine camera_make_circ_image_sub(inu00,inu11)
    implicit none
    integer :: inu00,inu11
    !
    ! Reset some non-essential counters
    !
    camera_subpixeling_npixfine = 0
    camera_subpixeling_npixtot  = 0
    !
    ! Compute the intensity of all pixels
    !
    do iy=1,camera_image_ny
       do ix=1,camera_image_nx
          !
          ! Set the ray variables
          !
          px = camera_zoomcenter_x + ((ix-1)-(0.5d0*(camera_image_nx-1)))*pdx
          py = camera_zoomcenter_y + ((iy-1)-(0.5d0*(camera_image_ny-1)))*pdy
          !
          ! Now compute the intensity of this pixel
          !
          ! Note that if nrrefine>0 this routine will call itself
          ! recursively until a desired spatial resolution is acquired
          ! so as to guarantee that all flux is picked up, but this
          ! recursion is limited to nrrefine depth levels.
          !
          call camera_compute_one_pixel(camera_nrfreq,inu00,inu11,px,py,pdx,pdy,    &
                                        camera_nrrefine,camera_intensity,-1)
          !
          ! Put the result into the image array
          !
          camera_rect_image(ix,iy,inu00:inu11) = camera_intensity(inu00:inu11)
          !
       enddo
    enddo
    !
    ! Add the discrete star sources. 
    !
    ! NOTE: Only for the mode in which stars are treated as point-sources.
    !
    if((camera_incl_stars.ne.0).and.(.not.star_sphere)) then
       !
       ! Do a check
       !
       if(.not.allocated(star_spec)) then
          write(stdo,*) 'WARNING in camera module: Stars not allocated.'
!          stop
       endif
       !
       ! Now add all stars
       !
       do istar=1,nstars
          !
          ! Set the starting position and direction of the ray, as well as
          ! the location in the image
          !
          call camera_set_ray_stars_pntsrc(istar,x,y,z,dirx,diry,dirz,px,py,distance)
          !
          ! Check if this star is located in the image
          !
          if((px.ge.camera_zoomcenter_x-camera_image_halfsize_x).and.  &
             (px.le.camera_zoomcenter_x+camera_image_halfsize_x).and.  &
             (py.ge.camera_zoomcenter_y-camera_image_halfsize_y).and.  &
             (py.le.camera_zoomcenter_y+camera_image_halfsize_y)) then
             !
             ! Find the indices of the pixel in which the star is
             !
             idxx = floor((px-(camera_zoomcenter_x-camera_image_halfsize_x))/pdx)+1
             idxy = floor((py-(camera_zoomcenter_y-camera_image_halfsize_y))/pdy)+1
             if((idxx.lt.1).or.(idxx.gt.camera_image_nx).or.&
                (idxy.lt.1).or.(idxy.gt.camera_image_ny)) stop 8309
             !
             ! Set the intensity to the stellar spectrum
             ! 
             do inu=inu00,inu11
                camera_intensity(inu) =                                  &
                     find_starlight_interpol(camera_frequencies(inu),istar)
             enddo
             !
             ! Do ray tracing
             !
             call camera_serial_raytrace(camera_nrfreq,inu00,inu11,x,y,z,dirx,diry,dirz,distance,  &
                                  celldxmin,camera_intensity)
             !
             ! Compute the ratio starsurface / pixelsurface
             !
             if(camera_localobserver) then
                factor = pi*(star_r(istar)/distance)**2 / (pdx*pdy)
             else
                factor = pi*star_r(istar)**2 / (pdx*pdy)
             endif
             !
             ! Check if the point source assumption is violated. 
             !
!             if(factor.gt.1.d0) then
!                write(stdo,*) 'ERROR in camera module: The image resolution is so'
!                write(stdo,*) '      high that the stellar surface is no longer'
!                write(stdo,*) '      small enough to be considered a point source.'
!                write(stdo,*) '  NOTE: In the (hopefully near) future a mode will'
!                write(stdo,*) '        be implemented to include non-point stars.'
!                stop
!             endif
             !
             ! Now modify the intensity of the pixel to include the star
             !
             camera_rect_image(idxx,idxy,inu00:inu11) =                           &
                   (1.d0-factor) * camera_rect_image(idxx,idxy,inu00:inu11) +     &  
                   factor * camera_intensity(inu00:inu11)
          endif
       enddo
    endif
  end subroutine camera_make_circ_image_sub
  !
end subroutine camera_make_circ_image


!-------------------------------------------------------------------
!           COMPUTE AND STORE POPULATIONS IN GLOBAL ARRAY
!-------------------------------------------------------------------
subroutine lines_compute_and_store_local_populations(action)
  implicit none
  integer :: action,index,ierr,ilevsub,ilevel,ispec
  type(amr_branch), pointer :: a
  double precision, allocatable :: pop(:,:)
  double precision :: velgrad,xav,yav,zav
  integer :: ixx,iyy,izz
  !
  ! Action=0 means do nothing, action=1 means compute if not yet computed,
  ! action=2 means re-compute.
  !
  if(action.eq.0) then
     return
  elseif(action.eq.1) then
     if(allocated(lines_levelpop)) return
  endif
  !
  ! Check
  !
  if(lines_nrlevels_subset_max.le.0) then
!     write(stdo,*) 'ERROR: lines_nrlevels_subset_max is zero...'
!     write(stdo,*) '       Cannot allocate level population array...'
!     stop
     return
  endif
  !
  ! Allocate local population array
  !
  allocate(pop(1:lines_maxnrlevels,1:lines_nr_species),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate local pop() array'
     stop
  endif
  !
  ! Remove any already existing array
  !
  if(allocated(lines_levelpop)) deallocate(lines_levelpop)
  !
  ! Create the level populations array
  !
  allocate(lines_levelpop(1:lines_nrlevels_subset_max,1:lines_nr_species,1:nrcells),STAT=ierr)
  if(ierr.ne.0) then
     write(stdo,*) 'ERROR: Could not allocate the lines_levelpop() array'
     stop
  endif
  !
  ! Message
  !
  write(stdo,*) 'Computing level populations, and storing them in global array...'
  call flush(stdo)
  !
  ! Checks
  !
  if(lines_mode.eq.1) then 
     if(.not.allocated(lines_partition_temperature)) then
        write(stdo,*) 'ERROR: wanting to compute level populations '
        write(stdo,*) '       according to LTE prescription, but'
        write(stdo,*) '       partition function array not allocated.'
        stop 
     endif
  endif
  if(.not.allocated(gastemp)) then
     write(stdo,*) 'ERROR: Wanting to compute level populations '
     write(stdo,*) '       according to LTE prescription, but'
     write(stdo,*) '       gastemp array not allocated.'
     stop
  endif
  if(.not.allocated(gas_chemspec_numberdens)) then
     write(stdo,*) 'ERROR: Wanting to compute level populations '
     write(stdo,*) '       according to LTE prescription, but'
     write(stdo,*) '       gas_chemspec_numberdens array not allocated.'
     stop
  endif
  if(lines_mode.ge.3) then
     if(.not.allocated(collpartner_numberdens)) then
        write(stdo,*) 'ERROR: Wanting to compute non-LTE level populations but'
        write(stdo,*) '       collpartner_numberdens(:,:) array not allocated.'
        stop
     endif
  endif
  !
  ! Now do the big loop over all cells
  !
  call amr_resetcount()
  call amr_nextcell(index)
  do while(index.ge.0)
     if(index.gt.0) then
        !
        ! Compute the local level populations
        !
        if(lines_mode.eq.1) then
           !
           ! LTE populations
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 call lines_compute_ltepop(ispec,             &
                      lines_nrlevels(ispec),gastemp(index),   &
                      gas_chemspec_numberdens(ispec,index),   &
                      pop(:,ispec))
              endif
           enddo
        elseif(lines_mode.eq.2) then
           !
           ! User-defined populations
           !
           ! Find the mid-cell position
           !
           if(amr_tree_present) then
              !
              ! If the AMR tree is available, then use that 
              !
              a => amr_index_to_leaf(index)%link
              xav = amr_finegrid_xc(a%ixyzf(1),1,a%level)
              yav = amr_finegrid_xc(a%ixyzf(2),2,a%level)
              zav = amr_finegrid_xc(a%ixyzf(3),3,a%level)
           else
              !
              ! For a regular grid, use ixx,iyy,izz
              !
              call amr_regular_get_vertex_ixyz(index,ixx,iyy,izz)
              xav = amr_finegrid_xc(ixx,1,0)
              yav = amr_finegrid_xc(iyy,2,0)
              zav = amr_finegrid_xc(izz,3,0)
           endif
           !
           ! User-defined population calculation
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 call userdef_compute_levelpop(ispec,            &
                      lines_nrlevels(ispec),index,               &
                      xav,yav,zav,                               &
                      gas_chemspec_numberdens(ispec,index),      &
                      pop(:,ispec))
              endif
           enddo
        elseif(lines_mode.eq.3) then
           !
           ! LVG (Sobolev) populations
           !
           ! First determine the velocity gradient
           !
           call lines_compute_velgradient(index,velgrad)
           !
           ! Now compute, with this velocity gradient, the
           ! LVG populations.
           !
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 if(lines_collisions_npartners(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no collision rates are loaded for'
                    write(stdo,*) '        this molecule, ispec=',ispec
                    stop
                 endif
                 if(lines_collisions_npartners_used(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no number densities of collision partners have been'
                    write(stdo,*) '        specified for this molecule, ispec=',ispec
                    stop
                 endif
                 call lines_compute_lvgpop(ispec,lines_nrlevels(ispec),            &
                      lines_nrlines(ispec),lines_collisions_npartners_used(ispec), &
                      lines_collisions_npartmax,gastemp(index),                    &
                      gas_chemspec_numberdens(ispec,index),                        &
                      collpartner_numberdens(:,index),velgrad,                     &
                      pop(:,ispec))
              endif
           enddo
        elseif(lines_mode.eq.4) then
           !
           ! Optically thin non-LTE populations
           !
           ! Now compute the populations
           ! 
           do ispec=1,lines_nr_species
              if(lines_species_fullmolec(ispec)) then
                 if(lines_collisions_npartners(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no collision rates are loaded for'
                    write(stdo,*) '        this molecule, ispec=',ispec
                    stop
                 endif
                 if(lines_collisions_npartners_used(ispec).eq.0) then
                    write(stdo,*) 'ERROR in line transfer: Cannot do non-LTE transfer because'
                    write(stdo,*) '        no number densities of collision partners have been'
                    write(stdo,*) '        specified for this molecule, ispec=',ispec
                    stop
                 endif
                 call lines_compute_optthinpop(ispec,lines_nrlevels(ispec),        &
                      lines_nrlines(ispec),lines_collisions_npartners_used(ispec), &
                      lines_collisions_npartmax,gastemp(index),                    &
                      gas_chemspec_numberdens(ispec,index),                        &
                      collpartner_numberdens(:,index),                             &
                      pop(:,ispec))
              endif
           enddo
        else
           write(stdo,*) 'ERROR: line_mode not known'
           stop
        endif
        !
        ! Now store these populations, but only those of the level
        ! subset. This is meant to save memory. By default, however, the
        ! subset is the full set.
        !
        do ispec=1,lines_nr_species
           if(lines_species_fullmolec(ispec)) then
              do ilevsub=1,lines_nrlevels_subset(ispec)
                 ilevel = lines_levels_subset(ilevsub,ispec)
                 lines_levelpop(ilevsub,ispec,index) = pop(ilevel,ispec)
              enddo
           endif
        enddo
        !
     endif
     call amr_nextcell(index)
  enddo
  !
  ! Deallocate
  !
  deallocate(pop)
  !
end subroutine lines_compute_and_store_local_populations




end module camera_module



